/*
 * ircfr - developed from ourmon src.
 *
 * promiscuous mode filter program.  stores last N IRC JOIN/PRIVMSG
 * messages and then exits.
 *
 * syntax:
 *	ircfr -d -s snaplen -interface -c message_count -f filename bpf_expression
 *
 *	-d debug mode - not background, -f is ignored (stdout is used) 
 *	-s pcap snap length
 *	-i pcap interface 
 *	-c count of PRIVMSG or JOINS - terminate at end of this number
 *	-f must be set unless -d is set.
 *	bpf_expression - kernel bpf filter expression, probably tcp
 *
 *	note: pid is stored in /var/run
 *	program runs in background unless -d is given.
 *
 * what it does:
 *	it writes IRC messages acc. to IRC parsing rules to the output filename.
 *	These may be PRIVMSG or JOIN messages but not PING/PONG etc.
 */

/*
 * ourmon was originally based in part on trafshow.
 * At this point, the heritage is non-obvious but is acknowledged.
 */
/*
 *      Copyright (c) 1993-1997 JSC Rinet, Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*
 * Copyright (c) 2006 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#ifdef LINUX
#include <sys/ioctl.h>
#else
#include <sys/filio.h>
#endif
#ifdef	HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#ifdef BSD
#include <net/bpf.h>
#endif
#include <pcap.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
 
#include "ircfr.h"

void cleanup(int arg);
void ignorehup(int arg);

int snaplen = 256;	/* length of saved portion of packet */

/* global variables */
char *program_name;		/* myself */
char *device_name;		/* network interface name */
char *device2_name;		/* network interface name */
char *conf_name;		/* name of config file */
char *output_name;
FILE *global_fd=NULL;		/* default to stdout */
int  dirset = 0;		/* saw -D -- note this is used to override any settings
				 * in ourmon.conf file
				 */
char *readfile;			/* use a tcpdump save file */
int debug = 0;

int timeout = 1000;		/* ms */

int g_bufsize = 0;		/* pcap/bpf input bufsize */

/* pcap handle */
pcap_t *pd;

int mcount = 0;
int Oflag = 0;

int
main(argc, argv)
	int argc;
	char **argv;
{
	extern int errno;
	int notready = 1;
	int op;
	bpf_u_int32 localnet, netmask;
	char *cp, *infile, *expr, ebuf[PCAP_ERRBUF_SIZE];
	struct bpf_program fcode;
	extern char *optarg;
	extern int optind, opterr;
	void usage(), onalarm(), onwinch(), vers();
	extern int abort_on_misalignment();
	extern pcap_handler lookup_if();
	int cnt;

	/* try to set some sane
	 * defaults
	 */
	localnet = netmask = 0;
	cnt = -1;

	device_name = NULL;

	if ((cp = (char *)strrchr(argv[0], '/')) != NULL)
		program_name = cp + 1;
	else	program_name = argv[0];

	if (abort_on_misalignment(ebuf) < 0) 
		error(0, ebuf);

	global_fd = stdout;		/* defaults to stdout */

	/* 
	 * options:
	 * -s snaplen
	 * -f output filename
	 * -c message count - terminate at this count
	 */
	while ((op = getopt(argc, argv, "s:i:c:f:d")) != EOF) {
		switch (op) {
		case 's':
			snaplen = atoi(optarg);
			break;
		case 'i':
			device_name = optarg;
			break;
		case 'c':
			mcount = atoi(optarg);
			break;
		case 'f':
			output_name = optarg;
			global_fd = fopen(output_name, "w");
		        if (global_fd == NULL) {
				fprintf(stderr, "cannot open %s, due to error: %s\n", output_name, strerror(errno));
				exit(1);
			}
			setbuf(global_fd, NULL);
			break;
		case 'd':
			debug = 1;
			setbuf(stdout, NULL);
			break;
		case 'h':
		default:
			usage();
		}
	}

	/* if no device specified, look for one with pcap_lookupdev
	*/
	if (device_name == NULL &&
	    (device_name = pcap_lookupdev(ebuf)) == NULL)
		error(0, ebuf);

	/* Attach pcap to the network interface 
	 * device_name - device to put in prom. mode
 	 * snaplen - number of bytes to capture.  
	 * note: 
 	 * promiscuous mode set to 1.
	 * note timeout is read timeout in ms (1 sec).
	 * psu mod: pcap_open_live2 passes bpf buffer size.
	 *	might use it here.
	 * note2:
	 * this routine really gets the bpf device up ...
	 */
	if ((pd = pcap_open_live(device_name, snaplen, 1, timeout, ebuf)) == NULL) {
		error(0, ebuf);
	}
		
	/* returns specified snapshot length
	*/
	if ((op = pcap_snapshot(pd)) > snaplen) 
		snaplen = op;

	/* returns link layer type (DLT)
	*/
	op = pcap_datalink(pd);

	/* determine network number and mask
	*/
	if (pcap_lookupnet(device_name, &localnet, &netmask, ebuf) < 0) {
		/* annoying
		warning(0, ebuf);
		*/
	}

	/* set expression pointer, if any
	*/
	expr = copy_argv(&argv[optind]);

	/* bpf compile
	*/
	if (pcap_compile(pd, &fcode, expr, Oflag, netmask) < 0 ||
	    pcap_setfilter(pd, &fcode) < 0)
		error(0, pcap_geterr(pd));

	/* ignored because we do not want to exit at boot
	*/
	(void) signal(SIGHUP, ignorehup);

	/* exit on SIGINT, SIGQUIT
	*/
	(void) signal(SIGINT, cleanup);
	(void) signal(SIGQUIT, cleanup);

	/* open syslog
	*/
        openlog("ircfr", LOG_CONS | LOG_PID , LOG_DAEMON);
        syslog(LOG_ERR, "ircfr is running");

	/* if not debug become a not very complex daemon
	*/
	if (!debug) {
		if (fork() == 0) {
			exit(0);
		}
	}

	storepid();


	/* pcap_loop
	 *	pd - pcap handle
	 *	cnt - # of packets to read, then exit, if -1, will loop forever
 	 *	lookup_if - callback
	 */
	if (pcap_loop(pd, cnt, lookup_if(op), NULL) < 0) {
		syslog(LOG_ERR, "ourmon pcap error: %s", pcap_geterr(pd));
		/*
		error(0, pcap_geterr(pd));
		*/
	}
	syslog(LOG_ERR, "ourmon pcap_loop exit : %s", pcap_geterr(pd));
	pcap_close(pd);

}

/* do whatever after a signal
*/
void
cleanup(int arg)
{
	syslog(LOG_ERR, "ircfr exit - cleanup(%d) called errno %m", arg);
	exit(0);
}

void
ignorehup(int arg)
{
	(void) signal(SIGHUP, ignorehup);
}

void
usage()
{
	fprintf(stderr,
		"Usage:	ircfr -d -s snaplen -i interface -c message_count -f filename bpf_expression\n\
		-d \t\tdebug flag \n\
		-s \t\tsnap length (total) of pkt \n\
		-i name\t\tnetwork interface name; ef0, sl0, ppp0, lo0, etc\n\
		-c number\ttotal irc message count before exit (PRIVMSG and JOIN)\n\
		-f file\t\toutput filename\n\
		bpf expr\t\ttcpdump filter expression\n",
		program_name);

	exit(1);
}

