/*
 * ircscan.c - build up irc info
 * 
 * determine minimal useful snaplength. 256 so far.
 *
 * design note:
 *
 * essentially we are doing 3 things here:
 *
 * 1. extracting a list of channels.  in irc channels are
 * case-insensitive, but this code ignores that, and in point
 * of fact is not 100% correct in terms of its parsing.  This
 * is because we favor stats and speed over perfect accuracy.
 * E.g., we could have two channels both named "foo" with
 * entirely different servers, but we ignore that.  If they
 * both lower-case they are treated as one channel.
 * The goal is to extract the channels, and a set of related ips.
 *
 * 2. we extract a set of IPs doing irc.  we note if an IP
 * appears to be an IRC server based on some very loose notions
 * in the IRC messages (JOIN and PRIVMSG only).
 *
 * 3. we compute raw stats for message counts.  
 *
 * at report time, we ask the syn list for a work weight for any
 * given IP.
 *
 * TBD: it might not hurt at some future data to do the following:
 *
 * 1. sample the destination ports.  At this point we learn nothing
 * about which ports are in use although we could.
 * 2. determine L3 destination counts. (cross module call to scanner module)
 */

/*
 * Copyright (c) 2004, 2005 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <string.h>

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif
#include <pcap.h>

#include "ircfr.h"	/* struct t_entry */
#include "ircscan.h"

static void isjoin(FILE *fd, int, struct in_addr, struct in_addr, 
	unsigned short, unsigned short, char *, int, unsigned char *, int );
static void isping(FILE *fd, struct in_addr, struct in_addr, 
	unsigned short, unsigned short, unsigned char *, int);
static void ispong(FILE *fd, struct in_addr, struct in_addr, 
unsigned short, unsigned short, unsigned char *, int, int);
static void isprivmsg(FILE *fd, int, struct in_addr, struct in_addr, unsigned short, unsigned short, char *, int, unsigned char *, int);

#define BLANK 0x20


/*
 * ircscan: parsing notes
 * overview:
 *	scan L7 payload for JOIN/PING/PONG/PRIVMSG irc messages
 *	goals include:
 *		figure out server vs host
 *		find channel string
 *		calculate IPs in channels
 *		calc some overall stats
 *
 *	Interested in channel names with JOINs/PRIVMSGs, and ip src
 *	relationships with JOIN/PING/PONG.
 *
 * privacy note:  no keys exposed here, and no internal message
 *	guts exposed either.
 *	
 * assumptions:
 *	only called for TCP packets.
 *	only called if there is any L7 payload at all.  not for TCP syns  
 *		without payload.
 *
 * note: CRNL (0xa0xd)
 *
 ****************************************
 * JOIN parsing: 
 *
 * initial:
 * A JOIN may begin the payload, or basically have a string followed by any number
 * of blanks.  This is actually a client->server message or server->client message.
 *
 * post JOIN:
 * any number of blanks, #, &, or :#, followed by channel string.
 * Terminated by blank or CRNL.  CR will do.
 *
 * must extract two things:
 * 1. # or : which indicates client->server or server->client.  E.g.,
 *	#bark indicates client server for channel bark.
 *	:bark indicates server client for channel bark.
 * 2. the channel string itself, sans initial indicator.
 *
 * rough patterns:
 *
 * 1. client to server message
 * ^JOIN[BLANK]#<channel>[BLANK][keystring][BLANK,CRNL]  
 * 2. server to client message
 * :<idstring>[BLANK]JOIN[BLANK]:#<channel>[BLANK,CRNL] 
 *
 * We ignore the keystring at this time.
 * 
 ****************************************
 * PRIVMSG parsing:
 *
 * notes:
 * 1. PRIVMSG like JOIN barring the lack of ":#channel" for server->client.
 * That is, #channel is used, not ":#channel".  
 *
 * 2. PRIVMSG #channel :msg ...   is basic structure of end part of PRIVMSG.
 * We do not care about the msg.
 *
 * once again client/server, server/client cases:
 *
 * rough patterns:
 *
 * 1. client to server message
 * ^PRIVMSG[BLANK]#<channelstring>[BLANK][:msg][BLANK,CRNL]  
 * 2. server to client message
 * :<idstring>[BLANK]PRIVMSG[BLANK]#<channelstring>[BLANK][:msg][BLANK,CRNL]  
 *
 * as with JOIN the existance of the :idstring[BLANK] tells us if we have client/server
 * or server/client.  
 *
 ****************************************
 * PING and PONG parsing (the same):
 *
 * 1. initial may start with
 * ^PING
 * 2. initial may start with 
 * <string>[BLANK]*PING
 *
 * remainder:
 * PING[BLANK]<string>
 * PING[CRNL]
 *
 * TBD: although we do not yet take advantage of this the PONG message
 * can be used for determining server->client messages as they have the
 * form:
 * 
 * :<identity>[BLANK]PONG
 *
 * in a fashion similar to JOIN/PRIVMSG.
 *
 * observations:
 *
 * 1. JOIN and PRIVMSG have same semantic for client to server, server to client 
 * ^JOIN is client to server
 * :id JOIN :#channel  is server to client
 *
 * note however that JOIN in the latter case has a :# semantic.  The channel name
 * does not use the '#'.  PRIVMSG does not use ':'.   Use of ':' with JOIN is currently used.
 * (possibly for no good reason).
 *
 * 2. PONG can be used for server ID and this should be taken advantage of.
 * :identity PONG identity :clientid  --- sent by server (ip src is server).  this is true.
 *
 * In point of fact it looks like PING/PONG may often have the same client/server nature; 
 * ^PING something - from client to server
 * :id PONG :id  - from server to client
 *
 * However this is not true for all PING/PONG messages in the wild.
 *
 *	returns:
 *		1 - if join
 *		2 - if ping
 *		3 - if pong
 *		4 - if privmsg
 *		0 - if not irc
 */
int
ircscan(struct t_entry *te, FILE *fd) 
{
	extern int debug;
	char channame[512];				/* max size of channel name */
	int chansize; 
	char *cpt;
	unsigned int ipsrc, ipdst;
	unsigned short dport;
	unsigned char *p;
	int psize;
	int sflag = 0;					/* ip src is server */

/*
	if (debug) {
		printf("DEBUG: IRCSCAN called \n");
	}
*/
	p     = te->p;
	psize = te->psize;

	/* check for possible match at start of
	 * messge, if so go to state 2 if NONE
	*/
	if (*p == 'J')
		goto state2;
	if (*p == 'P')
		goto state2;

	/* cycle thru NONBLANKS if any
	*/
	while(psize) {
		/* break on blanks
		*/
		if (*p == BLANK)
			goto blanks1;
		p++;
		psize--;
	}
	return(0);

	/* cycle thru BLANKS if any
	*/
blanks1:
	while(psize && (*p == BLANK)) {
		p++;
		psize--;
	}
	if (psize && ((*p == 'J') || (*p == 'P')))
		goto state2;
	/* no match 
	*/
	return(0);

state2:
	/* too small, bail
	*/
	if (psize < 6) {
		return(0);
	}

	/* JOIN ? */
	if ((*p == 'J') && (*(p+1) == 'O') && (*(p+2) == 'I') && (*(p+3) == 'N')) {
		p += 4;
		psize = psize - 4;
		/* too small, bail
		*/
		if (psize == 0)
			return(0);

		/* at this point must have blank
		*/
		if (*p != BLANK)
			return(0);

		/* move to word following blanks
		*/
		while (psize && (*p == BLANK)) {
			psize--;
			p++;
		}
		if (psize == 0) {
			return(0);
		}
		/* now at channel name, save it in local buffer
		if ((*p == '#') || (*p == '&')) {
		*/
		if ((*p == '#') || (*p == '&') || *p == ':') {
			chansize = 0;
			cpt = &channame[0];
			/* skip over any initial separation chars
			*/
			if ((*p == '#') ||  (*p == '&')) {
				psize--; p++;
			}
			else if (*p == ':') {
				/* protocol note:
				 *     JOIN # - indicates client to server
				 *     JOIN : - indicates server to client
				*/
				sflag = 1;		/* ip src is server */
				psize--; p++;
				/* TBD sanity check this
				*/
				if (*p == '#')
					psize--; p++;
			}
			/* rfc 1459
			<chstring> ::=
			    <any 8bit code except SPACE, BELL, NUL, CR, LF and comma (',')>
			*/
			while (psize && (*p != BLANK) && (*p != '\n') && (*p != '\r')) {
				*cpt++ = *p++;
				chansize++;
				if (chansize > 510)
					return(0);
				psize--;
			}
			if (psize == 0) {
				return(0);
			}
			channame[chansize] = 0;
			isjoin(fd, sflag, 
				te->src, te->dst, 
				te->dport, te->sport,
				channame, chansize, te->p, te->psize);
			return(ISJOIN);
		}
		return(0);
	}

	/* PING */
	else 
	if ((*p == 'P') && (*(p+1) == 'I') && (*(p+2) == 'N') && (*(p+3) == 'G')) {
		/* at this point must have blank or CRNL.  PING has at least one word
	         * behind it in all cases, although we are not absorbing it
		*/
		p += 4;
		psize = psize - 4;
		if (psize == 0)
			return;
		if (*p == BLANK) {
			isping(fd, te->src, te->dst, 
				te->dport, te->sport, te->p, te->psize);
			return(ISPING);
		}
		if (psize == 2 && ((*p == '\r') && (*p == '\n'))) {
			isping(fd, te->src, te->dst, 
				te->dport, te->sport, te->p, te->psize);
			return(ISPING);
		}
		return(0);
	}

	/* PONG */
	else 
	if ((*p == 'P') && (*(p+1) == 'O') && (*(p+2) == 'N') && (*(p+3) == 'G')) {
		/* determine if pong is server to client
		*/
		if (p != te->p) {
			sflag = 1;
		}
		/* at this point must have blank or CRNL
		 * PONG has at least one word behind it in all cases.
		*/
		p += 4;
		psize = psize - 4;
		if (psize == 0)
			return(0);
		if (*p == BLANK) {
			/* TBD: not using sflag yet.  check debug info again
			*/
			ispong(fd, te->src, te->dst, 
			te->dport, te->sport, te->p, te->psize, 0);
			return(ISPONG);
		}
		if ((psize == 2) && ((*p == '\r') && (*p == '\n'))) {
			ispong(fd, te->src, te->dst, 
			te->dport, te->sport, te->p, te->psize, 0);
			return(ISPONG);
		}
		return(0);
	}

	/* PRIVMSG - similar to join code above :blah PRIVMSG channel :blah*/
	else
	if ((*p == 'P') && (*(p+1) == 'R') && (*(p+2) == 'I') && (*(p+3) == 'V') && (*(p+4) == 'M') && (*(p+5) == 'S') && (*(p+6) == 'G')) {
		/* check to see if server is sending PRIVMSG
		 * idea: client PRIVMSGs do not have an identity string in front of them. 
		*/
		if (p != te->p) {
			sflag = 1;
		}
		p += 7;
		psize = psize - 7;
		if (psize == 0)
			return(0);
		
		if (*p != BLANK) {
			return(0);
		}

		/* move to word following blanks - this is channel
		*/

                while (psize && (*p == BLANK)) {
                        psize--;
                        p++;
                }
                if (psize == 0) {
                        return(0);
                }

               /* now at channel name, save it in local buffer
                if ((*p == '#') || (*p == '&')) {
                */
                if ((*p == '#') || (*p == '&') || *p == ':') {
                        chansize = 0;
                        cpt = &channame[0];
                        /* skip over any initial separation chars
                        */
                        if ((*p == '#') ||  (*p == '&')) {
                                psize--; p++;
                        }
                        else if (*p == ':') {
                                psize--; p++;
                                if (*p == '#')
                                        psize--; p++;
                        }
			/* rfc 1459
			<chstring> ::=
			    <any 8bit code except SPACE, BELL, NUL, CR, LF and comma (',')>
			*/
			while (psize && (*p != BLANK) && (*p != '\n') && (*p != '\r')) {
                                *cpt++ = *p++;
                                chansize++;
                                if (chansize > 510)
                                        return(0);
                                psize--;
                        }
                        if (psize == 0) {
                                return(0);
                        }
                        channame[chansize] = 0;
			isprivmsg(fd, sflag, te->src, te->dst, 
				te->dport, te->sport, 
				channame, chansize, te->p, te->psize);
			return(ISPRIVMSG);
		}
		return (0);
	}
	return(0);	
}


static void
dumpPayload(FILE *fd, unsigned char *p, int psize)
{
	int i;

	for ( i = 0; i < psize; i++) {
		if (*p == '\r')
			return;
		putc(*p++, fd);
	}
}

/*
 * process join message information.
 *
 * sflag: if true, ip src is irc server
 */

static void
isjoin(FILE *fd, int sflag, 
	struct in_addr ipsrc, struct in_addr ipdst, 
	unsigned short dport,
	unsigned short sport,
	char *chan, int cs, unsigned char *p, int psize)
{
	struct channel *nchan;
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;
	extern int mcount;

	char buf[1024];

	sprintf(buf, "%s", inet_ntoa(ipsrc));
	fprintf(fd, "\tIRCMSG: JOIN: s=%s -> d=%s dport=%d sflag=%d channel=%s clen=%d: p=[", buf, inet_ntoa(ipdst), dport, sflag, chan, cs);
	dumpPayload(fd, p, psize);
	fprintf(fd, "]\n");
	if (mcount == 1) {
		/* mcount will be 0 */
		exit(0);
	}
	else if (mcount == 0) {		/* go forever */
		;
	}
	mcount--;

}

static void
isping(FILE *fd, struct in_addr ipsrc, struct in_addr ipdst, unsigned short dport, unsigned short sport,
	unsigned char *p, int psize)
{
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;

	return;
/*
	if (debug) {
		char buf[1024];
		sprintf(buf, "%s", inet_ntoa(ipsrc));
		printf("\tIRCMSG: PING: s=%s d=%s dp=%d: p=[", buf, inet_ntoa(ipdst), dport);
		dumpPayload(fd, p, psize);
		printf("]\n");
	}
*/
}

static void
ispong(FILE *fd, struct in_addr ipsrc, struct in_addr ipdst, unsigned short dport, unsigned short sport,
	unsigned char *p, int psize, int sflag)
{
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;

	return;
/*
	if (debug) {
		char buf[1024];
		sprintf(buf, "%s", inet_ntoa(ipsrc));
		printf("\tIRCMSG: PONG: s=%s d=%s dp=%d: sflag=%d: p=[", buf, inet_ntoa(ipdst), dport, sflag);
		dumpPayload(fd, p, psize);
		printf("]\n");
	}
*/

}

static void
isprivmsg(FILE *fd, int sflag, struct in_addr ipsrc, struct in_addr ipdst, unsigned short dport, unsigned short sport,
	char *channel, int clen, unsigned char *p, int psize)
{
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;
	extern int mcount;

	char buf[1024];
	sprintf(buf, "%s", inet_ntoa(ipsrc));
	fprintf(fd, "\tIRCMSG: PRIVMSG: s=%s -> d=%s dport=%d sflag=%d, channel=%s clen=%d: p=[", buf, inet_ntoa(ipdst), dport, sflag, channel, clen);
	dumpPayload(fd, p, psize);
	fprintf(fd, "]\n");
	if (mcount == 1) {
		/* mcount will be 0 */
		exit(0);
	}
	else if (mcount == 0) {		/* go forever */
		;
	}
	mcount--;
}
