/*
 * hashblist.c - given a blacklist of IP addresses X, Y, Z
 *	auto store any packets from them into a tcpdump file.
 *
 * config assumptions:
 *
 *	We need to config and turn on a pcap dump capability
 *	We need to parse a list of IP addresses and put them in a hashfile.
 *
 * syntax:
 *	blist_dumpfile maxpackets /usr/dumps
 *		-> filename is /usr/dumps/blacklist.timestamp.dmp
 *		maxpackets may be 0 meaning infinity.
 *	blist_include tag filename (list of ips and ports)
 *	blist_include tag filename (list of ips and ports)
 *	blist_include tag filename (list of ips and ports)
 *	blist_include tag filename (list of ips and ports)
 *	any number of include files, only one trigger file.
 *
 *	config code turns the trigger on 
 *	there is only one trigger file.
 *
 *	we do our own writes as analyze_ip hands us the packet to check.
 *
 */

/*
 * Copyright (c) 2006 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <pcap.h>
#include <string.h>
#include <sys/stat.h>

#include "ourmon.h"
#include "hashblist.h"
#include "sync.h"
#include "barthash.c"		/* inlined function */

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

static
struct blist bl;

static void parseIp(char *, char *, char *, int, char *);

/* called due to blist_dumpfile config directive.
 * create pcap dump fd aimed at dirname/filename
 * config time code:
*/
void
config_blist_dumpfile(char *dirname, unsigned int pktcount)
{
        char ts[100];
        char filename[2100];
        struct stat statbuf;
        extern  int errno;
	extern pcap_t *pd;	/* in ourmon.c */
	char *dfile = "blacklist";
	struct blist *l = &bl;
	char *fn = filename;

	l->count = l->pktcount = pktcount; /* it can't be wrong as it can't be negative */
	l->match = 0;

        /* check the existence of directory 
	*/
        if( stat(dirname, &statbuf) < 0 ) {
                fprintf(stderr,"blist_dumpfile: stat of directory %s failed: error: %s\n", dirname, strerror(errno));
                exit(1);
        }
        if( S_ISDIR(statbuf.st_mode) == 0 )  {
                fprintf(stderr,"blist_dumpfile: dirname %s is not a directory\n", dirname);
                exit(1);
        }

	/* form the name of the dump file
	*/
        bzero(ts, 100);
        bzero(filename, 2100);  
 
        get_ts(ts);
        if( strlen(ts) == 0 ) {
                fprintf(stderr,"blist_dumpfile: get_ts timestamp call failed\n");
                exit(1);
        }

        /* form the config time base filename, which is used
         * at runtime to construct a dynamically instanced unique filename
        */
        if( (strlen(dirname)+strlen(dfile)) > (2048-1) ){
                fprintf(stderr,"blist_dumpfile: dump file name too long\n");
                exit(1);
        }
  
        /* filename = typename.<timestamp>.dmp
        */
        sprintf(filename,"%s/%s.<%s>.dmp", dirname, dfile, ts);

        if((l->pcapfd = pcap_dump_open( pd, (const char *)fn)) == NULL) {
                fprintf(stderr,
                        "blist_dumpfile pcap open failed: %s\n", pcap_geterr(pd));
                exit (1);
        }
	return;
} 

/* insert entries into the hash list from the blacklist file
 *
 * char *tag - one string tag for all the entries in the
 * char *filename - file with list of IP/ports
*/
void
config_blist(char *tag, char *filename)
{
	FILE *fd;
#define MAXBLISTLINE 1024
	char            linebuf[MAXBLISTLINE];
	int ccount; 
	char *opword;
	char           *result;
	int             lineNumber = 1;
	struct blist *l = &bl;
	char opbuf[1024];
	static int firsttime = 1;
	char *sltag;

	/* init hash array first time called
	 * so we can call this again to add more files with lists
	*/
	if (firsttime) {
		firsttime = 0;
		l->entries = 0;
		l->last = NULL;
		bzero (&l->hashlist[0], HSIZE * sizeof(struct bnode *));

		/* init the string list/s for either runtime memory of actions
		 * or config time memory of tag strings
		*/
		initString(&l->sl);
		initString(&l->sltags);
	}

	/* store the tag and later put a pointer to it in all
	 * hash entries
	 */
	sltag = storeString(&l->sltags, tag);
	if (sltag == NULL) {
		fprintf(stderr, "blist_include: ? out of malloc space for tag %s\n", tag);
		exit(1);
	}

	/* open the blacklist config file.  
	*/
	fd = fopen(filename, "r");
	if (fd == NULL) {
		fprintf(stderr, "blist_include open of blacklist config file %s failed\n", filename);
		exit(1);
	}

	/* XXX - this code is duplicated in monconfig.c, fix that somehow.
	 *
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		result = fgets(linebuf, MAXBLISTLINE, fd);
		if (result == NULL)
			break;

		ccount = strlen(linebuf);
		linebuf[ccount-1] = 0;

		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0)) {
			lineNumber++;
			continue;
		}
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */
		(void) parseIp(filename, linebuf, result, lineNumber, sltag);
		lineNumber++;
	} 
	fclose(fd);
	return;
}

static void
parseIp(char *filename, char *linebuf, char *result, int lineNumber, char *sltag)
{
	int rc;
	struct sockaddr_in sa;
	unsigned int ipaddr;
	struct bnode *n;   
	struct bnode *h, *prev;
	unsigned int mod;
	struct blist *l = &bl;
#define MAXLINE 512
	char dc[MAXLINE];
	char v1[MAXLINE];
	unsigned int port;
	extern int errno;

	sscanf(linebuf, "%s %s", dc, v1);

	rc = resolveAddress(dc, &sa);
	if (rc <= 0) {
		fprintf(stderr, "blacklist %s: include file parse: invalid ip address %s: at line %d\n", filename, linebuf, lineNumber);
		exit(1);
	}

	/* any is legal wildcard, represented internally by -1 */
	if (strncmp(v1, "any", 3)==0) {
		port = -1;		/* wildcard */
	} 
	else {
		/* numeric: decide if port is ok
		*/
		port = strtol(v1, NULL, 10);
		if (errno == EINVAL) {
			fprintf(stderr, "parseIp: invalid port: %s:%s\n", filename, linebuf);
			exit(1);
		}
		if (port > 65535 || port < 0) {
			fprintf(stderr, "parseIp: port out of range (0..65535): %s:%s\n", filename, linebuf);
			exit(1);
		}
	}

	/* in network byte order
	*/
	ipaddr = (unsigned int) sa.sin_addr.s_addr;

	/* search the hash/chain list to see if ip address is already inserted
	*/
	n = l->hashlist[mod=(barthash(ipaddr) % HSIZE)];
	for (; n!= NULL ; n=n->next) {
		/* found a match, no need to insert
		*/
		if (n && ( n->ipaddr == ipaddr && (n->port == port))) {
			fprintf(stderr, "warning: filename: %s has duplicate match with: ip/port %s:%s\n",
				filename, dc, v1);
			return;
		}
	}

	/* note this malloc is a config-time malloc only
	*/
	n = (struct bnode *) malloc(sizeof(struct bnode));
	if (n == NULL) {
		fprintf(stderr, "out of space with malloc in hashblist.c\n");
		exit(1);
	}
	n->next = NULL;
	n->tag = sltag;
	n->ipaddr = ipaddr;
	n->port = port;
	l->entries += 1;

	/* insert the pnode in the hash list
	*/
	h = l->hashlist[mod];
	if (h == NULL) {
		l->hashlist[mod] = n;   
	}
	else {
		n->next = h;
		l->hashlist[mod] = n;
	}
}

/* runtime: packet analysis routine
 * see if pkt should be shoved into tcpdump file
 */
void
search_blist(struct pcap_pkthdr *h, u_char *p, struct t_entry *te)
{
	struct bnode *n;
	int match = 0;
	struct blist *l = &bl;
	int mod;
	extern int debug;

	/* if the src or dst address in the packet is mentioned in the blacklist
	 * if this is the 1st time this has happened, we open a trigger file,
	 * and store any blacklist packets in that particular tcpdump
	*/
	unsigned int ipsrcaddr = te->src.s_addr;
	unsigned int ipdstaddr = te->dst.s_addr;
	unsigned short srcport = te->sport;
	unsigned short dstport = te->dport;
	unsigned short ipproto = te->proto;

	/* match if address is src address in packet AND
	 * either port if port not -1
	 * TBD:
	 * this is a little fuzzy and perhaps should be 
	 * ipsrc and src port or ipdst and dst port.
	*/
	n = l->hashlist[mod=(barthash(ipsrcaddr) % HSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ((n->ipaddr == ipsrcaddr) && ((n->port == -1) ||
		((n->port == te->sport) || (n->port == te->dport))))) {
			match = 1;
			break;
		}
	}
	if (!match) {
		n = l->hashlist[mod=(barthash(ipdstaddr) % HSIZE)];
		for (; n!= NULL ; n=n->next) {
			if (n && ((n->ipaddr == ipdstaddr) && ((n->port == -1) 
			||
			((n->port == te->sport) || (n->port == te->dport))))) {
				match = 1;
				break;
			}
		}
	}
	if (debug) {
		if (match)
			printf("BLACKLIST: match %d ip %x:%x s=%d d=%d\n", 
				match, ipsrcaddr, ipdstaddr, srcport, dstport);
	}

#ifdef THREAD
	spinlock(&locks[LOCK_BLIST]); 
#endif
	if (match) {
		l->match = 1;
		/* write packet via blist->pcapfd, te->h, te->op 
		*/
		if (l->pktcount==0 || l->count--) {
			pcap_dump((u_char *)l->pcapfd, h, p);
		}
		/* memorize what happened
		*/
		{
#define BUFSIZE 1024
			char buf[BUFSIZE];
			char addr1[BUFSIZE];
			int rc;
			rc = sprintf(addr1,"%s", intoa(ipsrcaddr));
			addr1[rc] = 0;
			rc = sprintf(buf, "(%s):%s.%d->%s.%d[%d]", 
				n->tag, addr1, srcport, 
				intoa(ipdstaddr), dstport, ipproto);
			buf[rc] = 0;
			storeString(&l->sl, buf);
		}
	}
#ifdef THREAD
 	unlock(&locks[LOCK_BLIST]);
#endif
} 

void 
print_blist(FILE *monfd)
{
	struct blist *l = &bl;

	if (l->match) {
		/* elog message if desired
		fprintf(monfd, "elog: IP blacklist storage has occured \n");
		*/
		dumpString(&l->sl, monfd, "elog: IP blacklist event:");
		zeroString(&l->sl);
		initString(&l->sl);
		
		l->match = 0;
	}
	/* flush the trigger file so it is in sync every 30 
	 * seconds
	*/
	pcap_dump_flush(l->pcapfd);
} 
