/*
 * hashsyn.c - count tcp syns with mention of rst/fin on a per ip src basis. 
 */

/*
 * Copyright (c) 2004 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <pcap.h>
#include <string.h>

/* use filter.h which is more general because we access
 * topn_syn in here
 */

/* 
#include "hashsyn.h"
*/
#include "ourmon.h"
#include "filter.h"
#include "trigger.h"
#include "bytecodes.h"
#include "barthash.c"
 
#ifdef SCANMESH
void scanmesh_candidate(struct synnode *);
#endif

/* email MTA ports of interest
*/
#define EMAILSCHECK (te->sport == 25 || te->sport == 587 || te->sport == 465)
#define EMAILDCHECK (te->dport == 25 || te->dport == 587 || te->dport == 465)

extern int debug;

/* hardwired for now */
int email_topn_count = 10;

static unsigned int websizes[6]; 

/* format string for printing out syn tuple
 * N items PLUS port tuples. includes port count.
*/
#define SYNFORMAT "%s:%u:%u:%u:%u:%u:%u:%u:%u:%x:%x:%x:%u:%u:%u:%u:%u:%s:%u:%d:%u:%u:%u:%u:%u:" 

void 
printSynTuple(FILE *fd, struct synnode *head, int newline) 
{
	char buf[1024];
	int z;

	snprintf(buf, 1024, "%s", intoa(head->dstip));

	fprintf(fd, SYNFORMAT,
		intoa(head->ipsrc), 			/* 0 */
		head->syncount, 			/* 1 */
		head->sacount,			/* 2 */
		head->fincount,			/* 3 */
		head->finssent,			/* 4 */
		head->rstcount,			/* 5 */
		head->tcpfrom,			/* 6 */
		head->tcpto,				/* 7 */
		head->icmpcount, 			/* 8 */
		head->flags, 				/* 9 */
		head->apptags.lower, 			/* 10 */
		head->apptags.upper, 			/* 11 */
		head->emulehits, 			/* 12 */
		head->ipds,				/* 13 */
		head->tpds, 	            		/* 14 */
		head->scount,             		/*15*/
		head->sports[0],          		/*16*/
		buf,          				/*17 - sampled src */
		head->p25_syn,       			/*18 - p25 syn count */
		head->p25ww,       			/*19 - p25 ww */
		head->sizes[0],       		/* 20 */
		head->sizes[1],       		/* 21 */
		head->sizes[2],       		/* 22 */
		head->sizes[3],       		/* 23 */
		head->pcount);			/* 24 */

	/* note: port tuple comes after this		  24 */
	for (z = 0; z < head->pcount; z++) {
		fprintf(fd, "%u,%u,", head->dports[z], head->dpcounts[z]);
	}

	if (newline)
		fprintf(fd, ":\n");
	else
		fprintf(fd, ":");
}

static
void __inline
check_p2p(struct synnode *n, struct t_entry *te, int ircresult)
{

	extern struct topn_syn topn_syn;
	/* irc check -- based on simple lex/scanning -- see ircscan.c
	 * data is stored here.
	*/
	if (ircresult) {
		if (ircresult == 1) {
			n->flags |= (IRCJOINS);
			te->flags |= (IRCJOINS);
		}
		else if (ircresult == 2) {
			n->flags |= (IRCPINGS);
			te->flags |= (IRCPINGS);
		}
		else if (ircresult == 3) {
			n->flags |= (IRCPONGS);
			te->flags |= (IRCPONGS);
		}
		else if (ircresult == 4) {
			n->flags |= (IRCPRIVMSG);
			te->flags |= (IRCPRIVMSG);
		}
		/*
		return;
		*/
	}
	/* bittorrent/gnutella etc... using bcmp at start of L7 payload
	 * if you have done this once per sample period do NOT
	 * do it again.
	*/
	if (te->psize >= 20) {
		/* bittorrent flow?  etc.
		*/
		if (((n->flags & BITTORRENT) == 0) && (bcmp(te->p, btcode, 20) == 0)) {
			n->flags |= (BITTORRENT);
			te->flags |= (BITTORRENT);
		}
		else if (((n->flags & GNUTELLA) == 0) && (bcmp(te->p, gnucode, 17) == 0)) {
			n->flags |= (GNUTELLA);
			te->flags |= (GNUTELLA);
		}
		else if (((n->flags & MORPHEUS) == 0) && (bcmp(te->p, morphcode, 20) == 0)) {
			n->flags |= (MORPHEUS);
			te->flags |= (MORPHEUS);
		}
		else if (((n->flags & KAZAA) == 0) && (bcmp(te->p, kazaacode, 20) == 0)) {
			n->flags |= (KAZAA);
			te->flags |= (KAZAA);
		}
	}
	/* edonkey check.  binary statistical guess
	 * it still is sucky and has too many false positives
	*/
	if (((n->flags & EDONKEY)==0)&&(te->psize >= 6)) {
		unsigned char c = *te->p;
		unsigned char c2 = *(te->p+5);	/* message type */

		/* 2nd byte chosen on basis of brief stat. study
		*/
		/* if edonkey payload byte */
		if (c == 0xe3) {
			/* 0x01 is hello, client-server
			 * 0x46 is msg sending part, client-client
			 */
			if (c2 == 0x01) { 
				n->emulehits++;
				if (n->emulehits > 2) {
					n->flags |= (EDONKEY);
					te->flags |= (EDONKEY);
				}
				/*
				return;
				*/
			}
		}
	}
}

static void  __inline
storePort( struct synnode *n, int dport);

static void  __inline
storeSPort( struct synnode *n, int sport);

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

#ifdef TESTMAIN
char *intoa(unsigned int addr);
#endif

/* frequency analysis for edonkey message bytes
*/
static int ede3[256];	/* store message byte after 1st e3 in payload */
static int edc5[256];	/* ditto on c5 */
static int edd4[256];   /* ditton on d4 */


/* called before list can be used 
 */
void init_synlist (struct synlist *l) 
{
	l->entries = 0;
	l->last = NULL;
	bzero (&l->hashlist[0], THSIZE * sizeof(struct synnode *));
	websizes[0] = websizes[1] = websizes[2] = websizes[3] = 0;
	websizes[4] = websizes[5] = 0;
}


/*
 * search_srcsyn - hash list lookup
 *
 * If found we increment various counters.
 *
 * if icmpflag is set, we have an icmp packet as the src packet, not a tcp packet.  
 * in that case, we wish to only increment the icmp error counter associated
 * with the destination IP address.	
 *
 * saflag means we should directly increment the sa counter.
 *
 * ircflag is a hint from the L7 irc parser about setting the irc bitflag.
 *
 * basic idea:
 *	1st deal with fin/reset/icmp in terms of lookup by destination address.
 *	icmp does not cause an insert.
 */ 
void
search_srcsyn(struct t_entry *te, struct synlist *l, int saflag, int ircresult, int icmpflag)
{
	struct synnode *n;
	int cflag = 0;
	int doinsert = 1;		/* assume true */
	struct synnode *h, *prev;
	unsigned int mod;
	int i;
	unsigned int ipsrcaddr = te->src.s_addr;
	unsigned int ipdstaddr = te->dst.s_addr;
	int result;

	/* first search on the ip destination address to increment 
	*  certain reverse counters
	*/
/*
	n = l->hashlist[mod=(ehash((unsigned char *)&ipdstaddr, 4) % THSIZE)];
*/
	n = l->hashlist[mod=(barthash(ipdstaddr) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipdstaddr)) {
			if (te->finflag) {
				n->fincount++;	/* fin returned to this ip source */
			}
			else if (te->resetflag) {
				n->rstcount++;	/* reset returned to this ip source */
				/* email mta ports may be 25, 587, 465 */
				if (EMAILSCHECK) {
					n->p25_rst++;
				}
			}
			else if (icmpflag) {
				n->icmpcount++;	/* icmp errors sent to this ip src */
			}

			if (!icmpflag) {
				n->tcpto++;	/* tcp pkts to this ip src */
				if (EMAILSCHECK) {
					n->p25_total++;
				}
			}

#ifdef MATCHDST
			/* add any new application flags detected in payload 
			 * this is done for the source and destination 
			 * perform pattern matching on payload - 
			 * only if there is payload
			 */
			if ((te->psize) &&  (patterns != NULL)) {
				result=patmatch_matchWithHistory(
					te->proto,	/* ip protocol */
					&n->apptags,	/* node tags field */
					te->p,		/* payload */
					patterns,	/* global pats */
					te->psize,	/* size of payload */
					&te->attempted,  /* match attempts */
					&te->appresult,  /* matches found */
					n->tcpfrom+n->tcpto,  /* pkt cnt */
					n->bytecount);   /* byte count */
				if (result==1) {
					syslog(LOG_ERR, 
					"error in pattern match of payload");
				}
			}
#endif

			break;
		}
	}

	/* end of icmp portion, tcp packets only from here on.
	 * in addition, we don't have any more screwball ideas about
	 * ip src vs ip dst.  we are interested in the packet from the ip src
	 * point of view  ONLY.
	*/
	if (icmpflag)
		return;

	/* web packet binning for size
	*/
	if ((te->sport == 80) || (te->sport == 443) || (te->dport == 80) || (te->dport == 443)) {
		websizes[te->sizeindex]++;
	}

	/* search the hash/chain list on the ip src.
	*/
/*
	n = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
*/
	n = l->hashlist[mod=(barthash(ipsrcaddr) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipsrcaddr)) {
			doinsert = 0;		/* we found it, we don't insert it */
			if (te->honeyflag) {
				n->flags |= (HONEYNET);
				te->flags |= (HONEYNET);
			}
			if (te->darkflag) {
				n->flags |= (DARKNET);
				te->flags |= (DARKNET);
			}
			n->sizes[te->sizeindex]++;         /* packet size histogram */
			n->tcpfrom++;		/* it's a tcp packet from the src */
			if (EMAILDCHECK) {
				n->p25_total++;	
			}	
			n->bytecount+=te->psize;
			if (te->synflag) {
				n->syncount++;
				/* if syn sent to port 25
				*/
				if (EMAILDCHECK) {
					n->p25_syn++;
				}
				/* if a syn with an ACK coming from an ip_src, then
				 * we count as syn/ack as well
				 */
				if (saflag) {
					n->sacount++;
				}
			}
			else if (te->finflag) {
				n->finssent++;	/* fin sent from this ip src */
				if (EMAILDCHECK) {
					n->p25_fin++;
				}
			}
			else if (te->resetflag) {
				; 
			}
			/* l7 connection checks on data packets only
			*/
			else {
				/* older but faster non PCRE checks
				*/
				check_p2p(n, te, ircresult);
			}
			/* a few more flag checks - only check for these if the user
			 * did not override the flag in the configuration file
			*/
			if ((te->sport == 80 || te->sport == 443)) {
				n->flags |= (WEBSERVER);
				te->flags |= (WEBSERVER);
			}	
			else if (EMAILDCHECK) {
				n->flags |= (EMAIL);
				te->flags |= (EMAIL);
			}
			/* do port sampling
			*/
			storePort(n, te->dport);
			storeSPort(n, te->sport);

			/* pattern match check -- see previous comment
			*/
			if ((te->psize) &&  (patterns != NULL)) {
				result=patmatch_matchWithHistory(
				te->proto,
				&n->apptags,
				te->p,
				patterns,
				te->psize,
				&te->attempted,
				&te->appresult,
				n->tcpfrom+n->tcpto,
				n->bytecount);
				if (result==1) {
					syslog(LOG_ERR, 
					"error in pattern match of payload");
				}
			}
			break;
		}
	}

	/* we found it therefore do not insert it */
	if (doinsert == 0) {
		return;
	}

	/* insert needed */

	n = (struct synnode *) malloc(sizeof(struct synnode));
	if (n == NULL) {
		syslog(LOG_ERR, "malloc failure in insert_srcsyn, entry/count %d",
			l->entries);
		return;
	}
	bzero(n, sizeof (struct synnode));
	l->entries += 1;
	n->tcpfrom = 1;
	n->ipsrc = ipsrcaddr;
	n->dstip = ipdstaddr;		/* one sample only */
	n->bytecount+=te->psize;
	n->p25_syn = n->p25_fin = n->p25_rst = n->p25_total = 0;
	n->sizes[0] = n->sizes[1] = n->sizes[2] = n->sizes[3] = 0;
	n->sizes[4] = n->sizes[5] = 0;
	n->sizes[te->sizeindex]++;         /* packet size histogram */
	/* mark ports unused */
	for ( i = 0; i < MAXPORTS; i++) {
		n->dports[i] = -1;
	}
	for ( i = 0; i < MAXPORTS; i++) {
		n->sports[i] = -1;
	}
	if (te->synflag) {
		n->syncount = 1;
		if (te->dport == 25) {
			n->p25_syn++;
			n->p25_total = 1;
		}
		if (saflag) {
			n->sacount = 1;
		}
		storePort(n, te->dport);	
		storeSPort(n, te->sport);	
	}
	else if (te->finflag) {
		n->finssent = 1;
		if (te->dport == 25) {
			n->p25_fin++;
			n->p25_total = 1;
		}
		n->tcpto = 1;
		storePort(n, te->dport);	
		storeSPort(n, te->sport);	
	}
	else if (te->resetflag) {
		/* not doing anything with resets */
	}
	else /* data */ {
		storePort(n, te->dport);	
		storeSPort(n, te->sport);	
		check_p2p(n, te, 0);
	}

	if (te->honeyflag) {
		n->flags |= (HONEYNET);
		te->flags |= (HONEYNET);
	}

	if (te->darkflag) {
		n->flags |= (DARKNET);
		te->flags |= (DARKNET);
	}

	/* PCRE patmatch check
	*/
	if ((te->psize) &&  (patterns != NULL)) {
		result=patmatch_matchWithHistory(
		te->proto,
		&n->apptags,
		te->p,patterns,
		te->psize,
		&te->attempted,
		&te->appresult,
		n->tcpfrom+n->tcpto,
		n->bytecount);
		if (result==1) {
			syslog(LOG_ERR, "error in pattern match of payload");
		}
	}

	/* insert the synnode in the hash list
	*/
	n->next = NULL;
/*
	h = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
*/
	h = l->hashlist[mod=(barthash(ipsrcaddr) % THSIZE)];
	if (h == NULL) {
		l->hashlist[mod] = n;
	}
	else {
		n->next = h;
		l->hashlist[mod] = n;
	}
}

/*
 * extract work weight associated with ip src address
 */
int
searchSyn(struct synlist *l, unsigned int ipsrcaddr)
{
	struct synnode *n;
	unsigned int mod;
	int t1;

	/* first search on the ip destination address to increment certain reverse counters
	*/
/*
	n = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
*/
	n = l->hashlist[mod=(barthash(ipsrcaddr) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipsrcaddr)) {
			t1 = n->tcpfrom + n->tcpto;
			if (t1 == 0)
				return(0);
			/* return the work weight as an integer between 0..100 */
			t1 = ((n->syncount + n->finssent + n->rstcount) / (t1 * 1.0)) * 100;                        
			return(t1);
		}
	}
	return(0);
}

/* we know an ip src address - we want a syn tuple associated with it
*/
struct synnode *
findSynTuple(struct synlist *l, unsigned int ipsrcaddr)
{
	struct synnode *n;
	unsigned int mod;
	int t1;

	/* first search on the ip destination address to increment certain reverse counters
	*/
/*
	n = l->hashlist[mod=(ehash((unsigned char *)&ipsrcaddr, 4) % THSIZE)];
*/
	n = l->hashlist[mod=(barthash(ipsrcaddr) % THSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipsrcaddr)) {
			return(n);
		}
	}
	return(NULL);
}

static void __inline
storePort(struct synnode *n, int dport)
{
	int i;

	for ( i = 0; i < MAXPORTS; i++) {
		/* have not seen the port and we have room,
		 * therefore store it and increment counter and return
		 */
		if (n->dports[i] == -1) {
			n->dports[i] = dport;
			n->dpcounts[i] = 1;
			n->pcount++;
			return;
		}
		/* found it, increment count, and return
		*/
		if (n->dports[i] == dport) {
			n->dpcounts[i]++;
			return;
		}
	}
	/* didn't find it, must not be room, 
	 * too bad 
	 */
	return;
}

static void __inline
storeSPort(struct synnode *n, int sport)
{
	int i;

	if (n->scount == MAXPORTS)
		return;
	for ( i = 0; i < MAXPORTS; i++) {
		/* have not seen the port and we have room,
		 * therefore store it and increment counter and return
		 */
		if (n->sports[i] == -1) {
			n->sports[i] = sport;
			/*
			n->spcounts[i] = 1;
			*/
			n->scount++;
			return;
		}
		/* found it, increment count, and return
		*/
		if (n->sports[i] == sport) {
			/*
			n->spcounts[i]++;
			*/
			return;
		}
	}
	/* didn't find it, must not be room, 
	 * too bad 
	 */
	return;
}

/*
 * deallocate malloced hash entries
 */

void
destroy_synlist(struct synlist *l) 
{
	int i;
	struct synnode *n;
	struct synnode *prev;
	for (i =0; i< THSIZE; i++) {
		n = l->hashlist[i];
		while(n != NULL) {
			prev = n;
			n = n->next;
			free(prev);
		}
	}
}

#ifdef TESTMAIN
static
dump_hash(struct synlist *l) {
	struct synnode *h;
	int counters[THSIZE];
	int i;
	int ls = 0;
	printf("No. of entries %d\n", l->entries);
	bzero(&counters, sizeof(int)*THSIZE);
	printf("\nhashdump---------\nlist entries %d\n",l->entries);
	for (i=0;i<THSIZE;i++) {
		h = l->hashlist[i];
		while(h != NULL) {
			ls = 1;
			printf("%d:", i);
			printf("ip src %s", intoa(h->ipsrc));
			printf(":s/f/r: %d %d %d", h->syncount,
				h->fincount, h->rstcount);
			counters[i]++;
			h = h->next;
		}
		if (ls) {
			printf(":%d \n",i);
		}
		ls = 0;
	}
	printf("\n");
	for (i = 0; i<THSIZE; i++) {
		if (counters[i] != 0) {
			printf("counters[%d]:%d buckets\n", i, counters[i]);
		}
	}
	printf("end hash dump---------\n");
}
#endif

/* pack the hashlist into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */
static
struct synnode *
pack_array(struct synlist *l, int scanflag, struct slist *isl, struct slist *tsl)
{
	struct synnode *head, *n;
	unsigned int c = l->entries;  /* count of hash entries allocated */
	int i = 0;
	int z;
	int hcounter = 0;
	WEIGHTCOUNTER t1;
	head = (struct synnode *) malloc(c * sizeof (struct synnode));

	if (head == NULL) {
		syslog(LOG_ERR, "malloc failure in syn/pack_array, count %d",
			l->entries);
		return (NULL);
	}
	bzero(head, c * sizeof(struct synnode));

	/* walk the hashlist
	*/
	for (i =0 ; i<THSIZE; i++) {
		n = l->hashlist[i];
		/* if n is non-nulln use, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			head[hcounter].ipsrc    = n->ipsrc;
			head[hcounter].dstip    = n->dstip;
			head[hcounter].syncount = n->syncount;
			head[hcounter].sacount  = n->sacount;
			head[hcounter].fincount = n->fincount;
			head[hcounter].finssent = n->finssent;
			head[hcounter].rstcount = n->rstcount;
			head[hcounter].tcpfrom  = n->tcpfrom;
			head[hcounter].tcpto    = n->tcpto;
			head[hcounter].icmpcount = n->icmpcount;
			head[hcounter].pcount   = n->pcount;
			head[hcounter].scount   = n->scount;
			head[hcounter].flags    = n->flags;
			head[hcounter].apptags = n->apptags;
			head[hcounter].emulehits = n->emulehits;
			/* no programmatic change as of yet
			 * 50 and 100 are combined to make 100
			 * 150 and 500 are combined to make 500
			*/
			head[hcounter].sizes[0] = n->sizes[0] + n->sizes[1];
			head[hcounter].sizes[1] = n->sizes[2] + n->sizes[3];
			head[hcounter].sizes[2] = n->sizes[4];
			head[hcounter].sizes[3] = n->sizes[5];
			if (scanflag) {
				head[hcounter].ipds = scanLookup(isl, head[hcounter].ipsrc);
				head[hcounter].tpds = scanLookup(tsl, head[hcounter].ipsrc);
			}
			/* compute p25 work weight
			*/
			head[hcounter].p25_syn = n->p25_syn;
			head[hcounter].p25_fin = n->p25_fin;
			head[hcounter].p25_rst = n->p25_rst;
			head[hcounter].p25_total = n->p25_total;
			/* note this value is an int
			*/
			if (head[hcounter].p25_total) {
				head[hcounter].p25ww = 
					(int) (((n->p25_syn + n->p25_fin + n->p25_rst) * 100)/
					n->p25_total);  
				if (head[hcounter].p25ww > 100) {
					head[hcounter].p25ww = 100;
				}
			} 
			else {
				head[hcounter].p25ww = 0;
				head[hcounter].p25_syn = 0;
			}

			t1 = n->tcpfrom + n->tcpto;
			/* compute the overall work weight
			*/
			if (t1 == 0) {
				head[hcounter].ww = head[hcounter].work = 0;
			}
			else {
				head[hcounter].ww = head[hcounter].work = 
					(n->syncount + n->finssent + n->rstcount) / t1;                        
				/* sort with w=100% (w varies from 0..1 here) by syn count by simply adding the syncount back in.
				 * it would be possible to do this within all weights by scaling
				 * the syncount by an absolute max. syns per second but it isn't
				 * clear as to whether or not that is needed.
				*/
				if (head[hcounter].work >= 1) {
					head[hcounter].work += n->syncount;
				}
			}

			for (z = 0; z < n->pcount; z++) {
				head[hcounter].dports[z] = n->dports[z]; 
				head[hcounter].dpcounts[z] = n->dpcounts[z];
			}
			for (z = 0; z < n->scount; z++) {
				head[hcounter].sports[z] = n->sports[z]; 
				head[hcounter].spcounts[z] = n->spcounts[z];
			}
			n = n->next;
			hcounter++;
		}
	}  
	return(head);
}

static
int
sort_by_syncount(e1,e2)
struct synnode *e1, *e2;
{
        if (e1->syncount > e2->syncount) return -1;
        if (e1->syncount < e2->syncount) return 1;
        return 0;
}

static
int
sort_by_25syncount(e1,e2)
struct synnode *e1, *e2;
{
        if (e1->p25_syn > e2->p25_syn) return -1;
        if (e1->p25_syn < e2->p25_syn) return 1;
        return 0;
}

static
int
sort_by_work(e1,e2)
struct synnode *e1, *e2;
{
        if (e1->work > e2->work) return -1;
        if (e1->work < e2->work) return 1;
        return 0;
}

/*
static
sort_by_scalar(e1,e2)
int *e1, *e2;
{
        if (*e1 > *e2) return -1;
        if (*e1 < *e2) return 1;
        return 0;
}
*/

/* given a synlist, sort it, and print it out (count entries max)
 *
 * inputs:
 *      fd - file for print output
 *      l - hash list
 *      count - max desired outputs for all filters
 */
void
print_synlist(FILE *fd, struct synlist *l, int count, int wflag, 
char *wormfile, int ipaflag, 
int sflag,
struct slist *isl, struct slist *tsl)
{
	extern struct topn_syn topn_syn;
	struct synnode *head;
	int i,k;
	int z;
	int c = l->entries;  /* no of hash elements */
	int us, them, wcount, wormcount, p2pcount;
	FILE *wfd = NULL;
	FILE *p2pfd = NULL;
	FILE *sfd = NULL;
	FILE *efd = NULL;
	FILE *potfd = NULL;
	int tcpworm;
	COUNTER ipdst_hits = 0;	
	COUNTER tcpdst_hits = 0;
	int usflag;
	COUNTER totalsizes[4];
	COUNTER totalp2p[4];

	totalsizes[0] = totalsizes[1] = totalsizes[2] = totalsizes[3] = 0;
	totalp2p[0] = totalp2p[1] = totalp2p[2] = totalp2p[3] = 0;

	/* these two things have to come out of a loop
	*/
	unsigned int usnet, usmask; 
	
	/* open tcpworm file for writing
	*/
	if (wflag) {
		wfd = fopen(wormfile, "w");
		if (wfd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", wormfile);  
			exit(1);
		}
	}
	us = them = wcount = p2pcount = wormcount = 0;

	/* open p2p file for writing
	*/
	if (topn_syn.p2pflag) {
		p2pfd = fopen(topn_syn.p2pfile, "w");
		if (p2pfd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", topn_syn.p2pfile);  
			exit(1);
		}
	}

	/* open candidate enterprise-only syndump for writing
	*/
	if (topn_syn.syndump_flag) {
		sfd = fopen(topn_syn.syndump, "w");
		if (sfd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", topn_syn.syndump);  
			exit(1);
		}
	}

	/* open email activity dump for writing
	 * we deem this important because of the possibility of
	 * machines generating spam due to exploits/botnets, etc.
	*/
	if (topn_syn.emailflag) {
		efd = fopen(topn_syn.emaildump, "w");
		if (efd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", topn_syn.emaildump);  
			exit(1);
		}
	}
	if (topn_syn.potflag) {
		potfd = fopen(topn_syn.potdump, "w");
		if (potfd == NULL) {
			syslog(LOG_ERR, "could not open %s:(%m)", 
				topn_syn.potdump);  
			exit(1);
		}
	}

	/* form the hashlist into a packed array
	*/
	head = pack_array(l, sflag, isl, tsl);
	if (head == NULL) {
		return;
	}
	/* quicksort it either by syn count or by adjusted worms at the top syncount
	 * sortbysyn is the default
	*/

	if (topn_syn.sortbysyn) {
		qsort((void *)head, c, sizeof(struct synnode), sort_by_syncount);
	}
	else {
		qsort((void *)head, c, sizeof(struct synnode), sort_by_work);
	}

	/* start mon.lite syn_list
	*/
	fprintf(fd, "syn_list: %u : ", l->entries); 

	/* loop thru the syn list and get the top N synners.
	*/
	for (i = 0; i < l->entries; i++) {

		/* calculate tcp-based p2p mon.lite stats
		 * totalsizes gives us all TCP using IP sources.
		 * p2p sizes gives us all TCP using IP sources who are doing p2p.
		*/

		totalsizes[0] += head[i].sizes[0];
		totalsizes[1] += head[i].sizes[1];
		totalsizes[2] += head[i].sizes[2];
		totalsizes[3] += head[i].sizes[3];
		if ((head[i].ipds > 10) && (head[i].tpds > 10) && (head[i].scount == 10)) {
			totalp2p[0] += head[i].sizes[0];
			totalp2p[1] += head[i].sizes[1];
			totalp2p[2] += head[i].sizes[2];
			totalp2p[3] += head[i].sizes[3];
		}

		/* produce top N syn tuple, stop at count
		 * this goes into mon.lite file.
		*/
		if ( i < count) {
			/* print to mon.lite
			*/
			printSynTuple(fd, head+i, 0);
		}

		/*
		 * from looking at the syndump.txt file (all syn list entries sorted)
		 * it appears that 20-40 is a reasonable range.  
		 * WORMWEIGHT
		 */

		/* syncount is unsigned and might be zero and we also want to catch cases of
		 * all fins but no syns, as well as no syns/fins, but we have resets.
		 */
		usflag = tcpworm = 0; 
		if (head[i].syncount > head[i].fincount) {
			if (((head[i].syncount - head[i].fincount) > topn_syn.wormdiff) ) {
				tcpworm = 1;
			}
		}
		/* mucho FINS sent, but not so much syns sent: nmap -sF */
		else if ((head[i].syncount <= head[i].finssent)) {
			if ((head[i].finssent - head[i].syncount) > topn_syn.wormdiff)  
				tcpworm = 1;
		}
		/* more than wormdiff packets sent and there exist resets.  this catches
		 * nmap scans with acks or just pushes(data) but no control
		 * TBD: possibly better if: rstcount > wormdiff and that's it?
		 * this version is looser.
		 */
		else if ((head[i].syncount == 0) && (head[i].rstcount > topn_syn.wormdiff)
			&& (head[i].tcpfrom > topn_syn.wormdiff)) {
				tcpworm = 1;
		}

		/* note if ip belongs to the home subnet or not
		*/
		if (ipaflag) {
			usflag = ishome(head[i].ipsrc);
		}
	
		if (tcpworm) {
			/* if the true work weight is greater than the config pre-filter
			 * weight, then we count things in tworm.  
			 * default: topn_syn.weight is 0.
			 */
			if ((int)(head[i].ww * 100) >= topn_syn.weight) { 
				wcount++;	/* count of worms acc. to topn_syn.weight */
				if (ipaflag) {
					if (usflag) {
						us++;
					}
					else  { 
						them++;
					}
				}
			}

			/* print to wormfile
			*/
			printSynTuple(wfd, head+i, 1);
			wormcount++;	/* total count of subset hosts in tcpworm.txt */
			/* if scanmesh is configured, hand it the
			*  syntuple to do with as it will
			*/
#ifdef SCANMESH
			if (scanmesh.configured) {
				if (debug)
					printf("SCANMESH: SCANMESH candidate call\n");
				scanmesh_candidate(head + i);
			}
#endif
		}
		/* if p2p file, write p2p tuples into it but only for home subnet
		*/
		if (p2pfd && ipaflag) {
			if (usflag && head[i].flags & P2PBITS) {
				printSynTuple(p2pfd, head+i, 1);
				p2pcount++;
			}
		}

		/* if any "potnet" violator put them in a file
		*/
		if (potfd) {
			if (head[i].flags & (DARKNET|HONEYNET)) {
				printSynTuple(potfd, head+i, 1);
			}
		}

		/* this may be viewed as a mini syndump.  it will include local web servers, web clients,
		 * attackers, p2p'ers, etc.  
		 * note: still sorted syn order
		 * 	all app/flag signature info is included (as not excluded)
		 * we do some filtering here
		 *	1. ip must belong to home team, this reduces fanout by an order of magnitude
		 *	2. must have 1 entry in L4 port dst array (this rules out some weak hosts)
		 *	3. also must have > 1 ip dst.  
		 */
		if (sfd && ipaflag) {
			if (usflag && head[i].pcount && head[i].ipds ) {
				printSynTuple(sfd, head+i, 1);
			}
		}
	}

	if (sfd) {
		fclose(sfd);
	}
	if (wfd) {
		fclose(wfd);
	}
	if (p2pfd) {
		fclose(p2pfd);
	}
	if (potfd) {
		fclose(potfd);
	}

	/* email syn dump file
	 * we must sort this because of the topn filtering.
	 * we sort by port 25 or other email port syn count.
	*/
	if (efd) {
		int emailcount = email_topn_count;
		qsort((void *)head, c, sizeof(struct synnode), sort_by_25syncount);
		for (i = 0; i < l->entries; i++) {
			/* email synner dump
			*/
			/* note if ip belongs to the home subnet or not
                         * we always put home ips in the list, topn or not.
			 */
			usflag = ishome(head[i].ipsrc);
			if (emailcount > 0 && (head[i].p25_syn > 0)) {
				emailcount--;
				printSynTuple(efd, head+i, 1);
			}
			else if (usflag && (head[i].p25_syn > 0)) {
				printSynTuple(efd, head+i, 1);
			}
		}
	}

	if (efd) {
		fclose(efd);
	}


#ifdef SYNDUMP
	/* this is a debug feature and used to dump the entire synlist which may not be optimal
	 * it is usually too big due to p2p'ing.
	*/
	sfd = fopen("/tmp/syndump.txt" , "w");
	if (sfd == NULL) {
		syslog(LOG_ERR, "could not open %s:(%m)", topn_syn.syndump);  
		exit(1);
	}
	/* write all the syn info to a file -- all entries
	*/
	for (i = 0; i < l->entries; i++) {
		printSynTuple(sfd, head+i, 1);
	}
	fclose(sfd);
#endif

#ifdef EMULESTATS
	/* this is just for gathering of stats for edonkey and has no runtime use
	*/
	{
		FILE *sfd;
		sfd = fopen("/tmp/emulestats.txt", "w");
		if (sfd) {
			fprintf(sfd, "C5C5c5 codes\n");
			for (i = 0; i < 256; i++) {
				fprintf(sfd, "(%x: %x) ", i, edc5[i] & 0xff);
				if (i % 16 == 0)
					fprintf(sfd, "\n");
			}
			fprintf(sfd, "\nE3E3e3 codes\n");
			for (i = 0; i < 256; i++) {
				fprintf(sfd, "(%x: %x) ", i, ede3[i] & 0xff);
				if (i % 16 == 0)
					fprintf(sfd, "\n");
			}
			fprintf(sfd, "\nD4D4d4 codes\n");
			for (i = 0; i < 256; i++) {
				fprintf(sfd, "(%x: %x) ", i, edd4[i] & 0xff);
				if (i % 16 == 0)
					fprintf(sfd, "\n");
			}
			fclose(sfd);
		}
	}
#endif
	
	fprintf(fd, "\n");

	/* print out tworm count for RRD analysis
	*/
	if (ipaflag) {
		fprintf(fd, "tworm: %d: %d: %d:\n", wcount, us, them);
	}

	/* print out p2p stats for RRD analysis
	*/
	fprintf(fd, "tcpp2p: %u: %u: %u: %u: %u: %u: %u: %u: %u: %u: %u: %u:\n", 
		totalsizes[0], totalsizes[1], totalsizes[2], totalsizes[3],
		totalp2p[0], totalp2p[1], totalp2p[2], totalp2p[3],
		websizes[0]+websizes[1], websizes[2]+websizes[3], 
		websizes[4], websizes[5]);

	free(head);

        /* check for trigger condition
	 * we trigger if the total count is greater than the threshold.
        */
	trigger_condition_check(TWORM, fd, wcount, "tworm");
}          

/*
 * is ip address in set of home ip addresses or no
 */
int
ishome(unsigned int ipsrc)
{
	int usflag = 0;
	int k;
	unsigned int usnet, usmask;

	for ( k = 0; k < topn_syn.homeips; k++) {
		usnet = topn_syn.ipaddr[k].s_addr;
		usmask = topn_syn.netmask[k].s_addr;
		if ((ipsrc & usmask) == usnet) {
			usflag = 1;
		}
	}
	return(usflag);
}


#ifdef TESTMAIN

this code is old and broken

static int f = 0;

unsigned int
getRandomWord()
{
        static unsigned int rw;

#define REALLYRANDOM
#ifdef REALLYRANDOM
        if (f == 0) {
                f = open("/dev/urandom", 0);
                if (f < 0) {
                        perror("open");
                        exit(1);
                }
        }
        read(f, &rw, sizeof(unsigned int));
#else
        rw = rand();
#endif
        return(rw);
}


functionTest()
{
	struct synlist l;
	unsigned int ipsrc;
	int rc;
	struct synnode *n = NULL;

	/* load sample ipkey and counter
	*/
	ipsrc = inet_addr("1.1.1.1");

	init_synlist(&l);

	n = search_srcsyn(NULL, &l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("search failed\n");
	} 
	
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	n = search_srcsyn(NULL, &l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 

	/* add 4 more nodes and dump the list
	*/
	ipsrc = inet_addr("1.1.1.2");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.3");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}
	ipsrc = inet_addr("1.1.1.4");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.5");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.6");
	rc = insert_srcsyn(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	/* now search for a given node, and increment its
	 * value
	 */
	n = search_srcsyn(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	/* now search for a previous node, note last
	 * as a simple cache should not be used here
	 */
	ipsrc = inet_addr("1.1.1.4");
	n = search_srcsyn(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	dump_hash(&l);

	print_synlist(stdout, &l, 10);

	destroy_synlist(&l);
}


/*
 * borrowed from trafshow.
 *
 * A faster replacement for inet_ntoa().
 */

char *
intoa(addr)
unsigned int addr;
{
	register char *cp;
	register unsigned int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	addr = ntohl(addr);
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return cp + 1;
}


main()
{
	functionTest();
}
#endif

