/*
 *
 * ipanalyze.c
 *
 * middle-level function that extracts most of the useful
 * bits from an IP packet, and shoves them in a structure
 * for use by cascading filters later.
 *
 */

/*
 * Copyright (c) 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>                                             
#include <stdlib.h>  
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <pcap.h>
#ifdef	TIME_WITH_SYS_TIME
#include <time.h>
#endif

#include "ourmon.h"
#include  "filter.h"
#include  "hashblist.h"
#include "spinlock.h"                                          
#include "sync.h"        

void dumpHexPayload(char *p, int psize);

/*
 * based on minimal upper-header info for
 * ip packets, we call filters
 *
 * basic flow:
 *	parse for various errors
 *	if 1st packet of ip datagram
 *		tcp
 *		udp
 *		icmp
 *		other
 *
 *	ip_filter(on various stats) and return
 *
 *	error_goto -- assume "non" ip and/or outrageous errors
 *
 * Side-effect:
 *	we set te->ipflag to 1 iff the te structure has been filled out ok in
 *		terms of data later possibly later used by the CBPF code.
 */

int nol2flag;

void
analyze_ip(h, p, ip, length, caplength, eh, te)
	struct pcap_pkthdr *h;
	unsigned char *p;
	register struct ip *ip; /* pointer to ip portion */
	int length;		/* length - of ip portion minus link */
	int caplength;		/* actual length of packet including link */
	struct ether_header *eh;
	struct t_entry *te;
{
	register iplen;
	int hlen;
	register u_char *cp;
	int tmp;
	int ipcalclen;		/* calculated ip pkt l7 payload length */
	extern int snaplen;

	/* this flag is used to decide whether or not to run the fixed_cast
	 * function later
	*/
	if (eh) {
		nol2flag = 0;
	}
	else {
		nol2flag = 1;
	}

	/* if no ethernet header process for all packet types including errors
	*/
	
#ifdef THREAD
        spinlock(&locks[LOCK_FIXED_IPPROTO]);                  
#endif
	/* six bins with more small bins for UDP.
	 * fixed_size histogram has not changed however.
	 * bin sizes: 77,110,250,500,1000, bigger
	 * note: historical 4 bin mechanism is now:
	 * 128/500/1000/big  (small change in tiny bin)
	 * NOTE: caplength includes IP header, L4 header.
	 */
	if ((!eh)) {
		if (caplength <= 77) {
			fixed_size.tiny++;
			te->sizeindex = 0;
		}
		else if (caplength <= 110) {
			fixed_size.tiny++;
			te->sizeindex = 1;
		}
		else if (caplength <= 250) {
			fixed_size.small++;
			te->sizeindex = 2;
		}
		else if (caplength <= 500) {
			fixed_size.small++;
			te->sizeindex = 3;
		}
		else if (caplength <= 1000) {
			fixed_size.medium++;
			te->sizeindex = 4;
		}
		else {
			fixed_size.large++;
			te->sizeindex = 5;
		}
	}
#ifdef THREAD
        unlock(&locks[LOCK_FIXED_IPPROTO]);                    
#endif

	/* if longer than snap (never mind the goto),
	 * ignore it
	 * TBD. count errors.
	 */
	if ((u_char *)(ip + 1) > snapend)  {
		goto too_short;
	}

	if (length < sizeof(*ip)) {
		goto too_short;
	}

	/* length of ip + rest
	*/
	iplen = ntohs(ip->ip_len);

	if (iplen < 1)  {
		goto too_short;
	}

	/* actual length compared to value in ip hdr
	*/
	if (length < iplen) {
		goto too_short;
	}

	/*
	 * If this is fragment zero, hand it to the next higher level protocol
	 * to fetch ports or icmp type else it is fragmented ip datagram.
	 */
	te->psize = te->sport = te->dport = 0;
	if ((ntohs(ip->ip_off) & 0x1fff) == 0) {
		/* advance to high level protocol header */
		hlen = ip->ip_hl << 2;
		cp = (u_char *)ip + hlen;
		if (ip->ip_p == IPPROTO_TCP) {
			/* tcp header too short
			*/
			if (cp + sizeof(struct tcphdr) > snapend ||
			    iplen - hlen < sizeof(struct tcphdr)) {
				goto too_short;
			}
#ifdef LINUX
#define th_sport source
#define th_dport dest
#endif
			te->sport = ntohs(((struct tcphdr *)cp)->th_sport);
			te->dport = ntohs(((struct tcphdr *)cp)->th_dport);
			te->synflag = te->finflag = te->resetflag = te->ackflag = 0;
#ifdef LINUX
                        if (((struct tcphdr *)cp)->ack) {
				te->ackflag = 1;
			}

                        if (((struct tcphdr *)cp)->syn) {
				te->synflag = 1;
			}
                        else if (((struct tcphdr *)cp)->fin) {
				te->finflag = 1;
			}
                        else if (((struct tcphdr *)cp)->rst) {
				te->resetflag = 1;
			}
#else
                        if ((((struct tcphdr *)cp)->th_flags &
                                (TH_SYN))) {
				te->synflag = 1;
                        }
                        else if ((((struct tcphdr *)cp)->th_flags &
                                (TH_FIN))) {
				te->finflag = 1;
                        }
                        else if ((((struct tcphdr *)cp)->th_flags &
                                (TH_RST))) {
				te->resetflag = 1;
                        }
                        if ((((struct tcphdr *)cp)->th_flags &
                                (TH_ACK))) {
				te->ackflag = 1;
			}
#endif
			/* determine protocol payload size if any
			*/
#ifdef LINUX
                        te->p = cp + (((struct tcphdr *)cp)->doff * 4);
			/* acc. to IP/TCP hdrs, payload is this long
			 * BUT remember that actually have a snaplen
			 */
			ipcalclen = iplen-hlen-(((struct tcphdr *)cp)->doff * 4);
#else
			/* BSD/solaris */
                        te->p = cp + (((struct tcphdr *)cp)->th_off << 2);
			ipcalclen=iplen-hlen-(((struct tcphdr *)cp)->th_off<<2);

#endif
                        te->psize = snapend - te->p;  /* length of L7 portion if any */
			/* with smallest ip pkts, we may have nulls at the end
			 * and we need to ignore them
			 */
			if (te->psize > ipcalclen) {
				te->psize = ipcalclen;
			}
#ifdef DEBUG
			/* goal here is to make sure we have the right payload length.
			 * unfortunately there are 2 test environments, batch/interface
			 * on every system.
			 */
                        if (debug) {
#ifdef LINUX
				printf("TCPPKT: length/caplength %d %d\n", length,
					caplength);
				tmp=iplen-hlen-
					(((struct tcphdr *)cp)->doff<<2);
				printf("TCPPKT: iplen %d hlen %d doff %d, total: %d\n", iplen, hlen, ((struct tcphdr *)cp)->doff, tmp);
#else
				printf("TCPPKT: length/caplength %d %d\n", length,
					caplength);
				tmp=iplen-hlen-
					(((struct tcphdr *)cp)->th_off<<2);
				printf("TCPPKT: iplen %d hlen %d th_off %d, total: %d\n", iplen, hlen, ((struct tcphdr *)cp)->th_off, tmp);
#endif
				printf("TCPPKT: snapend %x te->p %x, psize %d\n", 
					snapend, te->p, te->psize);
				dumpHexPayload((char *)te->p, te->psize);
				/*
				printf("\n");
				*/
                        }
#endif

		} else if (ip->ip_p == IPPROTO_UDP) {
			/* udp header too short
			*/
			if (cp + sizeof(struct udphdr) > snapend ||
			    iplen - hlen < sizeof(struct udphdr)) {
				goto too_short;
			}
#ifdef LINUX
#define uh_sport source
#define uh_dport dest
#define uh_ulen len
#endif
			te->sport = ntohs(((struct udphdr *)cp)->uh_sport);
			te->dport = ntohs(((struct udphdr *)cp)->uh_dport);

    			/* save payload size and set pointer to payload
                        */
                        te->p = cp + sizeof(struct udphdr);
                        te->psize = snapend - te->p;        /* length of L7 portion if any */
			ipcalclen = ntohs(((struct udphdr *)cp)->uh_ulen)-
				sizeof(struct udphdr);
			if (te->psize > ipcalclen) {
				te->psize = ipcalclen;
			}
		} else if (ip->ip_p == IPPROTO_ICMP) {
			/* icmp header too short 
			 * the only part of the icmp header we care about are the
			 * major/minor codes, which appear in the 1st two bytes
			*/
			if (cp + 2 > snapend || ((iplen - hlen) < 2)) {
				goto too_short;
			}
			te->sport = ((struct icmp *)cp)->icmp_type;
			te->dport = ((struct icmp *)cp)->icmp_code;

			/* stay out of ICMP payload if any for now. 	
			 * this sizeof for struct icmp is wrong anyway.
                        te->p = cp + sizeof(struct icmp);
                        te->psize = snapend - te->p;        
                        */
		}
	}

	te->src.s_addr = ip->ip_src.s_addr;
	te->dst.s_addr = ip->ip_dst.s_addr;
	te->proto = ip->ip_p;
	te->bytes = iplen;
	te->flags = te->honeyflag = te->darkflag = 0;
	if (fixed_honeynet.configured) {
		struct in_addr dst;

		dst.s_addr = te->dst.s_addr & fixed_honeynet.netmask.s_addr;
		if ( dst.s_addr == fixed_honeynet.netaddr.s_addr ) {
			te->honeyflag = 1;
		}

		if (debug && te->honeyflag) {
			printf("HONEY: %d: dst addr %s, dst port %u:", te->honeyflag,
				inet_ntoa(te->dst), ntohs(te->dport));
			printf("HONEY: honey addr %s\n", 
				inet_ntoa((struct in_addr)fixed_honeynet.netaddr));
		}
	}
	if (fixed_darknet.configured) {
		struct in_addr dst;

		dst.s_addr = te->dst.s_addr & fixed_darknet.netmask.s_addr;
		if ( dst.s_addr == fixed_darknet.netaddr.s_addr ) {
			te->darkflag = 1;
		}
	}

	/* only one flag can be set.  darkflag wins if collision
 	 * which means the 2 subnets overlap
	 */
	if (te->darkflag && te->honeyflag) {
		te->honeyflag = 0;
	}

	/* initialize structures for application tags */

	te->attempted.upper=0;
	te->attempted.lower=0;
	te->appresult.upper=0;
	te->appresult.lower=0;

	/* sanity check */

	if (te->psize < 0)  
		te->psize = 0;
	if (te->psize > snaplen)  
		te->psize = snaplen;

        /* TBD. fix this.  this is too high granularity       
        */                                                     
#ifdef THREAD
        spinlock(&locks[LOCK_ELGRANDE]);                  
#endif
        ip_filter(caplength, te);                        
#ifdef THREAD
        unlock(&locks[LOCK_ELGRANDE]);                    
#endif

	te->ipflag = 1;
	if (debug) {
		printf("analyze_ip te is set %x\n", te);
	}

	/* check blacklist if on - only for dumpfile storage
	*/
	if (blist_config) {
		search_blist(h, p, te);
	}
	return;

too_short:
	/* just count it
	*/
	non_ip_filter(caplength);

	return;
}

void
dumpHexPayload(char *p, int psize)  
{
	int cr = 0;
	while(psize--) {
		if (cr++ == 16) {
			cr = 0;
			printf("\n");
		}
		printf("%x:", *p++ & 0xff); 
	}
	printf("\n");
}



