/*
 * ircscan.c - build up irc info
 * 
 * determine minimal useful snaplength. 256 so far.
 *
 * design note:
 *
 * essentially we are doing 3 things here:
 *
 * 1. extracting a list of channels.  in irc channels are
 * case-insensitive, but this code ignores that, and in point
 * of fact is not 100% correct in terms of its parsing.  This
 * is because we favor stats and speed over perfect accuracy.
 * E.g., we could have two channels both named "foo" with
 * entirely different servers, but we ignore that.  If they
 * both lower-case they are treated as one channel.
 * The goal is to extract the channels, and a set of related ips.
 *
 * 2. we extract a set of IPs doing irc.  we note if an IP
 * appears to be an IRC server based on some very loose notions
 * in the IRC messages (JOIN and PRIVMSG only).
 *
 * 3. we compute raw stats for message counts.  
 *
 * at report time, we ask the syn list for a work weight for any
 * given IP.
 *
 * TBD: it might not hurt at some future data to do the following:
 *
 * 1. sample the destination ports.  At this point we learn nothing
 * about which ports are in use although we could.
 * 2. determine L3 destination counts. (cross module call to scanner module)
 */

/*
 * Copyright (c) 2004, 2005 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif
#include <pcap.h>

#include "ourmon.h"	/* struct t_entry */
/*
#include "ircscan.h"
*/
#include "filter.h"

#include "ehash.c"

static void isjoin(struct irclist *, int, struct in_addr, struct in_addr, 
	unsigned short, unsigned short, char *, int, unsigned char *, int );
static void isping(struct irclist *, struct in_addr, struct in_addr, 
	unsigned short, unsigned short, unsigned char *, int);
static void ispong(struct irclist *, struct in_addr, struct in_addr, 
unsigned short, unsigned short, unsigned char *, int, int);
static void isprivmsg(struct irclist *, int, struct in_addr, struct in_addr, unsigned short, unsigned short, char *, int, unsigned char *, int);

extern int searchSyn(struct synlist *, unsigned int);

static struct ircnode *
hash_ircnode(struct irclist *l, struct in_addr ip, int type, int sflag, unsigned short dport, unsigned short sport, int);

static void 
hash_channel(struct irclist *, struct ircnode *, struct ircnode *, unsigned short, char *, int, int); 

static struct ircnode * search_chan_list(struct channel *, struct ircnode *);
static void insert_chan_list(struct channel *, struct ircnode *);
static void free_channel(struct channel *);

#define BLANK 0x20

extern struct topn_irc topn_irc;

/* called before list can be used
*/
void
initIrc(struct irclist *l)
{
	bzero(l, sizeof(struct irclist));
}

destroyIrc(struct irclist *l)
{
	int i;
	struct ircnode *n;
	struct ircnode *prev;
	struct channel *cn;
	struct channel *cprev;

	/* destroy ircnode list
	*/
	for (i =0 ; i<IRCHASH; i++) {
		n = l->iphash[i];
                while(n != NULL) {
                        prev = n;
                        n = n->next;
                        free(prev);
                }
	}

	/* free channel list and associated per-channel space
	*/
	for (i =0 ; i<IRCHASH; i++) {
		cn = l->chanhash[i];
                while(cn != NULL) {
                        cprev = cn;
                        cn = cn->nextchan;
			free_channel(cprev);
                        free(cprev);
                }
	}
}

/*
 * ircscan: parsing notes
 * overview:
 *	scan L7 payload for JOIN/PING/PONG/PRIVMSG irc messages
 *	goals include:
 *		figure out server vs host
 *		find channel string
 *		calculate IPs in channels
 *		calc some overall stats
 *
 *	Interested in channel names with JOINs/PRIVMSGs, and ip src
 *	relationships with JOIN/PING/PONG.
 *
 * privacy note:  no keys exposed here, and no internal message
 *	guts exposed either.
 *	
 * assumptions:
 *	only called for TCP packets.
 *	only called if there is any L7 payload at all.  not for TCP syns  
 *		without payload.
 *
 * note: CRNL (0xa0xd)
 *
 ****************************************
 * JOIN parsing: 
 *
 * initial:
 * A JOIN may begin the payload, or basically have a string followed by any number
 * of blanks.  This is actually a client->server message or server->client message.
 *
 * post JOIN:
 * any number of blanks, #, &, or :#, followed by channel string.
 * Terminated by blank or CRNL.  CR will do.
 *
 * must extract two things:
 * 1. # or : which indicates client->server or server->client.  E.g.,
 *	#bark indicates client server for channel bark.
 *	:bark indicates server client for channel bark.
 * 2. the channel string itself, sans initial indicator.
 *
 * rough patterns:
 *
 * 1. client to server message
 * ^JOIN[BLANK]#<channel>[BLANK][keystring][BLANK,CRNL]  
 * 2. server to client message
 * :<idstring>[BLANK]JOIN[BLANK]:#<channel>[BLANK,CRNL] 
 *
 * We ignore the keystring at this time.
 * 
 ****************************************
 * PRIVMSG parsing:
 *
 * notes:
 * 1. PRIVMSG like JOIN barring the lack of ":#channel" for server->client.
 * That is, #channel is used, not ":#channel".  
 *
 * 2. PRIVMSG #channel :msg ...   is basic structure of end part of PRIVMSG.
 * We do not care about the msg.
 *
 * once again client/server, server/client cases:
 *
 * rough patterns:
 *
 * 1. client to server message
 * ^PRIVMSG[BLANK]#<channelstring>[BLANK][:msg][BLANK,CRNL]  
 * 2. server to client message
 * :<idstring>[BLANK]PRIVMSG[BLANK]#<channelstring>[BLANK][:msg][BLANK,CRNL]  
 *
 * as with JOIN the existance of the :idstring[BLANK] tells us if we have client/server
 * or server/client.  
 *
 ****************************************
 * PING and PONG parsing (the same):
 *
 * 1. initial may start with
 * ^PING
 * 2. initial may start with 
 * <string>[BLANK]*PING
 *
 * remainder:
 * PING[BLANK]<string>
 * PING[CRNL]
 *
 * TBD: although we do not yet take advantage of this the PONG message
 * can be used for determining server->client messages as they have the
 * form:
 * 
 * :<identity>[BLANK]PONG
 *
 * in a fashion similar to JOIN/PRIVMSG.
 *
 * observations:
 *
 * 1. JOIN and PRIVMSG have same semantic for client to server, server to client 
 * ^JOIN is client to server
 * :id JOIN :#channel  is server to client
 *
 * note however that JOIN in the latter case has a :# semantic.  The channel name
 * does not use the '#'.  PRIVMSG does not use ':'.   Use of ':' with JOIN is currently used.
 * (possibly for no good reason).
 *
 * 2. PONG can be used for server ID and this should be taken advantage of.
 * :identity PONG identity :clientid  --- sent by server (ip src is server).  this is true.
 *
 * In point of fact it looks like PING/PONG may often have the same client/server nature; 
 * ^PING something - from client to server
 * :id PONG :id  - from server to client
 *
 * However this is not true for all PING/PONG messages in the wild.
 *
 *	returns:
 *		1 - if join
 *		2 - if ping
 *		3 - if pong
 *		4 - if privmsg
 *		0 - if not irc
 */
int
ircscan(struct irclist *l, struct t_entry *te) 
{
	extern int debug;
	char channame[512];				/* max size of channel name */
	int chansize; 
	char *cpt;
	unsigned int ipsrc, ipdst;
	unsigned short dport;
	unsigned char *p;
	int psize;
	int sflag = 0;					/* ip src is server */

/*
	if (debug) {
		printf("IRCSCAN called \n");
	}
*/
	p     = te->p;
	psize = te->psize;

	/* check for possible match at start of
	 * messge, if so go to state 2 if NONE
	*/
	if (*p == 'J')
		goto state2;
	if (*p == 'P')
		goto state2;

	/* cycle thru NONBLANKS if any
	*/
	while(psize) {
		/* break on blanks
		*/
		if (*p == BLANK)
			goto blanks1;
		p++;
		psize--;
	}
	return(0);

	/* cycle thru BLANKS if any
	*/
blanks1:
	while(psize && (*p == BLANK)) {
		p++;
		psize--;
	}
	if (psize && ((*p == 'J') || (*p == 'P')))
		goto state2;
	/* no match 
	*/
	return(0);

state2:
	/* too small, bail
	*/
	if (psize < 6) {
		return(0);
	}

	/* JOIN ? */
	if ((*p == 'J') && (*(p+1) == 'O') && (*(p+2) == 'I') && (*(p+3) == 'N')) {
		p += 4;
		psize = psize - 4;
		/* too small, bail
		*/
		if (psize == 0)
			return(0);

		/* at this point must have blank
		*/
		if (*p != BLANK)
			return(0);

		/* move to word following blanks
		*/
		while (psize && (*p == BLANK)) {
			psize--;
			p++;
		}
		if (psize == 0) {
			return(0);
		}
		/* now at channel name, save it in local buffer
		if ((*p == '#') || (*p == '&')) {
		*/
		if ((*p == '#') || (*p == '&') || *p == ':') {
			chansize = 0;
			cpt = &channame[0];
			/* skip over any initial separation chars
			*/
			if ((*p == '#') ||  (*p == '&')) {
				psize--; p++;
			}
			else if (*p == ':') {
				/* protocol note:
				 *     JOIN # - indicates client to server
				 *     JOIN : - indicates server to client
				*/
				sflag = 1;		/* ip src is server */
				psize--; p++;
				/* TBD sanity check this
				*/
				if (*p == '#')
					psize--; p++;
			}
			/* rfc 1459
			<chstring> ::=
			    <any 8bit code except SPACE, BELL, NUL, CR, LF and comma (',')>
			*/
			while (psize && (*p != BLANK) && (*p != '\n') && (*p != '\r')) {
				*cpt++ = *p++;
				chansize++;
				if (chansize > 510)
					return(0);
				psize--;
			}
			if (psize == 0) {
				return(0);
			}
			channame[chansize] = 0;
			isjoin(l, sflag, 
				te->src, te->dst, 
				te->dport, te->sport,
				channame, chansize, te->p, te->psize);
			return(ISJOIN);
		}
		return(0);
	}

	/* PING */
	else 
	if ((*p == 'P') && (*(p+1) == 'I') && (*(p+2) == 'N') && (*(p+3) == 'G')) {
		/* at this point must have blank or CRNL.  PING has at least one word
	         * behind it in all cases, although we are not absorbing it
		*/
		p += 4;
		psize = psize - 4;
		if (psize == 0)
			return;
		if (*p == BLANK) {
			isping(l, te->src, te->dst, 
				te->dport, te->sport, te->p, te->psize);
			return(ISPING);
		}
		if (psize == 2 && ((*p == '\r') && (*p == '\n'))) {
			isping(l, te->src, te->dst, 
				te->dport, te->sport, te->p, te->psize);
			return(ISPING);
		}
		return(0);
	}

	/* PONG */
	else 
	if ((*p == 'P') && (*(p+1) == 'O') && (*(p+2) == 'N') && (*(p+3) == 'G')) {
		/* determine if pong is server to client
		*/
		if (p != te->p) {
			sflag = 1;
		}
		/* at this point must have blank or CRNL
		 * PONG has at least one word behind it in all cases.
		*/
		p += 4;
		psize = psize - 4;
		if (psize == 0)
			return(0);
		if (*p == BLANK) {
			/* TBD: not using sflag yet.  check debug info again
			*/
			ispong(l, te->src, te->dst, 
			te->dport, te->sport, te->p, te->psize, 0);
			return(ISPONG);
		}
		if ((psize == 2) && ((*p == '\r') && (*p == '\n'))) {
			ispong(l, te->src, te->dst, 
			te->dport, te->sport, te->p, te->psize, 0);
			return(ISPONG);
		}
		return(0);
	}

	/* PRIVMSG - similar to join code above :blah PRIVMSG channel :blah*/
	else
	if ((*p == 'P') && (*(p+1) == 'R') && (*(p+2) == 'I') && (*(p+3) == 'V') && (*(p+4) == 'M') && (*(p+5) == 'S') && (*(p+6) == 'G')) {
		/* check to see if server is sending PRIVMSG
		 * idea: client PRIVMSGs do not have an identity string in front of them. 
		*/
		if (p != te->p) {
			sflag = 1;
		}
		p += 7;
		psize = psize - 7;
		if (psize == 0)
			return(0);
		
		if (*p != BLANK) {
			return(0);
		}

		/* move to word following blanks - this is channel
		*/

                while (psize && (*p == BLANK)) {
                        psize--;
                        p++;
                }
                if (psize == 0) {
                        return(0);
                }

               /* now at channel name, save it in local buffer
                if ((*p == '#') || (*p == '&')) {
                */
                if ((*p == '#') || (*p == '&') || *p == ':') {
                        chansize = 0;
                        cpt = &channame[0];
                        /* skip over any initial separation chars
                        */
                        if ((*p == '#') ||  (*p == '&')) {
                                psize--; p++;
                        }
                        else if (*p == ':') {
                                psize--; p++;
                                if (*p == '#')
                                        psize--; p++;
                        }
			/* rfc 1459
			<chstring> ::=
			    <any 8bit code except SPACE, BELL, NUL, CR, LF and comma (',')>
			*/
			while (psize && (*p != BLANK) && (*p != '\n') && (*p != '\r')) {
                                *cpt++ = *p++;
                                chansize++;
                                if (chansize > 510)
                                        return(0);
                                psize--;
                        }
                        if (psize == 0) {
                                return(0);
                        }
                        channame[chansize] = 0;
			isprivmsg(l, sflag, te->src, te->dst, 
				te->dport, te->sport, 
				channame, chansize, te->p, te->psize);
			return(ISPRIVMSG);
		}
		return (0);
	}
	return(0);	
}


static void
dumpPayload(unsigned char *p, int psize)
{
	int i;

	for ( i = 0; i < psize; i++) {
		if (*p == '\r')
			return;
		putchar(*p++);
	}
}

/*
 * process join message information.
 *
 * sflag: if true, ip src is irc server
 */

static void
isjoin(struct irclist *l, int sflag, 
	struct in_addr ipsrc, struct in_addr ipdst, 
	unsigned short dport,
	unsigned short sport,
	char *chan, int cs, unsigned char *p, int psize)
{
	struct channel *nchan;
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;

	if (debug) {
		char buf[1024];
		sprintf(buf, "%s", inet_ntoa(ipsrc));
		printf("\tIRCMSG: JOIN: s=%s d=%s dport=%d sflag=%d chan=%s clen=%d: p=[", buf, inet_ntoa(ipdst), dport, sflag, chan, cs);
		dumpPayload(p, psize);
		printf("]\n");
	}

	/* update total joins
	*/
	l->totaljoins++;

	/* store both ip addresses and do stats, returning
	 * ircnode pointers.
	*/

	src = hash_ircnode(l, ipsrc, ISSRC, sflag, dport, sport, ISJOIN);
	dst = hash_ircnode(l, ipdst, ISDST, !sflag, sport, dport, ISJOIN);

	hash_channel(l, src, dst, dport, chan, cs, ISJOIN);
}

static void
isping(struct irclist *l, struct in_addr ipsrc, struct in_addr ipdst, unsigned short dport, unsigned short sport,
	unsigned char *p, int psize)
{
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;

	if (debug) {
		char buf[1024];
		sprintf(buf, "%s", inet_ntoa(ipsrc));
		printf("\tIRCMSG: PING: s=%s d=%s dp=%d: p=[", buf, inet_ntoa(ipdst), dport);
		dumpPayload(p, psize);
		printf("]\n");
	}
	l->totalpings++;

	src = hash_ircnode(l, ipsrc, ISSRC, 0, dport, sport, ISPING);
	dst = hash_ircnode(l, ipdst, ISDST, 0, sport, dport, ISPING);
}

static void
ispong(struct irclist *l, struct in_addr ipsrc, struct in_addr ipdst, unsigned short dport, unsigned short sport,
	unsigned char *p, int psize, int sflag)
{
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;

	if (debug) {
		char buf[1024];
		sprintf(buf, "%s", inet_ntoa(ipsrc));
		printf("\tIRCMSG: PONG: s=%s d=%s dp=%d: sflag=%d: p=[", buf, inet_ntoa(ipdst), dport, sflag);
		dumpPayload(p, psize);
		printf("]\n");
	}

	l->totalpongs++;
	src = hash_ircnode(l, ipsrc, ISSRC, sflag, dport, sport, ISPONG);
	dst = hash_ircnode(l, ipdst, ISDST, 0, sport, dport, ISPONG);
}

static void
isprivmsg(struct irclist *l, int sflag, struct in_addr ipsrc, struct in_addr ipdst, unsigned short dport, unsigned short sport,
	char *channel, int clen, unsigned char *p, int psize)
{
	extern int debug;
	struct ircnode *src;
	struct ircnode *dst;

	if (debug) {
		char buf[1024];
		sprintf(buf, "%s", inet_ntoa(ipsrc));
		printf("\tIRCMSG: PRIVMSG: s=%s d=%s dp=%d sflag=%d, chan=%s clen=%d: p=[", buf, inet_ntoa(ipdst), dport, sflag, channel, clen);
		dumpPayload(p, psize);
		printf("]\n");
	}

	l->totalprivmsgs++;
	src = hash_ircnode(l, ipsrc, ISSRC, sflag, dport, sport, ISPRIVMSG);
	dst = hash_ircnode(l, ipdst, ISDST, !sflag, sport, dport, ISPRIVMSG);

	hash_channel(l, src, dst, dport, channel, clen, ISPRIVMSG);
}

/*
 * hash an ircnode, lookup, create if necessary and link to hash list.
 * store stats.
 *
 * TBD: not doing anything with dport at this time.
 * sflag: if ISJOIN or ISPRIVMSG, then sflag == 1 indicates server, and 0 indicates client.
 */
static
struct ircnode *
hash_ircnode(struct irclist *l, 
	struct in_addr ip, int type, int sflag, 
	unsigned short dport, unsigned short sport, int command)
{
	struct ircnode *n, *hb;
	struct ircnode *fb;	/* first bucket */
	int mod;
	unsigned int ipaddr = ip.s_addr;
	int foundit = 0;

	/* look it up
	*/
        fb = hb = l->iphash[mod=(ehash((unsigned char *)&ipaddr, 4) % IRCHASH)];
        for (; hb!= NULL ; hb=hb->next) {
                if (hb && ( hb->ipsrc == ipaddr)) {
			foundit = 1;
			n = hb;
                        break;
                }
        }

	/* invariants.  1. hb is set to hash bucket slot.
	 *	        2. n is null or ptr to desired node
	 */

	/* if we didn't find it, create it, hook it up
	*/
	if (!foundit) {
		n = (struct ircnode *) malloc(sizeof(struct ircnode));
		if (n == NULL) {
			syslog(LOG_ERR, "malloc failure in hash_ircnode, entry count: %d",
				l->ipcount);
			return;
		}
		bzero(n, sizeof (struct ircnode));
		/* do irclist stats */
		l->ipcount += 1;
		/* insert the synnode in the hash list
		*/
		if (fb == NULL) {
			l->iphash[mod] = n;
		}
		else {
			n->next = fb;
			l->iphash[mod] = n;
		}
		n->ipsrc = ipaddr;
		n->dport = dport;
		n->sport = sport;
	}

	/* n is now pointer to desired node
	*/
	/* do stats
	*/
	if (command == ISJOIN) {
		if (sflag)
			n->serverhits++;
		if (type == ISSRC) {
			n->joinsrc++;
		}
		else {
			n->joindst++;
		}
	}
	else if (command == ISPING) {
		if (type == ISSRC) {
			n->pingsrc++;
		}
		else {
			n->pingdst++;
		}
	}
	else if (command == ISPONG) {
		if (type == ISSRC) {
			n->pongsrc++;
		}
		else {
			n->pongdst++;
		}
	}
	else if (command == ISPRIVMSG) {
		if (sflag)
			n->serverhits++;
		if (type == ISSRC) {
			n->privsrc++;
		}
		else {
			n->privdst++;
		}
	}
	return(n);
}


/*
 * search the channel link chain for a ircnode.
 * if found return a pointer to the node, else
 * return NULL
*/

static
struct ircnode *
search_chan_list(struct channel *c, struct ircnode *n)
{
	struct irclink *link = c->nextlink;

	if (link == NULL)
		return(NULL);

        for (; link != NULL ; link=link->nextlink) {
		/* match here on ipsrc addres
		*/
		if (link && (link->ircnode->ipsrc == n->ipsrc)) {
			/* found node */
			return(n);
		}
	}
	return(NULL);
}
 
/*
 * given a channel c and a ircnode n,
 * put the irc node at the beginning of the irc link chain
 * we must allocate a link node.
 *
 * note: assumptions here include
 *	we only put ircnode in channel ircnode link list
 *	if it is unique.  
 */
static
void 
insert_chan_list(struct channel *c, struct ircnode *n)
{
	struct irclink *link, *prev;

	link = (struct irclink *) malloc(sizeof(struct irclink));
	if (link == NULL) {
		syslog(LOG_ERR, "malloc failure in insert_chan_list\n");
		return;
	}
	link->nextlink = NULL;

	/* put the link node at the beginning of the list
	*/
	/* if the list is empty put it at the front
	*/
	if (c->nextlink == NULL) {
		c->nextlink = link;	
	}
	/* else put it at the front!  chaining it in.
	*/
	else {
		prev = c->nextlink;
		c->nextlink = link;
		link->nextlink = prev;
	}

	/* make the link point to the ircnode
	*/
	link->ircnode = n;
	n->channels++;					/* ircnode unique channel count */

	/* count the number of mallocs
	*/
	c->noips++;
}

void
dump_channels(struct irclist *l)
{
	int i;
	struct channel *n;

	/* walk the hashlist
	*/
	for (i =0 ; i<IRCHASH; i++) {
		n = l->chanhash[i];
		/* if n is non-nulln use, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			printf("IRCCHAN %s %d %d %d %d\n",
				n->name, n->noips, n->hits, n->joins, n->privmsgs);	
			n = n->nextchan;
		}
	}  
} 

/* free any malloc'ed space associated with a channel
*/
static
void
free_channel(struct channel *c) 
{
	struct irclink *link = c->nextlink;
	struct irclink *prev;

	/* first free the channel name 
	*/
	free(c->name);

	/* now free any links associated with the channel
	 * but not the ircnodes!
	*/
        while(link != NULL) {
        	prev = link;
                link = link->nextlink;
		if (prev)
                	free(prev);
        }
}


/*
 * we have a join or privmsg.  put channel name
 * into channel hashlist, do stats, and hookup src/dst nodes
 * into GACK linear list hanging off of channel.
 * TBD: dport ignored for now.
 *
 * 1. lookup in channel list
 * 2. if not found allocate it and put it into channel list.
 *	allocate space for channel name
 * 3. search on src, then dst, and make sure they are pointed to.
 */

static void
hash_channel(struct irclist *l, struct ircnode *src, struct ircnode *dst,
	unsigned short dport, char *channel, int chanlen, int command)
{
	struct channel *nchan, *hb, *fb;
	int mod;
	int foundit = 0;
	extern int debug;

	if (debug) {
		printf("IRCCHANHASH channel %s chanlen %d command %d\n",
			channel, chanlen, command);
	}

	/* look it up in channel hash list 
	 * we either find it or we do not find it.
	*/
        fb = hb = l->chanhash[mod=(ehash((unsigned char *)channel, chanlen) % IRCHASH)];
        for (nchan=NULL; hb != NULL ; hb=hb->nextchan) {
		/* note: match here only on channel name 
		*/
		if (hb && (strcmp(channel, hb->name) == 0)) {
			/* found node */
			foundit = 1;
			nchan = hb;
			break;
		}
	}

	/* if we did not find it, create a new channel node
	*/

	if (!foundit) {
		/* didn't find it, create it, stick it in, init it
		*/
		nchan = (struct channel *) malloc(sizeof(struct channel));
		if (nchan == NULL) {
			syslog(LOG_ERR, "malloc failure in hash_channel, channel count %d\n", l->chancount);
			return;
		}
                bzero(nchan, sizeof (struct channel));
		/* at this point, nchan is pointing to a channel node
		*/

                /* do irclist stats */
		l->chancount++;

                /* insert the synnode in the hash list
                */
                if (fb == NULL) {
                        l->chanhash[mod] = nchan;
                }
                else {
                        nchan->nextchan = fb;
                        l->chanhash[mod] = nchan;
                }
		/* save channel identity,  +1 to make sure null-terminated 
		*/
		nchan->name = (char *) malloc(chanlen+1);
		if (nchan->name == NULL) {
			syslog(LOG_ERR, "malloc failure in hash_channel, on name %s\n", channel);
			return;
		}
		bzero(nchan->name, chanlen+1);
		strncpy(nchan->name, channel, chanlen);
		if (debug) {
			printf("IRCCHAN, strncpy %s %s %d\n", nchan->name, channel, chanlen);
		}
	}

	/* global channel stats
	*/
	nchan->hits++;
	if (command == ISJOIN)
		nchan->joins++;
	else if (command == ISPRIVMSG)
		nchan->privmsgs++;

	/* check channel list for src node, if not present, link it in and count it
	*/
	if ( search_chan_list(nchan, src) == NULL ) {
		insert_chan_list(nchan, src);	/* put it in the channel list */
	}

	/* ditto for destination
	*/
	if ( search_chan_list(nchan, dst) == NULL ) {
		insert_chan_list(nchan, dst);	/* put it in the channel list */
	}

	if (debug) {
		dump_channels(l);
	}
}

/* pack the IRCNODE list into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */
static
struct ircnode *
pack_array(struct irclist *l)
{
	struct ircnode *head, *n;
	unsigned int c = l->ipcount;  /* count of hash entries allocated */
	int i = 0;
	int hcounter = 0;

	head = (struct ircnode *) malloc(c * sizeof (struct ircnode));
	if (head == NULL) {
		syslog(LOG_ERR, "malloc failure in irc pack_array, count %d",
			l->ipcount);
		return (NULL);
	}
	/* walk the hashlist
	*/
	for (i =0 ; i<IRCHASH; i++) {
		n = l->iphash[i];
		/* if n is non-nulln use, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			head[hcounter].ipsrc    = n->ipsrc;
			head[hcounter].sport    = n->sport;
			head[hcounter].dport    = n->dport;
			head[hcounter].joinsrc  = n->joinsrc;
			head[hcounter].joindst  = n->joindst;
			head[hcounter].pingsrc  = n->pingsrc;
			head[hcounter].pingdst  = n->pingdst;
			head[hcounter].pongsrc  = n->pongsrc;
			head[hcounter].pongdst  = n->pongdst;
			head[hcounter].privsrc  = n->privsrc;
			head[hcounter].privdst  = n->privdst;
			head[hcounter].channels = n->channels;
			head[hcounter].serverhits = n->serverhits;
			/* synthesize counters */
			head[hcounter].jointotal  = n->joinsrc+n->joindst;
			head[hcounter].pingtotal  = n->pingsrc+n->pingdst;
			head[hcounter].pongtotal  = n->pongsrc+n->pongdst;
			head[hcounter].privtotal  = n->privsrc+n->privdst;
			head[hcounter].nodehits  = head[hcounter].jointotal + head[hcounter].pingtotal 
				 + head[hcounter].pongtotal +  head[hcounter].privtotal;
			n = n->next;
			hcounter++;
		}
	}  
	return(head);
}

static int maxchanips;

static
struct channel *
cpack_array(struct irclist *l)
{
	struct channel *head, *n;
	unsigned int c = l->chancount;  /* count of hash entries allocated */
	int i = 0;
	int hcounter = 0;

	maxchanips = 0;
	head = (struct channel *) malloc(c * sizeof (struct channel));
	if (head == NULL) {
		syslog(LOG_ERR, "malloc failure in irc cpack_array, count %d",
			l->chancount);
		return (NULL);
	}
	/* walk the hashlist
	*/
	for (i =0 ; i<IRCHASH; i++) {
		n = l->chanhash[i];
		/* if n is non-nulln use, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			head[hcounter].name     = n->name;	/* !!! */
			if (n->noips > maxchanips) {
				maxchanips = n->noips;  /* botserver measure */
			}
			head[hcounter].noips    = n->noips;
			head[hcounter].hits     = n->hits;
			head[hcounter].joins    = n->joins;
			head[hcounter].privmsgs = n->privmsgs;
			head[hcounter].nextlink  = n->nextlink;
			n = n->nextchan;
			hcounter++;
		}
	}  
	return(head);
}

static
int
sort_by_totalcount(e1,e2)
struct ircnode *e1, *e2;
{
        if (e1->nodehits > e2->nodehits) return -1;
        if (e1->nodehits < e2->nodehits) return 1;
        return 0;
}

static
int
sort_by_ipcount(e1,e2)
struct channel *e1, *e2;
{
        if (e1->noips > e2->noips) return -1;
        if (e1->noips < e2->noips) return 1;
        return 0;
}

void
printIrc(FILE *fd, struct irclist *l, struct synlist *sl, int maxcount, char *ircfile)
{
	struct ircnode *head;
	struct channel *chanhead;
	int c = l->ipcount;
	int chancount = l->chancount;
	int i,j;
	FILE *ifd;
	struct irclink *nl;
	extern int debug;

	if (debug) {
		dump_channels(l);
	}
	
	/* print out global message stats to mon.lite in ircstats scalar which goes to an RRD
	 * this scalar tuple gives us counts for basic network-wide irc messages.
	*/
	fprintf(fd, "ircstats: %d: %d: %d: %d:\n", l->totaljoins, l->totalpings, l->totalpongs, l->totalprivmsgs);

	fprintf(fd, "ircmaxips: %u:\n", maxchanips); 

	/* counts for total number of irc nodes as RRDTOOL to mon.lite
	*/
	fprintf(fd, "ircnodes: %u:\n", c);

	/* next deal with ircnode stats
	 * 1. sort by most busy node. (total node hits which we must synthesize here)
	 * 2. sort by total joins (sum both counters), total pings, total pongs, total privmsgs
	*/
	ifd = fopen(ircfile, "w");
	if (ifd == NULL) {
		syslog(LOG_ERR, "OOPS - possible late config error: fopen failure on %s\n", ircfile);
		return;
	}

	/* convert hash list to dynamic array
	*/
	head = pack_array(l);
	if (head == NULL) {
		return;
	}

	/* sort the dynamic array of ircnodes first by total count
	*/
        qsort((void *)head, c, sizeof(struct ircnode), sort_by_totalcount);

	/* now print out the sorted array of ircnodes to irc.txt 
	 * tuple format here:  ip src: total msg count: jointotal: pingtotal: pongtotal: privmsgtotal:
 	 *	joinsrc:joindst:pingsrc:pingdst:pongsrc:pongdst:privmsgsrc:privmsgdst:channels:serverhits:
	 * 16 entries per tuple
	 * we expect backend to further process this file into a report.
	 * tuple type: IRCNODE
	*/
	fprintf(ifd, "#IRCNODE\n");
	fprintf(ifd, "#ipsrc:totalmsg:joins:pings:pongs:privmsgs:joinsrc:joindst:pingsrc:pingdst:pongsrc:pongdst:pmsrc:pmdst:channels:serverhits:synwork:sport:dport\n");
	for ( i = 0; i < c; i++) {
		/* print limited stats unless maxcount is 0
		*/
		if (maxcount)
			if (i == maxcount)
				break;
		fprintf(ifd, "%s:%u:%u:%u:%u:%u:%u:%u:%u:%u:%u:%u:%u:%u:%u:%u:%d:%u:%u:\n",
			intoa(head[i].ipsrc),
			head[i].nodehits,
			head[i].jointotal,
			head[i].pingtotal,
			head[i].pongtotal,
			head[i].privtotal,
			head[i].joinsrc,
			head[i].joindst,
			head[i].pingsrc,
			head[i].pingdst,
			head[i].pongsrc,
			head[i].pongdst,
			head[i].privsrc,
			head[i].privdst,
			head[i].channels,
			head[i].serverhits,
		        searchSyn(sl, head[i].ipsrc),		/* get work from synlist */	
			head[i].sport,
			head[i].dport 
			);
	}


	/* sort channels by most number of ips */
	chanhead = cpack_array(l);
	if (chanhead == NULL) {
		free(head);
		return;
	}

        qsort((void *)chanhead, chancount, sizeof(struct channel), sort_by_ipcount);

	/* irc channel tuple is as follows:
	 * channame: hits: joins: privmsgs: ipcount: ip_src1, ip_src2, ... ip_srcN:
	 * tuple type: IRCCHANNEL
	 */
	fprintf(ifd, "#channame:hits:joins:privmsgs:noips:ip list, ip, ip ...\n");
	/* now deal with channel output which goes into same irc.txt file
	*/
	fprintf(ifd, "#IRCCHANNEL\n");

	/* note: for the purposes of back-end parsing,
	 * a hash mark (#) can occur in a channel name, therefore
	 * from this point on it is NOT a comment.
	 */

	for ( i = 0; i < chancount; i++) {
		fprintf(ifd, "%s:%u:%u:%u:%u:",
			chanhead[i].name,
			chanhead[i].hits,
			chanhead[i].joins,
			chanhead[i].privmsgs,
			chanhead[i].noips 
		);
		nl = chanhead[i].nextlink;
		for ( j = 0; j < chanhead[i].noips; j++) {
			fprintf(ifd, "%s,", intoa(nl->ircnode->ipsrc));
			nl = nl->nextlink;
		}
		fprintf(ifd,":\n");
	}
	fclose(ifd);

	/* free up the various dynamic arrays for the ircnode
	*/
	free(head);
}
