/*
 * config.c - config info for ourmon utility
 * 
 * purpose: read/parse/store tag values from /etc/ourmon.conf configuration
 * file.
 * 
 * file syntax: ascii lines, one tag per line, may have N values depending on
 * tag skip lines/and comments, # in 1st column is comment
 *
 * All filters are configured with one line, the name of the filter
 * comes first.  optional parameters may follow on a per filter basis.
 * Filter spec is provided below.
 * 
 * tag syntax:
 * 
 * 0. pkts  
 *
 *	There is an implicit filter output that counts 	packets caught 
 *	and dropped.  These variables are zeroed
 *	at this point by the less than wonderful idea of
 *	zeroing out the BPF at write output file time.  
 *      The information comes from pcap (pcap_stats()).
 *	It allows us to determine if the filter/s are causing
 *	packet drops.  Note that two interfaces can be supported.
 *
 * 0.1 sysname.
 *
 *      Provide an optional label for the system.  If this
 * 	is provided.  E.g., "bob", then the pkts RRD will be named
 *	pkts.bob, and omupdate.pl will produce a unique per-probe
 *	pkts filter.   The resulting packets graph will have a label 
 *	like pkts.bob, and be per-probe.
 *	This allows a number of systems to run different
 *	filters and thus allow parallelism in the front-end.
 *	ourmon (the front-end) will place this sysname first
 *	(if provided) in the output mon.lite file.
 *
 * 0.2 nopkts
 *
 *	This is useful for simple parallelism.  If supplied
 *	in a config file, the pkts output is dropped and is
 *	not put into mon.lite.
 *
 * 1. fixed_ip
 * 
 *	syntax: fixed_ip
 *
 *      Fixed filter; i.e., keys are not dynamic/changeable.
 *	tcp/udp/icmp/remainder (4 keys).
 *	This filter is zeroed out at every write to output.
 *
 * 2. fixed_tcp3
 *
 *	syntax:  fixed_tcp3 tcp_port1  tcp_port2
 *
 *	Fixed filter.  captures and counts packets with tcp dest or
 *	src port equal to port1 or port2.  counts packets (extra)
 *	that do not belong to that filter spec of course.
 *	Note tcp_port1 for example should be a well-known port
 *	and will cause the count of packets with src OR dest port
 *	set to that value. This filter is zeroed out at every write 
 *	to output.
 *
 * 3. fixed_iprange1
 *
 *	syntax: iprange1  ipaddr1 ipnetmask1  ipaddr2 ipnetmask2 s1 s2
 *
 *	s1 and s2 are strings that must be supplied as they are used
 *	in the output. s1 should describe the 1st network, and s2
 *	should describe the 2nd network.
 *
 *	Measures total ip exchange in bytes between first ip network
 *	and second ip network.  Remainder to extra counter; i.e.,
 *	two numbers will be printed out.  More ipranges may be added
 *	later.  This is the semantic equal to tcpdump:
 *		net X and net Y
 *
 *  4. defunct ... 
 *
 *  5. fixed_cast 
 *	
 *	syntax:   fixed_cast  ip_addr  netmask
 *
 *	The focus of this filter is on dividing up packets
 *	into broadcast/multicast/unicast groups.  If the
 *	interface/s used are ethernet interfaces, then the
 *	filter bases its decisions on ethernet dst addresses.
 *	If the interface/s used are NOT ethernet interfaces,
 *	then IP addresses are used.  An IP address (which is used
 *	for guessing at the broadcast IP address) must be supplied.
 *	There are 4 buckets, multicast, unicast, broadcast
 *	and xtra.  Note that if we are running on ethernet,
 *	then decisions are 100% based on ethernet level information,
 *	and the results are not IP specific.  If non-ethernet interfaces
 *	are used,  the xtra bucket will capture all other traffic, which could
 *	include ARP, non-IP, and some directed broadcast, missed because
 *	the IP/netmask info above may miss some kinds of directed broadcast
 *	in a e.g., VLSM-based network.
 *
 * 6. fixed_size:  
 *
 *	syntax: fixed_size
 *
 *	4 buckets for measuring packet sizes (at layer 2):
 *	essentially tiny, small, medium, mtu-ish (1500)
 *
 *	tiny: <= 100
 *	small: < 500
 *	medium: < 1000
 *	large: <= 1500
 *
 *	Currently cannot reset bucket sizes.
 *
 * ------- above are static, remainder are dynamic -----------
 *
 *  7. topn_ip: 
 *	syntax: topn_ip N
 *
 *	N is the desired # of top N ip destinations to be output.
 *	N must be between 10..100 as a multiple of 10 (10, 20 ...100) 
 *      in order to produce 1 to 10 pics with 10 flows per pic.
 *	Information is sorted on a flow basis where a flow is roughly
 *	(ip source, ip dst, next protocol, L4 src port, L4 dst port).
 *	For ICMP L4 src port means the ICMP major code, and L4 dst
 *	port means the ICMP minor code.
 *
 *	The topn_ip list produces 4 lists:
 *	topn_ip
 *	topn_tcp
 *	topn_udp
 *	topn_icmp
 *
 *	The first item following the topn_ip, topn_tcp, etc., tag
 *	in the mon.lite file is the total count of flows during the
 *	sample period.  
 *
 *	topn_ip shows flows based on any ip protocol (tcp, udp, GRE, whatever).  
 *	TCP, UDP, and ICMP are specific to their ip layer protocol.  
 *	None of the graphs show "xtra" bytes.
 *
 *	In addition topn produces two additional mon.lite outputs.
 *
 *	hs_stat -  gives runtime stat information on the hashing algorithm
 *	used.  In general most of this information is only useful
 *	to an implementor.  However the hash list insert count is currently
 *	graphed by the back-end.  This is because inserts are caused by
 *	"new" flows, and inserts highly correlated with denial of service
 *	or scanning attacks as they generate great numbers of random 
 *	ip src, and ip dst addresses (or random ports).  Thus the
 *	back-end is currently graphing inserts.
 *
 *	The back-end also graphs the total number of flows seen by the topn
 *	mechanism.  This is also very useful for watching for random
 *	ip src/dst or port addresses, and can be useful in detecting
 *	scanning attacks.
 *
 *	topn_smallmesh basedirectory
 *
 *      If provided, 2-way flows for any IP protocol are collected.
 *      The flows collected must have a non-zero payload, and have
 *	"small" data payloads (< 600 bytes or so).  In general this
 *	rules out ping.  It does not rule out DNS packets from local
 *	DNS servers (this will require a future whitelist for this
 *	particular feature).  The output file is called smallmesh.txt.
 *
 *	2-way flows are flows between host A and B of the form:
 *	IPa.porta->IPb.portb, and IPb.portb->IPa.porta.  Both flows
 *	must have small data payloads.  
 *
 *  8.  topn_syn:
 *      syntax: topn_syn N, N < 100. 
 *
 *	N should be a multiple of 10, 10..100.  N is the top N TCP
 *	syn senders.  A list of ip addresses is kept.  For each ip
 *	address we count the number of syn packets where the ip address
 *	is the source.  We sort on this count.  In addition, we count
 *	the number of fins seen where the ip address is considered as the
 *	source address.  Resets are also counted, for resets the ip destination
 *	is the target ip.  Thus we are considering the number of syns and fins
 *	a given IP sends, and also the number of resets a given IP receives. 
 *	The output thus shows syn, fin, and reset counts.  We sort on syns,
 *	and simply present fins/resets as information.
 *
 *	An additional two syn scanning oriented functions are available 
 *	if turned on.  
 *
 *	syntax: 
 *	topn_syn_wormfile  basedirectory
 *
 *	If the wormfile is specified, the topn_syn print function will also
 *	produce a companion file called "tcpworm.txt".  
 *	This file will be written in the basedirectory
 *	(e.g., /home/mrourmon/web.pages).  The file is a list
 *	of ip source addresses that have satisfied the following
 *	constraint.  syns_sent - fins > 20.  
 *	The above is of course a rude heuristic but seems to be useful
 *	for capturing distributed denial of service attacks.
 *	The back-end tcpworm.pl app. is used to output the results in a readable form.
 *
 *	topn_syn_homeip	  ipaddress/netmask
 *
 *	Make sure this is correct.  It is useful.
 *	Example:
 *	topn_syn_homeip	  10.0.1.0/24
 *
 *	If homeip is provided, the wormfile output is also
 *	processed in RRDTOOL style to determine if the addresses belong to "us"
 *	(the home ip address range) or "them" (addresses do not
 *	belong in the home range).  An additional output is placed
 *	in the mon.lite file as follows:
 *
 *	tworm: totalcount: us_count : them_count :
 *
 *	Thus the ip addresses in tcpworm.txt are summarized in terms
 *	of a total count of IPs which is divided up into two subordinate counts
 *	of "us" versus "them".  One can thus determine if syn scanners
 *	are local, or remote, and how many total, or local scanning ips are active
 *	over time.  This data is captured by omupdate.pl, and can thus
 *	be basedlined via RRDTOOL.
 *
 *	At this time 1 to 10 home IP net/mask config commands may be given
 *      to ourmon.  This is the equivalent of a logical OR for the definition
 *	of home.  E.g.,
 *
 *	topn_syn_homeip 10.0.1.0/24 
 *	topn_syn_homeip 192.168.1.0/24 
 *
 *	defines the HOME as two subnets.
 *
 *	tworm_weight  weight
 *
 *	tworm_weight is a value that ranges from 0..100.  It is used
 *	to pre-filter the counts given in the tworm summary based
 *	on the built-in worm weight that is roughly # of tcp control
 *	packets divided by the total # of tcp packets per ip source.
 *	Thus the tworm output can be filtered say at 80%, or only
 *	ip sources with worm weights greater than or equal to 80 will be counted.
 *	This pre-filter on the worm weight at this time ONLY affects
 *	the tworm mon.lite count and does not affect the contents of tcpworm.txt.
 *	(This may change later).  Thus the filter weight defaults to 0.
 *
 *	topn_syn_sort  0
 *
 *	This is a boolean variable.  If not given, the topn syn list is sorted
 *	by the total number of syns.  If supplied with value set to 0, the topn syn list is pre-sorted
 *	so that all 100% work weight ip sources are shoved to the top of the sort,
 *	and any remaining tuples with a smaller work weight are then sorted by
 *	syns.  E.g., if you had two "synning" IP sources in the syn list,
 *	one with 100%, and two with 99% but more syns.  The 100% "synning" ip source
 *	would come first.  The two 99% rated ip sources would follow sorted by
 *	their respective syn counts.
 *	
 *	topn_syn_wormdiff 20
 *
 *	This number is used as the magic number to determine that subset of IP srcs from
 *	the TCP syn tuple that are determined to belong in the tcpworm.txt output file.
 *	A magic equation is used for each IP source address.  Per IP src: SYNS - FINS > wormdiff.
 *	If this is true, the ip source in question is placed in the output tcpworm.txt file.
 *	That set of IP sources is deemed to be "noisy" for some reason, and often contains
 *	syn scanners and other forms of attacks, although it also contains noisy web servers
 *	and P2P clients.  Of course the latter may not be desired at a given site.
 *	Note: if this value is set to 0, one is effectively dumping the entire synlist.
 *	The current default value is 20.
 *
 *	topn_syn_p2pfile  basedirectory
 *
 *	If the p2pfile option is specified, the topn_syn print function will also
 *	produce a companion file called "p2pfile.txt".  
 *	This file will be written in the basedirectory
 *	(e.g., /home/mrourmon/web.pages).  The file is a list
 *	of ip source addresses that have been identified as either a known
 *	P2P application like bittorrent, edonkey, gnutella, morpheus, or kazaa,
 *	or an IRC application.  The output is a syn tuple.  The internal flags
 *	field is set to indicate information about the type of app identified.
 *	The back-end tcpworm.pl app. is used to output the results in a readable form.
 *
 *	topn_syn_syndump  basedirectory
 *
 *	If this option is specified, the syndump of the home IPs only is put in the file called
 *	syndump.txt.  Warning: this file may be quite large.  Normally the 
 *	file size is reduced and parts of it are put into the
 *	twprom or p2pfiles.  On the other hand, this may be of use at a small installation
 *	or for experimental work.  
 *
 *	topn_syn_emaildump  basedirectory
 *
 *	If this option is specified, emaildump.txt is created in the
 *	output directory in syntuple format.  Any IP src that sent
 *	TCP syns to dst port 25 is included.  Note there is also
 *	a port 25 specific workweight, although it is provided
 *	in any output instance of the syn tuple.  This facility may
 *	be useful for catching spam machines.
 *
 *  9. topn_icmperror:
 *      syntax: topn_icmperror N, N < 100. 
 *
 *	This filter generates two different output lists.  One
 *	list called icmp_errorlist simply gives the top generators
 *	of ICMP errors based on IP src.  The second list is UDP
 *	focused, and uses a weighted function to rank systems
 *	according to UDP packets send/received times significant ICMP
 *	errors.  It is called the udperror_list.
 *
 *	N should be a multiple of 10, 10..100.  N is the top N IP
 *	source addresses that accrue ICMP errors back to them.
 *
 *	For the icmp_errorlist:
 *	A list of IP source addresses is kept.  For each IP
 *	address we count the number of icmp errors that include
 *	icmp redirects, unreachables (of any important kind), and ttl exceeded, 
 * 	where the ip dst of the icmp error messages is the same as that 
 *	of a known IP src ip address.  We sort on the total icmp error count.  
 *	The mon.lite output for the icmp_errorlist (with one tuple) is
 *	as follows:
 *
 *	icmp_errorlist: ip address: totalicmp_error_count: tcp_count:
 *		udp_count: icmp_ping_count: unreach_count: redirect_count:
 *		ttl_count: next tuple ...
 *
 *	This means for a given ip src address:
 *		total_icmp_error_count: all icmp errors returned to this addr
 *		tcp_count: total count of tcp packets sent from this addr
 *		udp_count: total count of udp packets sent from this addr
 *		icmp_ping_count:  pings sent (echo request only)
 *		unreach_count: ICMP unreachables returned to this address
 *		redirect_count: ICMP redirects returned to this address
 *		ttl_cou0nt: ICMP ttl exceeded returned to this address
 *
 *	The second list is a weighted UDP error list.  Its goal is to
 *	find UDP scanners based on ICMP errors returned to a given
 *	scanning IP source address.  The mon.lite output is as follows:
 *
 *	udperror_list: ip address: weight: udp_packets_sent: 
 * 		udp_packets_received:  total_icmp_errors: next tuple ...
 *
 *	This list is sorted according to a weight function. 
 *
 *	W = (udp_pkts_sent - udp_pkts_received) * total_icmp_errors.
 *	
 *	As before the ICMP errors are redirects, unreachables, and ttl
 *	exceeded.  It is likely that a UDP scanner is generating host
 *	unreachables, and possibly ICMP administrative prohibited 
 *	unreachables.  UDP pkts sent and recieved are both shown.
 *	Note: it is a bad sign if a system is sending UDP packets
 *	and not getting them back AND generating many ICMP errors. 
 *	It is possible that such a system is a media-based system
 *	that is simply misbehaving.  It is also possible that it is
 *	the slammer worm.
 *
 *  10. topn_port:
 *	syntax: topn_port N, N < 100.
 *
 *	N should be a multiple of 10, 10..100.  N is the top N ports
 *	src/dst for UDP or TCP flows.  Two lists are output sorted by
 *	bytes, one for TCP flows and one for UDP flows.  The number of
 *	flows seen during the period is the first number.  Subsequent
 *	numbers are 4-tuples of the form (byte, port, src_port_packet_count,
 *	dst_port_packet_count).  If a packet is sent to UDP dst port 512 or
 *	recv from UDP port 512, it counts against the port tuple 512.  
 *	The src/dst packet counters simply reflect the number of times a given
 *	port was seen.  
 *
 *  11. ip port scanning filters (2): 
 *      syntax: topn_scans N
 *              topn_port_scans N
 *
 *	N may be set in units of 10 from 10 to 100.
 *	The back-end produces a separate graph for each successive
 *	unit of 10. N specifies the top N ip address scanners.
 *
 *	topn_scans is a filter aimed at detection of IP destination 
 *	address scanning by a single IP source.  It produces tuples
 *	of the form: ip address: count, where count is the count
 *	of distinct ip destinations for that ip address during
 *	the sample period.  We sort on the count.  Thus this detects
 *	single ip src to many ip destination scans.
 *
 * 	topn_port_scans is aimed at scanning from one ip source
 *	of multiple TCP or UDP ports. 	topn_port_scans three
 *	outputs (three lists):
 *
 *	ip_portscan:
 *	tcp_portscan:
 *	udp_portscan:
 *
 *	Each has a 2-tuple consisting of an ip src address and a count.
 *	The count indicates the number of distinct L4 ports referenced
 *	by the ip src during the sample period.  tcp_portscan only
 *	counts tcp port scanners.  udp_portscan only counts udp port
 *	scanners.  ip_portscan counts both tcp and udp port scanners.
 *
 *	The underlying goal here is to detect manual scanning say
 *	done with nmap, or dynamic scanning done by viruses.  P2P
 *	apps may also be detected.
 *	
 *  12. bpf: 
 *
 *	Two possible forms:  a basic bpf spec, which has a fixed number.
 *	and a concatenated bpf which means a basic bpf specification,
 *	with > 1 bpf filters in the same RRD graph.  
 *
 *	The basic spec can be viewed as consisting of two graphs
 *		bpf bytes and extra (xtra); i.e., bytes that
 *		did not match the bpf graph.
 *	Each bpf specification must have a unique label which is
 *	used to label the graph.  
 *
 *	Each bpf string (the filter spec) must also have its own label.
 *
 *	Up to N (MAXBPFS) (a small number) of concatenated bpfs can
 *	be tacked onto a basic bpf, thus producing one rrd graph
 *	with up to MAXBPFS number of lines (plus the extra line) on
 *	the graph.
 *
 *	syntax:
 *
 *	The basic specification has this form:
 *
 *	bpf "label" "per-filter-label" "filter"
 *
 *	label - label for entire bpf specification
 *	per-filter-label - short name for the filter 
 *	filter - the bpf filter spec
 *
 *	The concatenated spec syntax:
 *
 *	bpf-next "per-filter-label" "filter"
 *
 * 	This form must follow the bpf form sequentially in
 *	the configuration file.
 *
 *	bpf-noxtra  turns off xtra tag (for now, simply
 *	doesn't store it.  still in perl backend)
 *
 *      cbpf "tag" "label" "per-filter-label" "filter"
 *      cbpf-next "tag" "per-filter-label" "filter"
 *
 *		tags	filter expressions
 *	      	net  	net/mask		DONE
 *		icmp	major minor		DONE
 *		ping    no filter spec		DONE
 *		tcprange  loport hiport
 *		udprange  loport hiport
 *		ipproto number
 *			note: as opposed to numbers,
 *			certain strings are understood.
 *			ip (means all)
 *			tcp/udp/icmp are also understood.
 *		tcpflag s f r
 *		port	t port u port		
 *		TBD: extend net to deal with > 1 pair.
 *		net	net/mask net/mask net/mask		up to say 5
 *
 * bpf-packets
 * bpf-bytes
 *
 *	The bpf graphs default to counting bytes (bpf-bytes).
 *	bpf-packets thus means that any bpf graphs that start
 *	after bpf-packets should count packets.  bpf-bytes 
 *	overrides bpf-packets mode.  It is not possible to
 *	change from packets to bytes *within* a bpf graph,
 *	but one can do it between graphs; e.g.,,
 *	
 *	bpf-packets
 *	bpf blah blah
 *	bpf-next
 *	bpf-bytes
 *	bpf blah blah
 *	bpf-noextra
 ****************************************************************
 * triggers  (automated packet capture)
 ****************************************************************
 * trigger overview:
 *
 * The trigger mechanism allows the probe to dynamically store
 * a set of "interesting" packets which may be filtered internally
 * with various BPF expressions and then written out to a tcpdump style dump file.
 *
 * Typically this is done when a back-end graphics "curve" exceeds
 * some configuration-time threshold integer peak where both the threshold and current values  
 * are calculated in terms of back-end values, either in bits per second, or packets per second.  
 * A per trigger type dump directory must be specified (say /usr/dumps) and each
 * trigger places a unique instance of a dump file filled with
 * the specified count of packets.  The instance file name is
 * always of the form:  trigger_tag.<timestamp>.dmp.  The timestamp
 * records both the time the dump was made and allows the dump
 * to have a unique filename.  Only one trigger of each kind may be active at a time.
 *
 * For example, one can have both the udp error weight trigger and the
 * tworm triggers active, but it is not possible to have two tworm
 * triggers.  Each trigger has some sort of BPF expression capability
 * associated with it, which is not user specifiable (except in the case
 * of BPF filter-set triggers).  For example the udp weight trigger
 * dynamically determines the top IP host address, and stores count
 * packets of the form:  "host IP and (udp or icmp)", thus storing
 * udp packets and any ICMP errors associated with the host IP that
 * caused the trigger to activate.  In general triggers are associated
 * with some topn lists, and the BPF filter-set BPF expressions. 
 *
 * Current trigger types include:
 *
 *	topn_ip - topn IP (any IP protocol) trigger on total flow count.
 *	tworm - total worm count.
 *	topn_icmperror (udplist) - udp error weight for a specific IP host.
 *	topn_scans - 1 trigger that is actually 4 triggers, 
 *		one per topn_scan list type (IP, L4, TCP, UDP).  This
		is a per IP host trigger. 
 *	(pkts)drops - triggers on too many drops from the pkt capture filter.
 *	bpf - any BPF expression used in a BPF filter-set may have a trigger.  
 *
 * Note: there is a maximum count of triggers.  (MAXTRIGGERS see trigger.h).
 *
 * It is important to note that triggers may be more useful in terms of
 * the packets captured if the particular BPF expression (implicit as with tworm,
 * or explicit as with a BPF filter-set expression) is more specific.
 * For example, a BPF expression that captures all TCP packets may not show
 * anything useful.  On the other hand, the udp error weight trigger only
 * captures UDP packets for a particular IP source.  Thus the tcpdump in
 * question may show the exact nature of an attack.  This problem can
 * be called the "signal to noise problem".    A BPF expression on the order
 * of "net X and tcp" may be more useful than just "tcp" or "net X" by themselves.
 *
 * In ourmon.conf, triggers in general have the following syntax:
 *
 * trigger_name threshold packet_count dump_dirname.
 *
 * 
 * threshold - specify in bits/sec or pkts/sec a trigger threshold.
 * When the named filter reaches that threshold, packet capture will begin.
 *
 * packet_count - terminate when <count> packets stored.
 *
 * dump_dirname - directory in which to store the dump file.  
 *
 * Packet dumping is turned off when either the count is exceeded
 * or the trigger threshold flow count is less than the current runtime value 
 * -- which ever comes first.  Packet dumping after being triggered cannot recur
 * between trigger on and trigger off times which are based on the
 * threshold being crossed (either above it or below it).  For example
 * assume the packet capture count is 100000, and the value is 2000.
 * If at mon.lite creation time, the value is 2000, a trigger_on event
 * is started and 100000 packets (barring trigger_off occuring because
 * threshold < value) will be stored.  If the count is exhausted and
 * the threshold is still > than v, no more packets will be stored in
 * the file.  Thus it is not possible for one trigger_on event to cause
 * more than one output file.  Only one trigger on event is possible
 * for the topn_ip type.  However this does not prevent other triggers
 * from working (say with tworm).  Also when this trigger is turned
 * off, it may be retriggered later in which case data will be put
 * into a separate file according to the timestamp.
 *
 * Trigger_on and trigger_off events are put into mon.lite as elog
 * messages and are recorded in the back-end daily event log.  As a result
 * it is possible to more or less determine when the trigger event
 * occurred and to easily see the event from the web interface.  
 * The trigger dump filename is passed to the event log.
 *
 * Note: at this time ourmon has no way to determine if the 
 * dynamically created file has been created successfully
 * or if there is runtime storage room within the file system
 * in question.  If the file cannot be created an elog message
 * will be sent.  ourmon will not exit.
 *
 * Note: in general, triggers are expressed in terms of their back-end
 * graphics realization.  If the back-end graphics is displayed as
 * pkts/sec, then the config threshold should be pkts/second. 
 * If the back-end value is count/period (30 seconds), then the
 * config threshold should be the count for the sample period.
 * E.g., the tworm trigger is IPs per sample period.  The 
 * drops trigger is packets per second.  
 *
 * 13.
 * trigger_topn_ip threshold packet_count dump_dirname 
 *
 * The flow count is graphed in the back-end graphics flow.html page.
 *
 * This trigger will dump packets when the flow count for
 * topn_ip flows exceeds the threshold.  Packets are stored in
 * a file with the form:
 *
 * 	topn_ip.<timestamp>.dmp.
 *
 * This particular trigger uses an internal
 * dynamic BPF expression to determine if packets should be stored.
 * At trigger_on time the trigger code looks to see if tcp, udp, or
 * icmp events are greater than the trigger, if so, the BPF expression
 * is made more narrow, e.g., "tcp".  If this is NOT the case, the
 * expression "ip" which is the default for this expression is used.
 * The BPF expression is only made more narrow if for example the TCP
 * flow count per second is greater than the threshold but the UDP and
 * ICMP flow counts are both less than the threshold.  This logic also
 * holds for UDP and ICMP flow counts.  When in doubt the trigger
 * expression is "ip".
 *
 * Note: this trigger is aimed at the total flow count, not any
 * individual single flow instance.  The flow count is the second
 * parameter in topn_ip, topn_tcp, etc. mon.lite outputs.
 *
 * 14.
 * trigger_worm threshold packet_count dump_dirname
 * 
 * This trigger applies to the tworm total count (not us, not them, total).  
 * If the total value for tworm counts (us+them) exceeds the supplied threshold value,
 * a packet dump will begin for count packets in the dump_dirname in
 * a dynamically created filename.  The filename has the form:
 *
 * 	tworm.<timestamp>.dmp.
 *
 * The semantics for this trigger are similar to the semantics for
 * the preceeding trigger.  However only TCP SYN packets are stored.  
 * Note it can be useful to use the saved back-end portreport.txt file to try and
 * reproduce the port signature report at the trigger time.  This can
 * easily help narrow down the attacking instances in terms of 
 * IP source addresses or destination TCP port numbers.  Put another way,
 * it is useful to know the IP address of the attacker a priori or know
 * a port signature in the case of multiple attackers.  The IP address of
 * the attackers can then be used to "search" the tcpdump file to see exactly
 * what the attackers were up to.  Also one may simply use ports of interest
 * as indicated in the portreport.txt file.  For example, assume that the portsignature.txt
 * file shows that port 6666 was attacked, then one can do:
 *
 * tcpdump -n -r "*dump*file" tcp and port 6666
 *
 * to narrow down the packet dump to a more interesting set of packets.
 *
 * 15.
 * bpf_trigger bpf_graph_name bpf_line_label threshold packet_count dump_dirname
 * 
 * This trigger works with any expression found in its corresponding BPF filter-set. 
 * Ourmon allows user specified BPF expressions which result in graphics-engine
 * RRDTOOL-based graphs.  Every BPF filter is identified by 'bpf_graph_name' and
 * 'bpf_line_label'. For any pre-declared BPF filter, a trigger can be associated with it
 * by declaring the filter's 'bpf_graph_name' and 'bpf_line_label'. Please see the 
 * BPF section above for the syntax and semantics of bpf filter-sets. 
 * Note the threshold value is compared to the internal count in terms of
 * packets or bytes -- whichever was used when the BPF filter-set was declared.
 * The dynamically created filename has the following form:
 *
 *      bpf_<bpf_graph_name>_<bpf_line_label>.<timestamp>.dmp.
 *
 * The semantics for this trigger are similar to that for the preceeding
 * trigger. Only those packets which match BPF filter spec are stored.
 *
 * It is useful here to refer to the back-end/graphics graphs in order to
 * determine a threshold value.  One may watch the graphs for a week or so
 * and then determine a threshold.  Note that in general the more specific
 * the BPF expression, the better as this will tend to help you filter
 * "signal" as opposed to "noise".
 *
 * 16.
 * topn_scans_trigger threshold packet_count dump_dirname
 *
 * The topn_scans_trigger works for four kinds of topn_scan list and
 * stores packets for each kind of topn_scan list separately. Please see the 
 * topn_scan section for the semantics of topn_scan. The trigger (set) for topn_scan will 
 * dynamically configure its BPF fiilter expression, and capture filename.
 *
 * For the basic ip_scan list, the trigger is applied to the number of destination 
 * addresses related to the IP src address which is the top one in the ip_scans list. 
 * If the number of destinations this src address has exceeds the supplied 
 * threshold value, a packet dump will begin. The capture filename has the form:
 *
 *      ip_scans.<timestamp>.dmp
 *
 * This trigger will capture and store any IP packets sent by the IP source in question.
 *
 * The trigger for the topn_scan_port list is applied to the total count of L4 ports 
 * for the first host in the ip_portscan list.
 * If the number of ports is greater than the threshold value, a packet dump
 * will begin and the filename has the form:
 *
 *      ip_portscan.<timestamp>.dmp
 *
 * The trigger for the topn_scan_port_tcp list is applied to the total count of TCP dst ports.
 * If the count of TCP ports for the top IP src in the TCP scan list exceeds the threshold,
 * packets will be stored in a filename with the form:
 *
 *      tcp_portscan.<timestamp>.dmp
 *
 * The trigger for the topn_scan_port_udp list is applied to the total count of UDP dst ports.
 * If the count of UDP ports for the top IP src address in the udp_portscan list exceeds the threshold 
 * value, packets will be stored in a filename with the form:
 *
 *      udp_portscan.<timestamp>.dmp
 *
 * Note: there is really only one trigger capability here, although it can be dynamically
 * narrowed a bit.  Only one scan trigger can be triggered at one time.  For example, it is
 * not possible to have the tcp_portscan and the udp_portscan both running concurrently.
 * They can however run in a serial fashion, one after the other.
 *
 * 17.
 * drop_trigger threshold packet_count dump_dirname
 *
 * This trigger is triggered when the current drop count (as seen in pkts.html)
 * exceeds the trigger value.  If packets are not dropped due to mundane reasons
 * (like too much ourmon work or not enough CPU power), packets may be dropped
 * due to DOS attacks.  Thus it is possible that this trigger might actually
 * capture a DOS attack. The capture filename has the form:
 *
 *      drop.<timestamp>.dump.
 *
 * The semantics for this trigger are similiar to that for the preceeding
 * trigger.  
 *
 * Note: the threshold here is supplied as if it is pkts/second so that it
 * matches the back-end RRDTOOL display (in other words the value will be multiplied
 * internally by 30 so for example if you supply 1000, that means the drop trigger will
 * be triggered at 30000 for 30000 drops per 30-second period).  Counts associated with the elog message
 * however are expressed with internal mon.lite values.  In other words threshold and packet
 * count are shown in counts per 30 second period.
 * 
 * 18.
 * udperror_trigger threshold packet_count dump_dirname
 *
 * This trigger is based on the udperror_list (topn_icmperror). 
 * Too many packets might mean  a UDP port scan attack. See the topn_icmperror section for
 * more information on the udperror_list. If the udperror weight calculated for the first host in udperror_list  
 * is greater than the threshold, a packet dump will begin for
 * the requested number of packets in a dynamically created filename in the supplied directory.
 * The capture filename has the form:
 *
 *      topn_udp_err.<timestamp>.dump
 *
 * This trigger will capture and store "udp" packets sent to and from
 * the top host in udperror_list.  The internal BPF expression used is:
 * "ip_src and (udp or icmp)".
 *
 ****************************************************************
 * END triggers  
 ****************************************************************
 *
 * 19. irc information.
 * 
 * topn_irc maxcount 
 *
 * Note both topn_irc and topn_irc_file MUST be defined or neither should
 *	be defined.  
 *
 * dump various irc stats. The mon.lite file gets a few RRD tuples related to IRC
 * message counts:
 *
 *		ircstats has total message counts for basic IRC messages:
 *		format: totaljoints, totalpings, totalpongs, totalprivmsgs
 *	
 *		ircnodes: one counter - counts total number of active IRC hosts
 *
 * At this point maxcount should be set to 0 and has no useful semantics.
 * It is a placeholder for possible future features.
 *
 * topn_irc_file basedirectory
 *
 * topn_irc must be set.
 *
 * produces filename irc.txt in dirname.
 *
 * if specified gives us irc host tuple dump for basic stats
 * of IRC message activity per IRC host.  This is the IRCHOST tuple.
 * Also gives us the IRCCHANNEL tuple, which is a list of
 * channels found in use abstracted from PRIVMSG and JOIN irc messages.	
 * Both host and channel irc tuples are put in irc.txt.
 * See info.html or ircscan.c for more information.
 * Backend must further process this file.  The irc.pl script is
 * used for doing that work.  
 *
 * 20.  fixed_l2proto:  
 *
 * L2 type check for ethernet only.  With other L2 headers,
 * it will simply produce no result at this time.
 * pkt counts for IP, IPv6, ARP, and other.
 *
 * 21. honeynet:
 *
 * honeynet  net/mask
 *
 * If a local network can be allocated that is empty, attempts to communicate
 * with the target net/mask will be tagged in various ways.  The first such
 * use is to indicate such a network as an additional tag in the tcp syn
 * portreport.  See the portreport module itself for more information.
 * The P application tag is used to flag packets sent to this net.
 *
 * 22. PCRE layer 7 pattern matching
 *
 * pattern <transport layer> <tagchar> <max packets> <max bytes> <pcre>
 * for example:
 *
 * pattern TCP B 100 2000 "^\x13bittorrent protocol"
 *
 * <transport layer> Only payloads on this transport layer will be matched.
 * Currently supports TCP, and UDP.  Other transport layer values are not supported.
 *
 * <tagchar> One alphabetical character [a-zA-Z] tag associated with the
 * pattern. This tag will be used to indicate a successful 
 * match in the application flags field of the portreport. 
 * Note that <tagchar> is independent from the layer being
 * matched. That is, if the same <tagchar> is associated with
 * a TCP and UDP payload then the second tag will be ignored as
 * it is considered to be a duplicate.
 *
 * Also, certain tags are reserved by ourmon and cannot be used.
 * A boot time config error will prevent those tags from being used.
 * 8 reserved tag letters:  B, G, K, M, H, E, e, and s  are reserved.
 * In addition P and p are reserved as well.
 *
 * <max packets> Maximum number of packets associated with a IP src that
 * will be searched for this pattern.
 *
 * <max bytes> Maximum bytes associated with an IP src that will be
 * searched for this pattern.
 *
 * The payload will be searched for this pattern until <max packets> or
 * <max bytes>, whichever comes first, is encountered. These can also have
 * the value DEFAULT that will cause the built in defaults, 8 for <max
 * packets> and 2048 for <max bytes>, to be used. These default values
 * have been adopted from the l7-filter project.
 * If a pattern has a value of zero for both the maximums then the pattern
 * will always be searched for in payload.
 *
 * <pcre> Packet payloads will be searched with this Perl compatible regular
 * expression (PCRE) to determine if the IP source and destination is
 * marked with the given tag. It is recommended that the pattern be
 * delimited, for example by using quotes, but this is not required.
 *
 * In addition the following two stateful tokens exist and apply to
 * the patterns that they encapsulate:
 *
 * pattern-keepNULL
 * pattern-removeNULL
 *
 * In order to interoperate with patterns 
 * found at: http://l7-filter.sourceforge.net/
 * by default NULL characters will be removed from the L7 payload before any
 * pattern matching is performed. This can be changed by preceding a
 * section of patterns with the tag "pattern-keepNULL". Using the tag
 * "pattern-removeNULL" before a section of patterns will restore the
 * default. For example,
 *
 * pattern-keepNULL
 * pattern TCP B 100 2000 "^\x13bittorrent protocol"
 * pattern-removeNULL
 * pattern TCP a DEFAULT DEFAULT ajprot\x0d\x0a
 *
 * 23. darknet:
 *
 * darknet  net/mask
 *
 * If a local network can be allocated that is empty, attempts to communicate
 * with the target net/mask will be tagged in various ways.  The first such
 * use is to indicate such a network as an additional tag in the tcp syn
 * portreport.  See the portreport module itself for more information.
 * The darknet net/mask offers a 2nd network that will be flagged
 * in addition to the honeynet tag.  Note that the network in question
 * can be a /32 if one host in particular is thought to be of interest.
 * The p application tag is used to flag packets sent to this net.
 * 
 * 24. pot file
 *
 * If topn_syn is defined, and the darknet and/or honeynet are defined
 * as well, any TCP syn tuple instances with either the hardwired P or p
 * flags will be put in their own "potdump.txt" TCP syn tuple file.
 * In order for that to happen the following "pot dump"  spec must be
 * supplied:
 *
 *	syntax: 
 *	topn_syn_potdump  basedirectory
 *
 * 25. blacklist directives:
 *	
 *	syntax:
 *	blist_dumpfile maxpackets directory
 *	blist_include tag blistfile
 *
 * This allows the specification of a blacklist list of IP addresses 
 * and ip ports associated with those hosts.
 *
 * The IP addresses if found in a packet will cause two things to happen.
 * 1. any packets sent to/from these 
 * IP addresses to be recorded in a tcpdump format file named 
 * 
 * blacklist.<timestamp>.dmp
 *
 * in the directory specified to the blist_dumpfile config directive.
 *
 * 2. an elog entry of the form "IP blacklist event <tag> <ip address> will
 * be sent to the event log in the backend.  The tag is any string
 * which may be used to produce a named file so IP addresses from different
 * sources can be matched to different event log messages.
 *
 * The specified port may be either a port number or the string any
 * without quotes as a wildcard for any port.  See the examples below.
 * Note that the port is either a source or destination port (so it's
 * a little fuzzy in that regard).
 *
 * maxpackets tells how many packets max to store.  Ourmon will stop storing
 * packets when this packet count is exhausted.  maxpackets == 0 means 
 * there is no limit on the packet count.  If the packet count is exhausted,
 * ourmon must be restarted to store more packets in a new dump file.
 * The include file must be a simple list of IP addresses, one per line.
 * Any errors in the include file will cause ourmon to halt with an
 * error message.  At runtime a hash list is used to check whether packets
 * should be stored in the dumpfile.  Packets stored are subject to the 
 * command-line snap length specified when ourmon is started.
 *
 * example:
 *
 * blist_dumpfile 1000 /usr/dumps
 * blist_include source1 /etc/blacklist1.txt
 * blist_include source2 /etc/blacklist2.txt
 *
 * The blacklist include directive may be used 1 or more times.
 * The blist_dumpfile directive should only be used once.  Both
 * must be specified.
 *
 * Include file address examples:
 *
 *	10.0.0.1	7777
 *	10.0.0.2	any
 *
 * 26. dns directives:
 *
 *	syntax:
 *	dns topn_number - turn dns stats, top n, blacklist, on
 *	dns 40 - top N dns queries will be 40
 *
 * dns N turns on the DNS facility which has three features:
 *	1. simple dns stats are put out in an RRD file.
 *	2. a top N dns query list is produced that includes
 *		4 record types (A, AAAA, MX, and CNAME).
 *	3. in addition a DNS blacklist may be included to
 *	create event log messages for known evil dns names.
 *
 * Note: currently only IPv4 UDP messages are examined for DNS
 * content.
 *
 *      DNS blacklist syntax:
 *	one or more dns include files:
 *
 *	dns_include /etc/dnslist.txt
 *	dns_include /etc/dnslist.txt
 *
 * syntax of dns include file:
 *
 * DNS blacklist entries should consist of DNS types and DNS type.
 * A, AAAA, CNAME, and MX records are supported.
 *
 *	example:
 *
 *	kazuyo.org A
 *
 * 27. scanmesh
 *
 *	scanmesh tcp_ww basedir (for scanmesh.txt)
 *
 *	Only store local (HOME) IP hosts (as found in hashsyn.c's port report)
 *	with a ww higher than tcp_ww.  These are output into scanmesh.txt.
 *	Followed by all sampled remote peers for the next 30 seconds.  
 *	Thus we can learn who is talking to some set of scanners.
 *
 * 28.  topn_ssh
 *
 *	topn_ssh basedir 
 *
 *	dumps variations of flow tuple to ssh.txt in base directory.
 *	This is based on topn_ip flows and cannot be done unless topn_ip
 *	flows is on.  All port tuples mentioning port 22 as src/dst
 *	are dumped.
 */

/*
 * Copyright (c) 2001,2002,2003,2004 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <pcap.h>
#include <errno.h>

#include "ourmon.h"
#include "filter.h"
#include "cprogram.h"
#include "stats.h"
#include "trigger.h"
#include "patmatch.h"

#define MAXLINE	1536

/* global filter variables *************************** */
struct fixed_ipproto fixed_ipproto;
struct fixed_honeynet fixed_honeynet;
struct fixed_honeynet fixed_darknet;
struct fixed_tcp3 fixed_tcp3;
struct fixed_iprange1 fixed_iprange1;
struct fixed_l2proto fixed_l2proto;
struct fixed_cast fixed_cast;
struct fixed_size fixed_size;
struct bpf_spec bpfs[MAXBPFSPECS];
char sysname[MAXSYSNAME];
char pktname[MAXSYSNAME];
int    pktnameflag = 0;
int    dopackets = 1;		/* nopkts turns this off */
struct ourstats os;

struct topn_ip topn_ip;
struct topn_syn topn_syn;
struct topn_icmperror topn_icmperror;
struct topn_port topn_port;

struct topn_sip topn_scans_input;
struct topn_sip topn_scans_output;
struct hashStats topn_scans_ip_stats;
 
struct topn_sip topn_scans_ports_input;
struct topn_sip topn_scans_ports_output;
struct hashStats topn_scans_ports_stats;

struct topn_sip topn_scans_ports_udp_input;
struct topn_sip topn_scans_ports_udp_output;
struct hashStats topn_scans_ports_udp_stats;

struct topn_sip topn_scans_ports_tcp_input;
struct topn_sip topn_scans_ports_tcp_output;
struct hashStats topn_scans_ports_tcp_stats;

extern struct trigger_cap trigger_cap[MAXTRIGGERS];
extern int bpf_trigger_cap[MAXBPFSPECS][MAXBPFS];
extern int trigger_count;

struct topn_irc topn_irc;

int bpfpackets = 0;	/* bpf graph is in bytes mode if 0, packets if 1 */
int removenull=1;       /* for pattern descriptions - NULLs are by 
			 * default removed from payload before matching 
			 */ 
#ifdef SCANMESH
struct scanmesh scanmesh;
#endif

/* no global structure so we remember a few things to
 * prove this is working
 */
int blist_config = FALSE;
int dns_config = FALSE;
unsigned int blist_maxpkts = 1;
void config_blist_dumpfile(char *, unsigned int);
void config_blist(char *, char *);
void config_dnslist(char *);

/*************************************************/

extern int debug;

/*
 * function prototype declarations
 */
static  switchOnOp();
int     resolveAddress(char *addr, struct sockaddr_in * to);
char*   extractString();
void 	parse_netmask(struct bpf_cparams *cps, char *n);
void    parse_icmp(struct bpf_cparams *cps, char *n);
void 	parse_tcprange(struct bpf_cparams *cps, char *n);
void 	parse_udprange(struct bpf_cparams *cps, char *n);
void 	parse_ports(struct bpf_cparams *cps, char *n);
void    parse_ipproto( struct bpf_cparams *cps, char *n);
void    parse_tcpflag( struct bpf_cparams *cps, char *n);

void
configInit()
{
	int i;

	fixed_ipproto.configured = FALSE;
	fixed_honeynet.configured = FALSE;
	fixed_darknet.configured = FALSE;
	fixed_tcp3.configured = FALSE;
	fixed_iprange1.configured = FALSE;
	fixed_cast.configured = FALSE;
	fixed_size.configured = FALSE;
	fixed_l2proto.configured = FALSE;
	topn_ip.configured = FALSE;
	topn_ip.sflow = FALSE;
	topn_ip.flowfile[0] = '\0';
	topn_port.configured = FALSE;
	topn_syn.configured = FALSE;
	topn_syn.wormfile[0] = '\0';
	topn_syn.wflag = FALSE;
	topn_syn.weight = 0;
	topn_syn.ipaflag = FALSE;
	topn_syn.homeips = 0;
	topn_syn.sortbysyn = TRUE;
	topn_syn.wormdiff = 20;
	topn_syn.p2pflag = FALSE;
	topn_syn.syndump_flag = FALSE;
	topn_syn.emailflag = FALSE;
	topn_icmperror.configured = FALSE;
        topn_scans_input.configured = FALSE;
        topn_scans_output.configured = FALSE;
        topn_scans_ports_input.configured = FALSE;
        topn_scans_ports_output.configured = FALSE;
        topn_scans_ports_tcp_input.configured = FALSE;
        topn_scans_ports_tcp_output.configured = FALSE;
        topn_scans_ports_udp_input.configured = FALSE;
        topn_scans_ports_udp_output.configured = FALSE;
	topn_irc.configured = FALSE;
	topn_irc.fileflag = FALSE;

	bzero(sysname, MAXSYSNAME);
	bzero(pktname, MAXSYSNAME);

	bzero(&trigger_cap, sizeof(struct trigger_cap[MAXTRIGGERS]));
        bzero(&bpf_trigger_cap, sizeof(int) * MAXBPFSPECS * MAXBPFS);
	trigger_count = 0;

	for (i = 0; i < MAXBPFS; i++) {
		bpfs[i].configured = FALSE; 	
	}
}

/*
 * read the config file and load its values
 * 
 * return: 0: ok -1: fatal error (although we return)
 */
int
configRead(configName, bpfop, snaplen, netmask, dirset, outdir)
char *configName;	/* name of config file */
int  bpfop;		/* bpf op */
int  snaplen;		/* bpf snaplen */
unsigned int netmask;  /* bpf netmask */
int  dirset;		/* outdir overrides any config path setting */
char *outdir;		/* directory path to use */
{
	FILE           *fd;
	char            linebuf[1024];
	char            opbuf[1024];
	char           *opword;
	char           *result;
	int             lineNumber = 1;
	int 		i, j;
	char 		*ptr;
	int 		ccount;

	fd = fopen(configName, "r");
	if (fd == NULL) {
		fprintf(stderr, "ourmon: open of config file %s failed\n",
		       configName);
		exit(1);
	}

	configInit();

	/*
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		result = fgets(linebuf, MAXLINE, fd);
		if (result == NULL)
			break;

		ccount = strlen(linebuf);
		linebuf[ccount-1] = 0;

		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0)) {
			lineNumber++;
			continue;
		}
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */
		switchOnOp(bpfop, snaplen, netmask, opword, result, lineNumber, dirset, outdir);
		lineNumber++;
	};
	fclose(fd);

	if (debug) {
		configDump();
	}

	/*
	 * semantic checks if any
	 */
	if (topn_irc.configured && !topn_irc.fileflag) {
		fprintf(stderr, "topn_irc and topn_irc_file both need to be defined\n");
		exit(1);
	}
	return (0);

}  /* end configRead() */

static
switchOnOp(bpfop, snaplen, netmask, op, linebuf, line, dirset, outdir)
	int	       bpfop;
	int 	       snaplen;
	unsigned int  netmask;
	char           *op;
	char           *linebuf;
	int             line;
	int		dirset;
	char 		*outdir;
{
	char            dc[MAXLINE];
	char            v1[MAXLINE];
	char            v2[MAXLINE];
	char            v3[MAXLINE];
	char            v4[MAXLINE];
	char            v5[MAXLINE];
	char            v6[MAXLINE];
	int             rc;
	int 		oopsflag = 0;
	struct sockaddr_in sa;
	char*		p;
	extern char *   device_name;
	int             readpos;
	char            inchar;
	unsigned int 	i1;
	extern int 	errno;

	v1[0] = v2[0] = v3[0] = v4[0] = v5[0] = v6[0] = '\0';

	/* mn ifname ip */
	if (strcmp(op, "fixed_ipproto") == 0) {
		fixed_ipproto.configured = TRUE;
	} else if (strcmp(op, "sysname") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		strncpy(sysname, v1, MAXSYSNAME-1);
	} else if (strcmp(op, "pktname") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		strncpy(pktname, v1, MAXSYSNAME-1);
		pktnameflag = 1;
	} else if (strcmp(op, "nopkts") == 0) {
		dopackets = 0;
	} else if (strcmp(op, "fixed_tcp3") == 0) {
		fixed_tcp3.configured = TRUE;
		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		fixed_tcp3.tcp_port1 = atoi(v1);
		fixed_tcp3.tcp_port2 = atoi(v2);
		if (fixed_tcp3.tcp_port1 == 0 || fixed_tcp3.tcp_port2 == 0) {
			fprintf(stderr, "fixed_tcp3 needs two non-zero port arguments\n");
			fprintf(stderr, "syntax: fixed_tcp3 port1 port2\n");
			exit(1);
		}
	} else if (strcmp(op, "fixed_iprange1") == 0) {
		fixed_iprange1.configured = TRUE;

		sscanf(linebuf, "%s %s %s %s %s %s %s", dc, v1, v2, v3, v4, v5, v6);
		/* labels first
		*/
		strncpy(fixed_iprange1.s_ipnet1, v5, 1024);
		strncpy(fixed_iprange1.s_ipnet2, v6, 1024);

		/* now 4 "ip" addresses, two of which should actually
		 * be subnet masks
		 */
		rc = resolveAddress(v1, &sa);

		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid ip address 1 %s\n", v1);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_iprange1.ipaddr1.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v2, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid mask 1 %s\n", v2);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_iprange1.netmask1.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v3, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid ip address 2 %s\n", v1);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_iprange1.ipaddr2.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v4, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_iprange1: invalid netmask 2 %s\n", v1);
			exit(1);
		}
		fixed_iprange1.netmask2.s_addr = sa.sin_addr.s_addr;

		/* in network byte order
		*/

		fixed_iprange1.ipnet1.s_addr = fixed_iprange1.ipaddr1.s_addr
					& fixed_iprange1.netmask1.s_addr;
		fixed_iprange1.ipnet2.s_addr = fixed_iprange1.ipaddr2.s_addr
					& fixed_iprange1.netmask2.s_addr;

	} else if (strcmp(op, "fixed_cast") == 0) {
		fixed_cast.configured = TRUE;

		sscanf(linebuf, "%s %s %s", dc, v1, v2);

		/* now 2 "ip" addresses, ip addr and net mask
		 */
		rc = resolveAddress(v1, &sa);

		if (rc <= 0) {
			fprintf(stderr, "fixed_cast: invalid ip address %s\n", v1);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_cast.ipaddr1.s_addr = sa.sin_addr.s_addr;

		rc = resolveAddress(v2, &sa);
		if (rc <= 0) {
			fprintf(stderr, "fixed_cast: invalid mask %s\n", v2);
			exit(1);
		}
		/* in network byte order
		*/
		fixed_cast.netmask1.s_addr = sa.sin_addr.s_addr;

		/* now compute all 0's and all 1's bcast
		*/

		fixed_cast.ipnet0.s_addr = fixed_cast.ipaddr1.s_addr
					& fixed_cast.netmask1.s_addr;

		fixed_cast.ipnet1.s_addr = fixed_cast.ipnet0.s_addr
					| ~fixed_cast.netmask1.s_addr;

	} else if (strcmp(op, "fixed_size") == 0) {
		fixed_size.configured = TRUE;
	} else if (strcmp(op, "fixed_l2proto") == 0) {
		fixed_l2proto.configured = TRUE;

	/* bpf label bpf-label bpf-expr */
	} else if (strcmp(op, "bpf") == 0) {
		p = extractString(linebuf, v1, 1024);	/* graph label */
		p = extractString(p, v2, 1024);		/* line label */
		p = extractString(p, v3, 1024);		/* bpf expression */
		bpf_config(FALSE, bpfop, snaplen, netmask, FALSE, v1, v2, v3, NULL);
	} else if (strcmp(op, "bpf-packets") == 0) {
		bpfpackets = 1;		/* now count packets */
	} else if (strcmp(op, "bpf-bytes") == 0) {
		bpfpackets = 0;		/* now count bytes */
	/* bpf-next bpf-label bpf-expr */
	} else if (strcmp(op, "bpf-next") == 0) {
		p = extractString(linebuf, v1, 1024);
		p = extractString(p, v2, 1024);
		bpf_config(FALSE, bpfop, snaplen, netmask, TRUE, NULL, v1, v2, NULL);
	/* this is a variation on bpf-next internally
	*/
	} else if (strcmp(op, "bpf-noxtra") == 0) {
		bpf_config(TRUE, bpfop, snaplen, netmask, TRUE, NULL, NULL, NULL, NULL);
/*
 *      cbpf "tag" "label" "per-filter-label" "filter"
 */
	} else if (strcmp(op, "cbpf") == 0) {
		p = extractString(linebuf, v1, 1024);	/* tag */
		p = extractString(p, v2, 1024);		/* graph label */
		p = extractString(p, v3, 1024);		/* line label */
		p = extractString(p, v4, 1024);		/* expression */
		bpf_config(FALSE, bpfop, snaplen, netmask, FALSE, v2, v3, v4, v1);
/*
 *      cbpf "tag" "per-filter-label" "filter"
 */
	} else if (strcmp(op, "cbpf-next") == 0) {
		p = extractString(linebuf, v1, 1024);
		p = extractString(p, v2, 1024);
		p = extractString(p, v3, 1024);
		bpf_config(FALSE, bpfop, snaplen, netmask, TRUE, NULL, v2, v3, v1);
	} else if (strcmp(op, "topn_ip") == 0) {
		topn_ip.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		topn_ip.actualn = atoi(v1);
		if (topn_ip.actualn > MAXTOPN) {
			fprintf(stderr, "topn_ip topN limit is %d\n",
				MAXTOPN);
			exit(1);
		}
		/* initialize the list mechanism
		*/
		initList(&topn_ip.iplist);

/*
 *	topn_smallflow  basedirectory + smallflow.txt
 */
	} else if (strcmp(op, "topn_smallflow") == 0) {
		char *fn = "smallflow.txt";
		int l = strlen(fn);
		
		if (topn_ip.configured == FALSE) {
			fprintf(stderr, "topn_smallflow: topn_ip needs to turned on\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_smallflow path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_ip.flowfile, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_ip.flowfile, "%s/%s", v1, fn);
		}
		topn_ip.sflow = TRUE;
/*
 *	topn_ssh  basedirectory + ssh.txt
 */
	} else if (strcmp(op, "topn_ssh") == 0) {
		char *fn = "ssh.txt";
		int l = strlen(fn);
		
		if (topn_ip.configured == FALSE) {
			fprintf(stderr, "topn_ssh: topn_ip needs to turned on\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_ssh path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_ip.sshfile, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_ip.sshfile, "%s/%s", v1, fn);
		}
		topn_ip.ssh = TRUE;

	/* trigger_topn_ip count dirname for dump
	*/
        } else if (strcmp(op, "trigger_topn_ip") == 0) {
                if(topn_ip.configured == FALSE) {
                        fprintf(stderr, "trigger_topn_ip need top_ip configured\n");
                        exit(1);
                }
                sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);
 
		/* flow count per second
		*/
                if( trigger_config(TOPNIP, v1, v2, v3, 30) <  0 ) {
                        fprintf(stderr,"trigger_topn_ip configuration failed\n");
                        exit(1);
                }

                /* parse the bpf expression.  The default expression
		 * is "ip".   which doesn't matter but oh well.
		*/
                if ( trigger_bpf_config(TOPNIP, bpfop, snaplen, netmask, NULL) < 0 ) {
                        fprintf(stderr,"trigger_topn_ip bpf parse failed\n");
                        exit(1);
                }
	/*
	 *      trigger_worm trigger_threshold packet_count tworm_dump_file
	 */
        } else if(strcmp(op, "trigger_worm") == 0) {
                if(topn_syn.configured == FALSE){
                        fprintf(stderr,"trigger_worm: topn_syn needs to turned on\n");
                        exit(1);
                }
                if (topn_syn.wflag == FALSE) {
                        fprintf(stderr,"trigger_worm: tcp worm file needs to be set\n");
                        exit(1);
                }
                sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);

		/* ip count per period
		*/
                if( trigger_config(TWORM, v1, v2, v3, 0) < 0 ) {
                         fprintf(stderr,"trigger_worm: configure failed\n");
                         exit(1);
                }

                if( trigger_bpf_config(TWORM, bpfop, snaplen, netmask, NULL) < 0 ) {
                        fprintf(stderr,"trigger_worm: bpf parsing failed\n");
                        exit(1);
                }
	/*
	 * bpf_trigger bpf_graph_name bpf_line_label threshold packet_count directory
        */
        } else if (strcmp(op, "bpf_trigger") == 0) {
		int trigger_type;

                p = extractString(linebuf, v1, 1024);
                p = extractString(p, v2,1024);
                sscanf(p, "%s %s %s", v3, v4, v5);

                /* allocate a trigger capability
		 * this is simply an integer index into the trigger
		 * capability array.  the function itself complains.
		 */
                trigger_type = bpf_allo_trigger_cap();
		if (trigger_type == -1) {
			fprintf(stderr,"fatal: out of trigger_capabilities\n");
			exit(1);
		}
                /* configure the trigger cap 
		 * 	with threshold, target packet_count, output directory
		*/
                if (trigger_config(trigger_type, v3, v4, v5, 0) < 0) {
			exit(1);
		}
  
                /* complete the trigger cap  of bpf type  
                 *      with its name and save the index of
                 *      bpf_trigger in trigger_cap to the table
                 */
                if ( bpf_trigger_completion(trigger_type, v1, v2) < 0 ) {
                        fprintf(stderr,"trigger_topn_ip bpf parse failed\n");
                        exit(1);
                }
	/*
 	 * topn_scans_trigger threshold packet_count directory
	 */
        } else if (strcmp(op, "topn_scans_trigger") == 0) {
                if(topn_scans_input.configured == FALSE) {
                        fprintf(stderr,"topn_scans_trigger need topn_scans configured\n");
                        exit(1);
                }
                sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);
		/* this is ips per period
		*/
                if(trigger_config(TOPNSCANS, v1, v2, v3, 0) < 0 ) {
                        fprintf(stderr,"topn_scans_trigger configuration failed\n");
                        exit(1);
                }

                if ( trigger_bpf_config(TOPNSCANS, bpfop, snaplen, netmask, NULL) < 0 ) {
                        fprintf(stderr,"topn_scans_trigger bpf parse failed\n");
                        exit(1);
                }
 	
	/*
 	 * drop_trigger threshold packet_count directory
	 */

	} else if (strcmp(op, "drop_trigger") == 0 ) {
		
                sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);

		/* note: drops/sec in back-end RRD graph
		*/
                if( trigger_config(DROP, v1, v2, v3, 30) < 0 ) {
                         fprintf(stderr,"drop_trigger: configuration failed\n");
                         exit(1);
                }

                if( trigger_bpf_config(DROP, bpfop, snaplen, netmask, NULL) < 0 ) {
                        fprintf(stderr,"drop_trigger: bpf parse failed\n");
                        exit(1);
                }
	/*
 	 * udperror_trigger threshold packet_count directory
	 */

	} else if (strcmp(op, "udperror_trigger") == 0 ) {
		if ( topn_icmperror.configured == FALSE) {
			fprintf(stderr, "udperror_trigger needs topn_icmperror configured. \n");
			exit(1); 		
		}
		sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);
		
		if( trigger_config(TOPNUDPERR, v1, v2, v3, 0) < 0 ) {
                        fprintf(stderr,"udperror_trigger configuration failed\n");
                        exit(1);
                }
                if ( trigger_bpf_config(TOPNUDPERR, bpfop, snaplen, netmask, NULL) < 0 ) {
                        fprintf(stderr,"udperror_trigger bpf parse failed\n");
                        exit(1);
                }

	} else if (strcmp(op, "topn_syn") == 0) {
		topn_syn.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		topn_syn.actualn = atoi(v1);
		if (topn_syn.actualn > MAXTOPN) {
			fprintf(stderr, "topn_syn topN limit is %d\n",
				MAXTOPN);
			exit(1);
		}
		/* initialize the list mechanism
		*/
		init_synlist(&topn_syn.synlist);
/*
 *	topn_syn_wormfile  basedirectory + tcpworm.txt
 */
	} else if (strcmp(op, "topn_syn_wormfile") == 0) {
		char *fn = "tcpworm.txt";
		int l = strlen(fn);
		
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_wormfile: topn_syn needs to turned on\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_syn path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_syn.wormfile, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_syn.wormfile, "%s/%s", v1, fn);
		}
		topn_syn.wflag = TRUE;
/*
 *	topn_syn_p2pfile  basedirectory + p2pfile.txt
 */
	} else if (strcmp(op, "topn_syn_p2pfile") == 0) {
		char *fn = "p2pfile.txt";
		int l = strlen(fn);
		
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_p2pfile: topn_syn needs to turned on\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_syn p2p directory path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_syn.p2pfile, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_syn.p2pfile, "%s/%s", v1, fn);
		}
		topn_syn.p2pflag = TRUE;

/*
 *	topn_syn_syndump  basedirectory + syndump.txt
 */
	} else if (strcmp(op, "topn_syn_syndump") == 0) {
		char *fn = "syndump.txt";
		int l = strlen(fn);
		
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_syndump: topn_syn needs to turned on first\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_syn syndump directory path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_syn.syndump, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_syn.syndump, "%s/%s", v1, fn);
		}
		topn_syn.syndump_flag = TRUE;

	} else if (strcmp(op, "topn_syn_emaildump") == 0) {
		char *fn = "emaildump.txt";
		int l = strlen(fn);
		
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_emaildump: topn_syn needs to turned on first\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_syn emaildump path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_syn.emaildump, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_syn.emaildump, "%s/%s", v1, fn);
		}
		topn_syn.emailflag = TRUE;

	/* potdump
	*/
	} else if (strcmp(op, "topn_syn_potdump") == 0) {
		char *fn = "potdump.txt";
		int l = strlen(fn);
		
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_potdump: topn_syn needs to turned on first\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_syn potdump path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_syn.potdump, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_syn.potdump, "%s/%s", v1, fn);
		}
		topn_syn.potflag = TRUE;

	/* blist_dumpfile maxpackets dumpdir
	*/
	} else if (strcmp(op, "blist_dumpfile") == 0) {
		
		if (blist_config == TRUE) {
			fprintf(stderr, "only one blist_dumpfile directive allowed\n");
			exit(1);
		}
		blist_config = TRUE;
		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		int l = strlen(v2);
		
		if (strlen(v2) > (1022-l)) {
			fprintf(stderr, "blist_dumpfile directory path too long\n");
			exit(1);
		}

		errno = 0;

		/* note strtol is limited to a long if decimal
		*/
		blist_maxpkts = strtol(v1, NULL, 0);  /* this allows hex to be specified */
		if (errno == EINVAL) {
			fprintf(stderr, "blist_dumpfile: invalid packetsize: %s\n", v1);
			exit(1);
		}
		config_blist_dumpfile(v2, blist_maxpkts);

	/* blist_include tag filename
	*/
	} else if (strcmp(op, "blist_include") == 0) {

		if (blist_config == FALSE) {
			fprintf(stderr, "sorry, need blist_dumpfile directive first\n");
			exit(1);
		}
		rc = sscanf(linebuf, "%s %s %s", dc, v1, v2);
		if (rc != 3) {
			fprintf(stderr, "blist_include: invalid syntax: blist_include \"tag\" file\n");
			exit(1);
		}
		p = extractString(v1, v4, 1024);	/* tag limited to 1k bytes */
		int l = strlen(v2);
		if (strlen(v2) > (1022-l)) {
			fprintf(stderr, "blist_include file path too long\n");
			exit(1);
		}
		config_blist(v4, v2);
	/* dns
	 * turn dns on AND request mon.lite metrics
	*/
	} else if (strcmp(op, "dns") == 0) {
		int count;
		
		rc = sscanf(linebuf, "%s %s", dc, v1);
		if (rc != 2) {
			fprintf(stderr, "dns topn_value (1..100)\n");
			exit(1);
		}
		dns_config = TRUE;
		enable_dns(atoi(v1));

	/* dns_include filename
	*/
	} else if (strcmp(op, "dns_include") == 0) {
		
		if (dns_config == FALSE) {
			fprintf(stderr, "sorry, need dns_config directive first\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		int l = strlen(v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "dns_include file path too long\n");
			exit(1);
		}
		config_dnslist(v1);

	/*
	 * tworm_weight 0..100
	 */
	} else if (strcmp(op, "tworm_weight") == 0) {
		
		if (topn_syn.wflag == FALSE) {
			fprintf(stderr, "tworm_weight: topn_syn.wormfile needs to be set\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		topn_syn.weight = atoi(v1);
		if ((topn_syn.weight < 0) || (topn_syn.weight > 100)) {
			fprintf(stderr, "tworm_weight value must be between 0 and 100 inclusive\n");
			exit(1);
		}
 
	} else if (strcmp(op, "topn_syn_wormdiff") == 0) {
		
		if (topn_syn.wflag == FALSE) {
			fprintf(stderr, "tworm_weight: topn_syn.wormfile needs to be set\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		topn_syn.wormdiff = atoi(v1);
		if ((topn_syn.wormdiff < 0)) {
			fprintf(stderr, "topn_syn.wormdiff must be non-negative.\n");
			exit(1);
		}

	} else if (strcmp(op, "topn_syn_sort") == 0) {
		
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_sort: topn_syn needs to be set before this switch is set.\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		topn_syn.sortbysyn = atoi(v1);
		if ((topn_syn.sortbysyn < 0) || (topn_syn.sortbysyn > 1)) {
			fprintf(stderr, "topn_syn_sort value must be 0 or 1 inclusive.\n");
			exit(1);
		}

/*
 *	topn_syn_homeip	  ipaddress/netmask
 */
	} else if (strcmp(op, "topn_syn_homeip") == 0) {
		struct bpf_cparams home_cps;

		if (topn_syn.homeips >= MAXHOMEIPS) {
			fprintf(stderr, "too many home ips: modify MAXHOMEIPS in filter.h\n");
			exit(1);
		}
		if (topn_syn.configured == FALSE) {
			fprintf(stderr, "topn_syn_homeip: topn_syn needs to turned on\n");
			exit(1);
		}
		if (topn_syn.wflag == FALSE) {
			fprintf(stderr, "topn_syn_homeip: topn_syn_wormfile needs to be set first\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		/* store here because parse_netmask is destructive
		*/
		strncpy(topn_syn.ipmask[topn_syn.homeips], v1, 511);
		parse_netmask(&home_cps, v1);
		topn_syn.ipaddr[topn_syn.homeips].s_addr = home_cps.net.s_addr;
		topn_syn.netmask[topn_syn.homeips].s_addr = home_cps.netmask.s_addr;
		topn_syn.ipaflag = TRUE;
		topn_syn.homeips++;

	/*
	 *	honeynet net/mask
	 */
	} else if (strcmp(op, "honeynet") == 0) {
		struct bpf_cparams honey_cps;

		fixed_honeynet.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		parse_netmask(&honey_cps, v1);
		fixed_honeynet.netaddr.s_addr = honey_cps.net.s_addr;
		fixed_honeynet.netmask.s_addr = honey_cps.netmask.s_addr;
	/*
	 *	darknet net/mask
	 */
	} else if (strcmp(op, "darknet") == 0) {
		struct bpf_cparams honey_cps;

		fixed_darknet.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		parse_netmask(&honey_cps, v1);
		fixed_darknet.netaddr.s_addr = honey_cps.net.s_addr;
		fixed_darknet.netmask.s_addr = honey_cps.netmask.s_addr;

	/* maxcount if 0, means all
	*/
	} else if (strcmp(op, "topn_irc") == 0) {
		topn_irc.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		topn_irc.maxcount = atoi(v1);

		/* initialize the list mechanism
		*/
		initIrc(&topn_irc.irclist);
	/*
	 *	topn_irc_file  basedirectory + irc.txt
	 */
	} else if (strcmp(op, "topn_irc_file") == 0) {
		char *fn = "irc.txt";
		int l = strlen(fn);
		
		if (topn_irc.configured == FALSE) {
			fprintf(stderr, "topn_syn_ircfile: topn_irc needs to turned on first\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		
		if (strlen(v1) > (1022-l)) {
			fprintf(stderr, "topn_irc directory path too long\n");
			exit(1);
		}
		if (dirset) {
			sprintf(topn_irc.ircfile, "%s/%s", outdir, fn);
		}
		else {
			sprintf(topn_irc.ircfile, "%s/%s", v1, fn);
		}
		topn_irc.configured = TRUE;
		topn_irc.fileflag = TRUE;

	} else if (strcmp(op, "topn_icmperror") == 0) {
		topn_icmperror.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		topn_icmperror.actualn = atoi(v1);
		if (topn_icmperror.actualn > MAXTOPN) {
			fprintf(stderr, "topn_icmperror topN limit is %d\n",
				MAXTOPN);
			exit(1);
		}
		/* initialize the port list mechanism
		*/
		init_icmplist(&topn_icmperror.icmplist);


	} else if (strcmp(op, "topn_port") == 0) {
		topn_port.configured = TRUE;
		sscanf(linebuf, "%s %s", dc, v1);
		topn_port.actualn = atoi(v1);
		if (topn_port.actualn > MAXTOPN) {
			fprintf(stderr, "topn_port topN limit is %d\n",
				MAXTOPN);
			exit(1);
		}
		/* initialize the port list mechanism
		*/
		init_plist(&topn_port.tcplist);
		init_plist(&topn_port.udplist);

	} else if (strcmp(op, "topn_scans") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		topn_scans_input.actualn = atoi(v1);
		topn_scans_output.actualn = topn_scans_input.actualn;
		topn_scans_input.configured = 1;
		topn_scans_output.configured = 1;
		if (topn_scans_input.actualn > MAXTOPN) {
			fprintf(stderr, "topn_scans topN limit is %d\n", 
				MAXTOPN);
			exit(1);
               	}
		/* initialize the list mechanism
		*/
		initScanList(&topn_scans_input.iplist);
		initScanList(&topn_scans_output.iplist);

	} else if (strcmp(op, "topn_port_scans") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		topn_scans_ports_input.actualn = atoi(v1);
		
		topn_scans_ports_output.actualn = 
			topn_scans_ports_input.actualn;
		topn_scans_ports_input.configured = 1;
		topn_scans_ports_output.configured = 1;

		topn_scans_ports_tcp_output.actualn = 
			topn_scans_ports_tcp_input.actualn =
			topn_scans_ports_input.actualn;
		topn_scans_ports_input.configured = 1;
		topn_scans_ports_output.configured = 1;
		
		topn_scans_ports_udp_output.actualn = 
			topn_scans_ports_udp_input.actualn =
			topn_scans_ports_input.actualn;
		
		topn_scans_ports_input.configured = 1;
		topn_scans_ports_output.configured = 1;

		topn_scans_ports_tcp_input.configured = 1;
		topn_scans_ports_udp_input.configured = 1;
		
		topn_scans_ports_udp_output.configured = 1;
		topn_scans_ports_tcp_output.configured = 1;
		
		if (topn_scans_ports_input.actualn > MAXTOPN) {
			fprintf(stderr, "topn_port_scans topN limit is %d\n", 
				MAXTOPN);
			exit(1);
               	}

		/* initialize the list mechanism
		*/
		initScanList(&topn_scans_ports_input.iplist);
		initScanList(&topn_scans_ports_output.iplist);

		initScanList(&topn_scans_ports_tcp_input.iplist);
		initScanList(&topn_scans_ports_tcp_output.iplist);
	
		initScanList(&topn_scans_ports_udp_input.iplist);
		initScanList(&topn_scans_ports_udp_output.iplist);
	
	} else if (strcmp(op, "pattern-removeNULL") == 0) {
		removenull=1; /*remove NULLs from payload before matching*/
	} else if (strcmp(op, "pattern-keepNULL") == 0) {
		removenull=0; /*no changes to payload before matching*/
	} else if (strcmp(op, "pattern") == 0) {

		/* user supplied regular expression using patmatch library
		*/
		sscanf(linebuf, "%s %s %s %s %s %n", dc, v1, v2, v3, v4, &readpos); 

		/* advance pointer into buffer to cover all that has been read so far
		 * just the regular expression should remain. Only accept regular
		 * expression up to maximum size
		 */

		linebuf+=readpos;

		/* only accept one character tags
		*/
		inchar=v2[0];

		/* check for hardwired tags.  error to try and reuse.
		 */
		switch (inchar) {
			/* TCP bittorrent */
			case 'B': 
				oopsflag = 1;
				break;
			/* TCP gnutella */
			case 'G': 
				oopsflag = 1;
				break;
			/* TCP kazaa */
			case 'K': 
				oopsflag = 1;
				break;
			/* TCP edonkey */
			case 'e': 
				oopsflag = 1;
				break;
			/* TCP morpheus */
			case 'M': 
				oopsflag = 1;
				break;
			/* TCP saw web port */
			case 'H': 
				oopsflag = 1;
				break;
			/* TCP saw email port */
			case 'E': 
				oopsflag = 1;
				break;
			/* UDP spim port */
			case 's': 
				oopsflag = 1;
				break;
		}
		if (oopsflag) {
			fprintf(stderr,"pattern is attempting to use reserved hardwired tag: (%c) on line [%d]\n",
				inchar,line);
			exit(1);
		}
		/*now add the new regular expression to list of all*/
		if (patmatch_addPattern(v1, inchar, v3, v4, removenull, linebuf, &patterns)==1) {
			fprintf(stderr,"could not load pattern from configuration file. tag: (%c) on line [%d]\n",inchar,line);
			patmatch_printPatterns(patterns);
			exit(1);
		}
#ifdef SCANMESH
        /* scanmesh tcp_ww basedir */
	/* TBD - need way to specify filepath without -D
	*/
        } else if (strcmp(op, "scanmesh") == 0) {
                char *fn = "scanmesh.txt";
                scanmesh.configured = TRUE;

		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		int l = strlen(v2);
		
		if (strlen(v2) > (1022-l)) {
			fprintf(stderr, "scanmesh directory path too long\n");
			exit(1);
		}

		errno = 0;

		/* note strtol is limited to a long if decimal
		*/
		scanmesh.ipmeshlist.tcp_ww = strtol(v1, NULL, 0);  /* this allows hex to be specified */
		if (errno == EINVAL || scanmesh.ipmeshlist.tcp_ww < 0 || 
			scanmesh.ipmeshlist.tcp_ww > 100) {
			fprintf(stderr, "scanmesh: invalid workweight 0..100: %s\n", v1);
			exit(1);
		}

		if (dirset) {
			sprintf(scanmesh.ipmeshfile, "%s/%s", outdir, fn);
		}
		else {
			sprintf(scanmesh.ipmeshfile, "%s/%s", v2, fn);
		}
 
                init_scanmesh(&scanmesh.ipmeshlist);
#endif
	} else {
		fprintf(stderr, "switchOnOp(): line [%d], unknown config op (%s) in config file\n", line, op);
		exit(1);
	}
}

/*
 * compile bpf expression into fcode bpf byte array
 *
 * note: errexit: exits at config time, does not exit at runtime
 *
 * return:
 *	< 0: compile error.
 *	otherwise: ok.
 */
int
local_pcap_compile_nopcap(
		int snaplen_arg, 
		int linktype_arg, 
		struct bpf_program *program,
                char *buf, 
		int optimize, 
		unsigned int mask, int errexit)
{
        pcap_t *p;
        int ret;
                        
        p = pcap_open_dead(linktype_arg, snaplen_arg);
        if (p == NULL)
                return (-1);
        ret = pcap_compile(p, program, buf, optimize, mask);
	if (ret < 0 && errexit) {
		fprintf(stderr, "expr: (%s): pcap syntax error: %s\n",
			buf,
			pcap_geterr(p));
		exit(1);
	}
        pcap_close(p);
        return (ret);
}

static int last_bpf_index = 0;
/* global so runtime analysis routines
 * can learn if there is any point in
 * calling bpf_filter
 */
int any_bpf = 0;

/* struct bpf_spec bpfs[MAXBPFSPECS]; */

/*
 * compile a bpf/rrdtool graph
 *
 * if ctag is set, we use a hardwired C program,
 * not a bpf spec.
 */

bpf_config(
	int xtraflag, 
	int bpfop, 		/* bpf param only: needed for bpf link-layer */
	int snaplen, 		/* bpf param only: snap length */
	unsigned int netmask,  /* bpf param only: interface netmask */
	int ntag,		/* "next tag": if FALSE, then bpf, else bpf_next */
	char *label, 		/* graph label */
	char* blabel, 		/* line or per bpf label */
	char* bpfexpr,		/* bpf expression or c program expression */
	char *ctag 		/* c program tag */
)
{
	int i;
	struct bpf_spec *b;
	int gotone = 0;		/* gotone is the fcode index */

	if (debug) {
		printf("ntag %d, label %s bl %s bpf %s\n", ntag,
			label, blabel, bpfexpr);
	}
	/* if new bpf graph
	*/
	if (!ntag) {
		/* check for bpf spec array exhaustion
		*/
		if (last_bpf_index >= MAXBPFSPECS) {
			fprintf(stderr, "out of bpf specs, max %d\n",
				MAXBPFSPECS);
			exit(1);
		}
		/* using the label, search for a free bpf structure 
	 	 * the label must be unique
		*/
		b = &bpfs[0];
		for ( i = 0; i < MAXBPFSPECS; i++) {
			if (b->configured == FALSE) {
				break;
			}
			if (strcmp(label, b->prog_label[0]) == 0) {
				fprintf(stderr, 
					"ourmon: bpf label %s already used\n",
		       			label);
				exit(1);
			}
			b++;
		}
		/* we have a bpf spec structure
		*/
		/* allocate it and store labels
		*/
		last_bpf_index = i;	/* next allocation starts from here */
		any_bpf = TRUE;
		b->configured = TRUE;
		strncpy(b->label, label, MAXLABEL-1);
		strncpy(b->prog_label[0], blabel, MAXBPFLABEL-1);
		b->prog_count = 1;
		b->bpfpackets = bpfpackets;  /* set mode bytes/packet cnt */
	}
	/* deal with concatenation case; that is, we must have
	 * a previous bpf_spec with an available bpf fcode array slot in it
	 */
	else {
		if (any_bpf == FALSE) {
			fprintf(stderr, "%s must be used after a previous bpf tag\n", bpfop);
			exit(1);
		}
		b = &bpfs[last_bpf_index];

		/* if xtra flag is set, we do not compute
		 * the xtra (remainder) bytes
		 */
		if (xtraflag) {
			b->noxtra = TRUE;
			return;
		}
		/* it cannot be the 1st one.  we check the fcode
		 * to see if it is set.
		*/
		for ( i = 1; i < MAXBPFS; i++) {
			if (b->progkind[i] == BPFFREE) {
				gotone = i;
				break;
			}
		}
		/* gotone is the fcode index
		*/
		if (gotone == 0) {
			fprintf(stderr, "all bpf slots in bpf spec used up, max is %d\n", MAXBPFS);
			exit(1);
		}
		/* ok we have a free fcode slot
		*/
		strncpy(b->prog_label[gotone], blabel, MAXBPFLABEL-1);
		b->prog_count++;	/* count of fcodes in use */
	}

	if (ctag) {
		if (strcmp(ctag, "net") == 0) {
			(void) parse_netmask(&b->cps[gotone], bpfexpr);
			b->c_program[gotone] = c_network;
			if (debug)
				printf("c_network addr %x gotone %d c_program %x\n", c_network, gotone, b->c_program[gotone]);
		}
		else if (strcmp(ctag, "ping") == 0) {
			/* no expression string */
			b->c_program[gotone] = c_ping;
			if (debug)
				printf("c_ping addr %x gotone %d c_program %x\n", c_network, gotone, b->c_program[gotone]);
		}
		else if (strcmp(ctag, "icmp") == 0) {
			(void) parse_icmp(&b->cps[gotone], bpfexpr);
			b->c_program[gotone] = c_icmp;
		}
		else if (strcmp(ctag, "tcprange") == 0) {
			(void) parse_tcprange(&b->cps[gotone], bpfexpr);
			b->c_program[gotone] = c_tcprange;
		}
		else if (strcmp(ctag, "udprange") == 0) {
			(void) parse_udprange(&b->cps[gotone], bpfexpr);
			b->c_program[gotone] = c_udprange;
		}
		else if (strcmp(ctag, "tcpflag") == 0) {
			(void) parse_tcpflag(&b->cps[gotone], bpfexpr);
			b->c_program[gotone] = c_tcpflag;
		}
		else if (strcmp(ctag, "ports") == 0) {
			(void) parse_ports(&b->cps[gotone], bpfexpr);
			if (debug)
				printf("c_ports (%s) gotone %d c_program %x\n", bpfexpr, gotone, b->c_program[gotone]);
			b->c_program[gotone] = c_ports;
		}
		else if (strcmp(ctag, "ipproto") == 0) {
			(void) parse_ipproto(&b->cps[gotone], bpfexpr);
			b->c_program[gotone] = c_ipproto;
		}
		else {
			fprintf(stderr, "unknown or missing ctag\n");
			exit(1);
			
		}
		b->progkind[gotone] = BPFCPROG;
		return;
	}
	/* compile the bpf expression
	 * note: if fatal error, this routine does not
	 * return.
	 */
	(void) local_pcap_compile_nopcap(snaplen, bpfop,
	    &b->fcodes[gotone],  bpfexpr, 0, netmask, 1); 

	b->progkind[gotone] = BPFBPF;

	if (debug) {
		printf("fcodes[%d] len: %d\n", 
			gotone, b->fcodes[gotone].bf_len);
	}
}

/*
 * debug - dump configuration variables
 *
 * BUGGY: write code to dump out config-time trigger info here.
 */
configDump()
{
	int             i;
	struct bpf_spec *b;
	int j;

	printf("---------config dump-------------------\n");
	if (sysname[0]) {
          	printf("sysname: [%s]\n", sysname);
	}
	if (pktname[0]) {
          	printf("pktname: [%s]\n", pktname);
	}
       	printf("pkts: %d\n", dopackets);

	if (fixed_ipproto.configured) {
          	printf("fixed_ipproto.configured %d\n", fixed_ipproto.configured);
	}
	if (fixed_tcp3.configured) {
		printf("fixed_tcp3: port1 %d\n", fixed_tcp3.tcp_port1);
		printf("fixed_tcp3: port2 %d\n", fixed_tcp3.tcp_port2);
	}
	if (fixed_honeynet.configured) {
		printf("honeynet: netaddr %s, ", 
			inet_ntoa((struct in_addr)fixed_honeynet.netaddr));
		printf("honeynet: netmask %s, ", 
			inet_ntoa((struct in_addr)fixed_honeynet.netmask));
	}
	if (fixed_darknet.configured) {
		printf("darknet: netaddr %s, ", 
			inet_ntoa((struct in_addr)fixed_darknet.netaddr));
		printf("darknet: netmask %s, ", 
			inet_ntoa((struct in_addr)fixed_darknet.netmask));
	}
	if (fixed_iprange1.configured) {
		printf("iprange1: ipaddr1 %s, ", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipaddr1));
		printf("netmask1 %s\n", 
			inet_ntoa((struct in_addr)fixed_iprange1.netmask1) 
			);
		printf("ipaddr2 %s, ", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipaddr2));
		printf("netmask2 %s\n", 
			inet_ntoa((struct in_addr)fixed_iprange1.netmask2) 
			);
		printf("ipnet1 %s, ", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipnet1));
		printf("ipnet2 %s\n", 
			inet_ntoa((struct in_addr)fixed_iprange1.ipnet2));
		printf("labels %s %s\n",
			fixed_iprange1.s_ipnet1,
			fixed_iprange1.s_ipnet2);
	}

	if (topn_ip.configured) {
		printf("topn_ip: actualn %d\n", topn_ip.actualn);
		if (topn_ip.ssh) {
			printf("topn_ssh: file <%s>\n", topn_ip.sshfile);
		}
	}
	
 	if (topn_syn.configured) {

                printf("topn_syn: actualn %d\n", topn_syn.actualn);

		if (topn_syn.wflag) {
			printf("topn_syn: wormfile <%s>\n", topn_syn.wormfile);
			printf("topn_syn_wormdiff %d\n", topn_syn.wormdiff);
		}

		if (topn_syn.p2pflag) {
			printf("topn_syn: p2pfile <%s>\n", topn_syn.p2pfile);
		}

		if (topn_syn.syndump_flag) {
			printf("topn_syn: syndump <%s>\n", topn_syn.syndump);
		}

		if (topn_syn.emailflag) {
			printf("topn_syn: emaildump <%s>\n", topn_syn.emaildump);
		}

		if (topn_syn.ipaflag) {
			for (i = 0; i < topn_syn.homeips; i++) {
				printf("topn_syn: ipaddr %s \n", 
					inet_ntoa((struct in_addr)topn_syn.ipaddr[i]));
				printf("topn_syn: mask %s \n", 
					inet_ntoa((struct in_addr)topn_syn.netmask[i]));
			}
		}

		if (topn_syn.weight) {
			printf("topn_syn: weight %d\n", topn_syn.weight);
		}

                for ( j = 0; j < MAXTRIGGERS; j++) {
                        if (trigger_cap[j].configured) {
                                printf("trigger_cap value %d\n", trigger_cap[j].N_value);
				printf("trigger_cap count %d\n", trigger_cap[j].count);
				printf("trigger_cap filename %s\n", trigger_cap[j].base_filename);
                        }
                }

		printf("topn_syn_sort value %d\n", topn_syn.sortbysyn);
        }

 	if (topn_icmperror.configured) {
                printf("topn_icmperror: actualn %d\n", 
			topn_icmperror.actualn);
        }

	if (topn_port.configured) {
		printf("topn_port: actualn %d\n", topn_port.actualn);
	}

        if (topn_scans_input.configured) {   
                printf("topn_scans: actualn %d\n",
                       topn_scans_input.actualn);
        }

        if (topn_scans_ports_input.configured) {
                printf("topn_port_scans: actualn %d\n",
                       topn_scans_ports_input.actualn);
 
                printf("topn_port_scans_udp: actualn %d\n",
                       topn_scans_ports_udp_input.actualn);
 
                printf("topn_port_scans_tcp: actualn %d\n",
                       topn_scans_ports_tcp_input.actualn);
        }

	if (fixed_cast.configured) {
		printf("fixed_cast: bcast 1's %s \n", 
			inet_ntoa((struct in_addr)fixed_cast.ipnet1));
		printf("fixed_cast: bcast 0's %s \n", 
			inet_ntoa((struct in_addr)fixed_cast.ipnet0));
	}

	if (fixed_size.configured) {
		printf("fixed_size: ON \n");
	}
	if (fixed_l2proto.configured) {
		printf("fixed_l2proto: ON \n");
	}

	/* bpf program dump */

	if (any_bpf) {
		b = &bpfs[0];
		for ( i = 0; i < MAXBPFSPECS; i++, b++) {
			if (b->configured == FALSE) {
				break;
			}
			printf("bpf_spec[%d] label %s\n", i, b->label);
			printf("bpf_spec[%d] programs %d\n", i, b->prog_count);
			printf("bpf_spec[%d] NO xtraflag %d\n", i, b->noxtra);
			printf("bpf_spec[%d] bpfpacket flag %d\n", i, b->bpfpackets);
			for ( j = 0; j < MAXBPFS; j++) {
				printf("bpf_spec[%d] fcode[%d] label %s\n",
					i, j, b->prog_label[j]);
				printf("bpf program kind %d\n",
					b->progkind[j]);
				if (b->progkind[j] == BPFCPROG) {
					if (b->cps[j].net.s_addr) {
						printf("net: %s:\n",
						inet_ntoa((struct in_addr)b->cps[j].net));
						printf("mask: %s:\n",
						inet_ntoa((struct in_addr)b->cps[j].netmask));
					}
					if (b->cps[j].icmpflag) {
						printf("icmp: %d %d\n", b->cps[j].icmp_major, 
							b->cps[j].icmp_minor);
					}
					if (b->cps[j].tportcount) {
						int k;
						printf("cps[j].tportcount %d\n", b->cps[j].tportcount);
						for (k = 0; k < b->cps[j].tportcount; k++) {
							printf("cbpf: tport %d\n", b->cps[j].tports[k]);
						}
					}
					if (b->cps[j].uportcount) {
						int k;
						printf("cps[j].uportcount %d\n", b->cps[j].uportcount);
						for (k = 0; k < b->cps[j].uportcount; k++) {
							printf("cbpf: uport %d\n", b->cps[j].uports[k]);
						}
					}
					if (b->cps[j].ipprotoflag != 0) {
						printf("flag %d: b->cps[j].ipproto %d\n", 
							b->cps[j].ipprotoflag, b->cps[j].ipproto);
					}
					if (b->cps[j].synflag != 0) {
						printf("cbpf synflag set\n");
					}
					if (b->cps[j].finflag != 0) {
						printf("cbpf finflag set\n");
					}
					if (b->cps[j].resetflag != 0) {
						printf("cbpf resetflag set\n");
					}
				}
				else if (b->progkind[j] == BPFBPF) {
					printf("bpfexpr: bpf_spec[%d] fcode[%d] len %d\n",
						i, j, b->fcodes[j].bf_len);
				}
				else if (b->progkind[j] == 0) {
					break;
				}
			}
		}
	}
	if (topn_irc.configured) {
		printf("topn_irc %d\n", topn_irc.maxcount);
	}
	if (topn_irc.fileflag) {
		printf("topn_irc_file %s\n", topn_irc.ircfile);
	}
	if (patterns!=NULL) {
		printf("---------layer 7 patterns start ---------\n");
		patmatch_printPatterns(patterns);
		printf("---------layer 7 patterns end ---------\n");
	}
#ifdef scanmesh
	if (scanmesh.configured) {
		printf("scanmesh.configured %d\n", scanmesh.configured);
		printf("scanmesh.ipmeshlist.tcp_ww %d\n", scanmesh.ipmeshlist.tcp_ww);
		printf("scanmesh.ipmeshfile %s\n", scanmesh.ipmeshfile);
	}
#endif
	printf("---------config dump-------------------\n");
}

void
parse_netmask(struct bpf_cparams *cps, char *n)
{
	char *m;
	int bits;
	unsigned int mask = 0;
	struct sockaddr_in sa;
	int rc;

	m = strchr(n, '/');

	if (m) {
		bits = strtoul(m+1, 0, 0);
		if (bits) {
			mask = 0xffffffff << (32 - bits);	
		}
	}
	*m = '\0';
	rc = resolveAddress(n, &sa);
	if (rc <= 0) {
		fprintf(stderr, "invalid ip address %s\n", n);
			exit(1);
	}
	/* store the network and mask values in net byte order
	*/
	cps->net.s_addr = sa.sin_addr.s_addr;
	cps->netmask.s_addr = htonl(mask);
	if (debug) {
		printf ("address %x, mask %x\n", cps->net.s_addr,
			cps->netmask.s_addr);
	}

	/* do this computation now so we don't have to do it
	 * at runtime.
	 */
	cps->net.s_addr = cps->net.s_addr & cps->netmask.s_addr;

	if (debug) {
		printf ("address %x %x, mask %x %x\n", cps->net.s_addr,
			ntohl(cps->net.s_addr), 
			cps->netmask.s_addr,
			ntohl(cps->netmask.s_addr));
	}
}

void
parse_icmp( struct bpf_cparams *cps, char *n )
{
	int rc;
	int i1;
	int i2;

	if (!n) {
		fprintf(stderr, "parse_icmp: no input params - should be similar to %s:\n", "3 3");
		exit(1);
	}
	rc = sscanf(n, "%d %d", &i1, &i2);
	if (rc < 2) {
		fprintf(stderr, "parse_icmp: invalid string <%s>, should be similar to: %s\n", n, "3 3");
		exit(1);
	}
	if (debug)
		printf ("parse_icmp: %d %d %d\n", rc, i1, i2);
	cps->icmp_major = i1;
	if (i1 > 17 || i1 < 0) {
		fprintf(stderr, "parse_icmp: icmp major value invalid %d\n", i1);
		exit(1);
	}
	cps->icmp_minor = i2;
	if (i2 < 0) {
		fprintf(stderr, "parse_icmp: icmp major value invalid %d\n", i1);
		exit(1);
	}
	cps->icmpflag = 1;
}

void
parse_tcprange(struct bpf_cparams *cps, char *n )
{
	int rc;
	int i1;
	int i2;

	rc = sscanf(n, "%d %d", &i1, &i2);
	if (rc < 2) {
		fprintf(stderr, "parse_tcprange: invalid string <%s>, should be similar to: %s\n", n, "6881 6899");
		exit(1);
	}
	if (debug)
		printf ("parse_tcprange: %d %d %d\n", rc, i1, i2);
	if (i1 > 65536 || i1 < 0) {
		fprintf(stderr, "parse_tcprange: tcp port range value invalid %d\n", i1);
		exit(1);
	}
	cps->tports[0] = i1;
	if (i2 > 65536 || i2 < 0) {
		fprintf(stderr, "parse_tcprange: tcp port range value invalid %d\n", i2);
		exit(1);
	}
	cps->tports[1] = i2;
}

void
parse_udprange(struct bpf_cparams *cps, char *n )
{
	int rc;
	int i1;
	int i2;

	rc = sscanf(n, "%d %d", &i1, &i2);
	if (rc < 2) {
		fprintf(stderr, "parse_udprange: invalid string <%s>, should be similar to: %s\n", n, "0 10");
		exit(1);
	}
	if (debug)
		printf ("parse_udprange: %d %d %d\n", rc, i1, i2);
	if (i1 > 65536 || i1 < 0) {
		fprintf(stderr, "parse_udprange: udp port range value invalid %d\n", i1);
		exit(1);
	}
	cps->uports[0] = i1;
	if (i2 > 65536 || i2 < 0) {
		fprintf(stderr, "parse_udprange: udp port range value invalid %d\n", i2);
		exit(1);
	}
	cps->uports[1] = i2;
}

/* more pairs than 10 are ignored by scanf
*/
/* 
 * ugly in terms of the scanf as opposed to strtok
 * but it works.  since this entire C bpf lib is a *hack*,
 * then don't lose sleep over it.
 */
void
parse_ports(struct bpf_cparams *cps, char *n )
{
	int rc;
	char tokens[MAXPORTPAIRS*2][100];
	int ports[MAXPORTPAIRS*2];
	int i;
	char *opword1, *opword2;
	int tcount = 0;
	int ucount = 0;

	int paircount = 0;
	i = 0;
	while(1) { 
		if (i == 0) {
			opword1 = strtok(n, " \t");
			if (!opword1) {
				fprintf(stderr, "parse_ports parse error, no operators\n");
				exit(1);
			}
		}
		else {
			opword1 = strtok(NULL, " \t");
			if (!opword1) 
				break;
		}
		strncpy(tokens[i], opword1, 100);
		if (strcmp(tokens[i], "t") == 0) {
			tcount++;
			if (tcount > MAXPORTPAIRS) {
				fprintf(stderr, "parse_ports parse error may only have %d t ops\n", MAXPORTPAIRS);
				exit(1);
			}
		}
		else if (strcmp(tokens[i], "u") == 0) {
			ucount++;
			if (ucount > MAXPORTPAIRS) {
				fprintf(stderr, "parse_ports parse error may only have %d u ops\n", MAXPORTPAIRS);
				exit(1);
			}
		}
		else {
			fprintf(stderr, "parse_ports invalid operator <%s>: must be t or u\n", tokens[i]);
			exit(1);
		}
		opword2 = strtok(NULL, " \t");
		if (!opword2) {
			fprintf(stderr, "parse_ports parse error, odd count of operators\n");
			exit(1);
		}
		ports[i] = atoi(opword2);
		paircount++;
		if (paircount > MAXPORTPAIRS) {
			fprintf(stderr, "parse_ports parse error, too many port pairs: MAX ok is %d\n", MAXPORTPAIRS);
			exit(1);
		}
/*
printf("paircount %d:%d, %s, %s: tokens[i] %s: ports[] %d:\n", paircount, MAXPORTPAIRS, opword1, opword2, tokens[i], ports[i]);
*/
		i++;
	}

	/* should be divisible by 2
	*/
	if (debug) {
		printf("paircount %d\n", paircount);
		for ( i = 0; i < paircount; i++) {
			printf("%s %d\n", tokens[i], ports[i]);
		}
	}
	/* now walk the inputs, and check them, if ok
	 * store in cbpf c param structure
	 */
	cps->uportcount = 0;
	cps->tportcount = 0;
	i = 0;
	while (paircount) {
		if (strcmp(tokens[i], "t") == 0) {
			if (ports[i] < 0 || ports[i] > 65536) {
				fprintf(stderr, "parse_ports: invalid port <%d>\n", ports[i]);
				exit(1);
			}
			cps->tports[cps->tportcount++] = ports[i];
		}
		else if (strcmp(tokens[i], "u") == 0) {
			if (ports[i] < 0 || ports[i] > 65536) {
				fprintf(stderr, "parse_ports: invalid port <%d>\n", ports[i]);
				exit(1);
			}
			cps->uports[cps->uportcount++] = ports[i];
		}
		else {
			fprintf(stderr, "parse_ports: invalid string <%s>, should be either t or u\n", tokens[i]);
			exit(1);
		}
		i++;
		paircount--;
	}
}

void
parse_ipproto( struct bpf_cparams *cps, char *n )
{
	int rc;
	int i1;
	extern int errno;

	if (strcmp(n,"ip") == 0) {
		i1 = -1;	/* wildcard, means catch all of the ipv4 packets */
	} else if (strcmp(n,"tcp") == 0) {
		i1 = 6;
	}
	else if (strcmp(n,"udp") == 0) {
		i1 = 17;
	}
	else if (strcmp(n,"icmp") == 0) {
		i1 = 1;
	}
	else {
		errno = 0;
		i1 = strtol(n, NULL, 10);
		if (errno == EINVAL) {
			fprintf(stderr, "parse_ipproto: invalid op: %s\n", n);
			exit(1);
		}
	}
	if (i1 < -1 || i1 > 255) {
		fprintf(stderr, "parse_ipproto: ip protocol invalid integer: %d\n", i1);
		exit(1);
	}
	cps->ipproto = i1;
	cps->ipprotoflag = 1;	/* for debugging really */
}

void
parse_tcpflag( struct bpf_cparams *cps, char *n )
{
	int rc;
	int i1;
	extern int errno;

	if (strcmp(n,"s") == 0) {
		cps->synflag = 1;
	} else if (strcmp(n,"f") == 0) {
		cps->finflag = 1;
	}
	else if (strcmp(n,"r") == 0) {
		cps->resetflag = 1;
	}
	else {
		fprintf(stderr, "parse_tcpflag: invalid op: %s\n", n);
		exit(1);
	}
}

#ifdef TESTMAIN
int debug = 1;

char *program_name;
char *device_name = "lo0";	/* just for testing */
pcap_t *pd;
struct patdescr *patterns=NULL;

void
dump_te()
{
}

main(argc, argv)
int argc; 
char **argv;
{
	unsigned int netmask = htonl(0xff000000);  /* assume lo0 */

	configRead(argv[1],1,68,netmask, 0, NULL);
	/* implicit
	configDump();
	*/
}
#endif

