/*
 * store a string in a hashlist.  
 *
 * The stringlist structure must be supplied
 */

#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include "ehash.c"
#include "stringstore.h" 

/* stores the string if it has been seen before
 * else simply counts it
 * returns:
 *	char * non-NULL: match AND string returned.
 *	NULL: no space
*/
char *
storeString(struct stringlist *sl, char *str) 
{
	struct ssnode *n, *h;
	int mod;
	unsigned int hash;
	int l;

	/*
	 * hash the string
	 * search for the string
	*/
	l = strlen(str);
	hash = ehash(str,l);
	n = sl->hashlist[mod=(hash % HSIZE)];
	for (; n!= NULL ; n=n->next) {
		/* we may have found it, hash matches
		*/
		if (n && (n->hash == hash)) {
			/* really does match */
			if (strcmp(str, n->s)== 0) {
				n->count++;
				return (n->s);
			}
			/* it didn't match */
		}
	}

	
	/* you have to malloc the node too
	*/
	n = (struct ssnode *) malloc(sizeof(struct ssnode));
	if (n == NULL) {
		syslog(LOG_ERR, "malloc failure in storeString, entry/count %d",
			sl->entries);
		return (NULL);
	}
	/* now malloc the string
	*/
	n->s = (char *) malloc(l+1);
	if (n->s == NULL) {
		syslog(LOG_ERR, "malloc failure in storeString, entry/count %d",
			sl->entries);
		return (NULL);
	}
	n->s[l] = 0;
	/* store hash and copy string
	*/
	strncpy(n->s, str, l);
	n->hash = hash;
	n->count = 1;

	/* count total entries
	*/
	sl->entries++;

	/* insert the node in the hash list
	*/
	n->next = NULL;
	h = sl->hashlist[mod];
	if (h == NULL) {
		sl->hashlist[mod] = n;
	}
	else {
		n->next = h;
		sl->hashlist[mod] = n;
	}
	return(n->s);
} 

/*
 * print the set of strings out
 * we free the list entries after we print them out
 *
 *	walk the hashlist
 *	for any node found
 *		print to FILE with prefix first, then string
 *		free the node
*/
void
dumpString(struct stringlist *sl, FILE *fd, char *prefix) 
{

	struct ssnode *n;
	unsigned int c = sl->entries;  /* count of hash entries allocated */
	int i = 0;

	/* walk the hashlist
	*/
	for (i = 0 ; i < HSIZE; i++) {
		n = sl->hashlist[i];
		/* if n is non-null, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			fprintf(fd, "%s: %s count %d\n", prefix, n->s, n->count);
			n = n->next;
		}
	}  
} 

void
initString(struct stringlist *sl) 
{
	sl->entries = 0;
	bzero (&sl->hashlist[0], HSIZE * sizeof(struct ssnode *));
}

/* init the hash structure
*/
void
zeroString(struct stringlist *sl) 
{
	struct ssnode *n;
	struct ssnode *prev;
	int i;

	/* first get rid of malloc'ed storage
	*/
	for (i = 0; i < HSIZE; i++) {
		n = sl->hashlist[i];
		while(n != NULL) {
			prev = n;
			n = n->next;
			free(prev->s);
			free(prev);
		}
	}

}

#ifdef TESTMAIN
main()
{
	struct stringlist sl;
	char * rc;
	int i;

	for (i = 0; i < 100; i++) {
		initString(&sl);

		rc = storeString(&sl, "1.2.3.4->2.2.2.2");
		if (rc)
			printf("rc %s\n", rc);
		rc = storeString(&sl, "1.2.3.4->2.2.2.2");
		if (rc)
			printf("rc %s\n", rc);
		rc = storeString(&sl, "2.2.3.4->2.2.2.2");
		if (rc)
			printf("rc %s\n", rc);

		dumpString(&sl, stdout, "warning");

		zeroString(&sl);
	}

}
#endif
