#ifdef THREAD
/*
 * Copyright (c) 2007 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include "sync.h"
#include "spinlock.h"
#include "ourpcap.h"

void inline aquire(lock_t *lock)
{
	spinlock(lock);
	lock->owner = getpid();
}

void inline release(lock_t *lock)
{
	lock->owner = 0;
	unlock(lock);
}

// releases all locks calling thread is holding
void inline releaseall(lock_t *locks, int len)
{
	int i;
	pid_t pid = getpid();
	for (i = 0; i < len; i++) {
		// ensure calling thread owns lock
		if (locks[i].owner == pid) {
			release(&locks[i]);
		}
	}	
}

void sync_init(int nthreads)
{
	numpids = nthreads + 1;
	numthreads = nthreads;
	pids = malloc(sizeof(pid_t) * numpids); // add 1 for parent pid
	if (!pids) {
		perror("sync_init(): malloc error\n");
		exit(1);
	}
	pids[0] = getpid();
	
	threads = malloc(sizeof(thread_t) * numthreads);
	if (!threads) {
		perror("sync_init(): malloc error\n");
		exit(1);
	}	
	pending = 0;
	memset(locks, 0, NUMLOCKS);
}

// if sync_alert(), and calling thread will busy loop until sync_go() is called.  Only children should
// call this function.
inline int sync_waitforparent()
{
	if (pending) {
		aquire(&pendinglock);
		pending++;
		release(&pendinglock);
		// spin until parent tells children to go by setting pending
		// to zero;
		while (pending != 0)
			;
//		printf("child done waiting\n");

	}
}

inline int sync_pending()
{
	return pending;
}

// set flag telling all threads there is a pending action
inline void 
sync_alert()
{
//	printf("waiting for children to line up\n");
	aquire(&pendinglock);
	pending = 1;
	release(&pendinglock);
}

inline int 
isparent()
{
	pid_t pid = getpid();
	if (pid == pids[0])
		return 1;
	else
		return 0;
}

// if sync_alert() has been called, then calling thread will wait until
// all other threads are waiting in sync_waitforparent().  Only parent
// thread should call this function.
inline void sync_waitforchildren()
{	
	if (pending) {
		while (pending != numthreads + 1)
			;
	}
}

inline void sync_go()
{
//	printf("Telling child to go\n");
	aquire(&pendinglock);
	pending = 0;
	release(&pendinglock);
//	printf("GO!\n");

}

// The child thread will execute this function
int func(void *arg)
{
	thread_t* me;
	me = (thread_t *)arg;

	sync_waitforparent();
	if (ourpcap_loop(me->pd, me->cnt, lookup_if(me->dlt), me) < 0) {
		perror("oops\n");
		syslog(LOG_ERR, "ourmon pcap error: %s", pcap_geterr(me->pd));
	}
	pcap_close(me->pd);
	return 0;
}

inline void sync_createthreads(int nthreads, pcap_t *p, int dlt, int cnt)
{
	int i;
	sync_init(nthreads);
	sync_alert();
	for (i = 1; i < nthreads + 1; i++) {
		threads[i].pd = ourpcap_clone(p);
		if (!threads[i].pd) {
			perror("pcap_clone() failed\n");
			exit(1);
		}
		threads[i].dlt = dlt;
		threads[i].cnt = cnt;
		pids[i] = thread_create(func, &threads[i]);
	}
}

/* slay the child threads
 * make sure you don't slay yourself
 * as pids[0] is the parent
 */
inline void sync_destroythreads(thread_t *threads, int nthreads)
{
	int i;
	int rc = 0;
	pid_t pid = getpid();

	if (nthreads == 0) 
		return;

	syslog(LOG_ERR, "ourmon: killing nthreads %d", nthreads);
	for (i = 1; i <= nthreads; i++) {
		rc = kill(pids[i], SIGKILL);
/*
		syslog(LOG_ERR, "i %d, pid %d, rc %d", i, pids[i], rc);
*/
/*
		free(threads[i].stack);
*/
		
	}
}
#endif


