#! /usr/bin/perl

#
# tcpworm.pl
#
# syntax:
#
#  tcpworm.pl [TOO MANY FLAGS] -z directory OR input_file1..n  > output_file
#
# FLAGS:
# -4: limit summarization to inclusion of a specific port (4 for 445).
# -d: turn debug on
# -D: dbfile for this instance of tcpworm.pl (allows different
#	front-ends on same back-end system).
#	note: if no -D default is /tmp/portdb.  
# 	This file may grow (although it doesn't seem to get too big in
#	real-life).
#	note: /tmp/foo here produces /tmp/foo.db.
# -e: email syn report.  -e topn_count, default is 0 for topn_count.
#	count must be supplied and limits results to top_n N value.  0 means no limit.
#	-e may be used with either the 30-second port report OR in summarization mode.
#	It is intended to only be used with the emaildump.txt input provided from
#	the front-end.  All results are sorted by email syn count, outputs
#	include email_syncount/email_ww values.  the topn_count is intended to
#	limit the amount of DNS work done in summarization mode (which can kill
#	you on remote IPs).
# -E: eventfile
#	events are logged to the event file as supplied.  This is only used
#	for the system event log.
# -h: limit results to home network/mask (e.g., 10.0.0.0/8).  
# -H: limit results to addresses other than home network/mask (e.g., 10.0.0.0/8).  
# -i: if summarization, sort by instance count not ip address.
# -l: logdir - where is base log directory
# -N: do dns processing for ip addresses.  note this can take
#	a long time and should not be attempted during the 30 second ourmon
#	sample period.  batch processing is probably ok.
# -p: produce separate port signature file output.  Normally
#	it is not produced.   
# -P: port report is produced WITHOUT p2p hint flag info.  Note this can be used
#	for a separate run of tcpworm.pl in case you want a portreport.txt
#	file that is not intended for public consumption.
# -r net/mask: defaults to off.  addresses with a high ww (based on -W) 
#	may cause an event log message to be created.  an event log
#	message is issued iff the IP address in question falls in the supplied
#	subnet.
# -R net/mask: if event log in use, create a message for any 
#	host in the net/mask range that has gotten a P or D in flags.
#	defaults to off.  does NOT involve work weight.  suggest
#	it only be used with potdump output from front-end.
# -S: plus filename - use port db and put messages in port signature output log
# -t: sort by max total pkt count in all cases
# -w: limit results to work weight > N.  This is a filter and may be used
#	to get a portreport for only those hosts with a high work weight.
#	N must be in the range 0..100.
# -W: similar to -w, but -w applies to all IPs addresses and is a filter.
#	-W sets a work weight used to detect if 30-second events should
#	be reported to the -E event file.
# -z: a directory of input files - supply directory name
#
#  input_files: 1..N front-end output files in syn tuple format: tcpworm.txt
#  output_file: tcpworm long report (details, although it includes
#		the non-P2P version of the portreport)
#
# in general:
#	process tcpworm.txt file and produce stats, port signature
#	report and other things.
#
# we run tcpworm.pl in 2 possible ways:
#
#	1. with one input file for 30-second summary.  1 input, 1 output.
#
#	2. with many input files for hourly/daily summarization.  many inputs.  1 output.
#	in the latter case the long output format is not used so there is output to stdout.  
#	long summarization does NOT use
#	the following switches are also ignored if supplied: 
#
#	-D dbfile
#	-E eventfile
#
# tuple spoke here:
#	SYNTUPLE
#
# port store:
# ipport{ipdstport} = count
# WORMipport{ipdstport} = count
#
# ip address:
# ipaddrs{ipid} = count
#
# portSignature{} used for storing per ip_src portSig info
# portSignature{$ipsrc} = $nps;
# portSignature_flags{$ipsrc} = $flags;
# portSignature_work{$ipsrc} = $percent;
# portSignature_from{$ipsrc} = $tcpfrom;
# portSignature_to{$ipsrc} = $tcpto;
# portSignature_sample{$ipsrc} = store the number of dst packets sampled and the number of pkts sent.
# portSignature_server{$ipsrc} = synacks divided by syns.  
#
# unique port sig report:
# upsSig{} used for storing unique port sig tuple.
# index is e.g., "135,139,445" as string.  frequency/count not included.
# value is (date: count of instances seen: ip_src list (,,,,)),
# thus key: (date: instances: ip_src,...)
#
# this info is stored in the dbm database if it is new.
# it is looked up from the dbm database, and if not new this is shown
# in the report with the date and the original set of ips.

# this is the default portdb.db file for port signatures
# at this point it always exists but may be overridden
# with a runtime switch
#
# EVENTS:
#	we document the events in the event log possible with tcpworm.pl.
#
# 1. TCP port report ww threshold violation - 30 second event
#	format: tcp port report line
#	meaning: the IP in question had a ww higher than the threshold
#		this event is not stored in the event log database.
#		It may be filtered for inside home IP use (all IPs by deafult).
#		see -r 
# 2. honeypot check - hourly summarization
#	format: "honeypot check on: <IP address>"
#	meaning: the IP address in question wrote into the honeypot address space
#		this event is stored in the event log database and is only done
#		once per day (bad idea?).  This means the host in question shows
#		a P for an apps flag. The port report is not given and must be looked up.
# 3. darknet check - hourly summarization.
#	format: "darknet check on: <IP address>"
#	meaning: the IP address in question wrote into the darknet address space
#		this event is stored in the event log database and is only done
#		once per day (bad idea?).  Note that this means an associated
#		TCP port report (not given in the event) shows a D for an apps flag.
#
$dbmfile = "/tmp/portdb";
$logdir = "/home/mrourmon/logs";

# no events unless we are told we have an event file
# no default 
$eventflag = 0;

use Socket;
use Getopt::Std;
use File::Basename();

$stime = time();

# global variables
# set to 1 for dns lookup.  dns lookup is SLOW
# so this is off by default.  
$dodns = 0;

# HACK: ignore
# set output name for synthesized counters in file to add to mon.lite
$monlite2 = "/tmp/mon.xxx";
$xtramon = 0;		# turn off if we don't want syn. counters

# set to top N for number of sorted ports to be output in report
$topnports = 20;

# set top N ip addresses for wormy ip address report
$topnipaddr = 20;

# debug on, off by default
$debug = 0;

# need at least N emule packets to be identified before we declare it emule
# 1 however seems to suffice at this point
$emulecount = 1;

# GUESS
# guesstimator basic counters
$g_totalpop = 0;        # total population
$g_p2psig = 0;          # total p2p sigs only B/G
$g_p2pguess = 0;        # total p2p guesses
$g_p2psig_match = 0;    # total matches guess/sigs
$g_p2pguess_nomatch = 0;# total p2p guesses but no sig match
$g_p2psig_nomatch = 0;  # total p2p sigs but no guess match
# END GUESS

%options=();
getopts("R:S:W:r:l:e:4:z:ditNP:p:D:E:w:h:H:",\%options);
# like the shell getopt, "d:" means d takes an argument
#print "Unprocessed by Getopt::Std:\n" if $ARGV[0];
#foreach (@ARGV) {
#  print "$_\n";
#}
if (defined $options{d}) {
	$debug = 1;
}
if (defined $options{N}) {
	$dodns = 1;
}
$portreport = "";

$doapps = 1;
if (defined $options{p}) {
	$portreport = $options{p};
	#$portreport2 = $portreport .  "2";
}

# if -P is used, we suppress any application hint info
# this is a PSU hack and is not expected to be used elsewhere.
#
if (defined $options{P}) {
	$portreport = $options{P};
	#$portreport2 = $portreport .  "2";
	$doapps = 0;
}

$dbmfile_flag = 0;		# don't do dbm thing
if (defined $options{D}) {
	$dbmfile = $options{D};
	$dbmfile_flag = 1;
}

# -r restrict flag
$restrict_flag = 0;
if (defined $options{r}) {
	$restrict_flag = 1;
	$restrict_netmask = $options{r};	# netmask address should be 10/8 or the like.
}

# restrict tcp ww event log message to P or D flags
$restrict_topot = 0;
if (defined $options{R}) {
	$restrict_topot = 1;
	$restrict_netmask = $options{R};	# netmask address should be 10/8 or the like.
}

# -s port signature log filename 
$portsiglog_flag = 0;		
if (defined $options{S}) {
	$portsiglog_filename = $options{S};
	$portsiglog_flag = 1;
}

# -W workweight to drive event logging only
# defaults to 0
#
$ww_event_flag = 0;		
$ww_event_value = 0; 
if (defined $options{W}) {
	$ww_event_value = $options{W};
	$ww_event_flag = 1;
}

$emailsyns = 0;
$emailcount = 0;

# email style reports and/or summary
if (defined $options{e}) {
	$emailsyns = 1;			# capability is on
	$emailcount = $options{e};	# limit report to these values
}

# event log file (may not be same file every time)
if (defined $options{E}) {
	$eventfile = $options{E};	# name of ASCII log file
	$eventflag = 1;			# capability is on
}

# sort by instance count
$sortbyinstance = 0;
if (defined $options{i}) {
	if ($emailsyns) {
		print "if -e, -i or -t cannot be used\n";
		exit(1);
	}
	$sortbyinstance = 1;
}

# sort by total count
$sortbytotal = 0;
if (defined $options{t}) {
	if ($emailsyns) {
		print "if -e, -i or -t cannot be used\n";
		exit(1);
	}
	$sortbytotal = 1;
}

if ($sortbytotal && $sortbyinstance) {
	print "either -i or -t but not both\n";
	exit(1);
}

# use this workweight to filter results.  default is no filter.
$filter_workweight = 0;
if (defined $options{w}) {
	$filter_workweight = $options{w};	# new workweight filter to use
	if ($filter_workweight < 0) {
		usage();			# user is idiot
	}
	if ($filter_workweight > 100) {
		$filter_workweight = 100;
	}
}

# filter output based on a net/mask pair.  i.e., ip must be "inside"
# net/mask.  
$home_netmaskflag = 0;
if (defined $options{h}) {
	$home_netmask = $options{h};	# netmask address should be 10/8 or the like.
	$home_netmaskflag = 1;		# capability is on
}

# -h and -H are mutually exclusive
if (defined $options{H}) {
	if ($home_netmaskflag) {
		usage();
	}
	$home_netmask = $options{H};	# netmask address should be 10/8 or the like.
	$home_netmaskflag = 2;		# not capability is on
}

$portlimit = -1;
if (defined $options{4}) {
	$portlimit = $options{4};	# this port must be uttered
}

# override of default logdir
if (defined $options{l}) {
	$logdir = $options{l};	
}

# event db if used is located in logdir
$event_dbmfile = $logdir . "/eventdb";

$zflag = 0;
if (defined $options{z}) {
        $directory = $options{z};       # get input file directory
	$zflag = 1;

        opendir (DIR, $directory) or
            die "can't open $directory directory: $!\n";
        @FILES = readdir DIR;
        closedir DIR;
        $size = scalar @FILES;
        while ($size > -1) {
                # prepend directory path to files
                if ( -f "$directory/$FILES[$size]" ) {
                        $FILES[$size] = $directory.'/'.$FILES[$size];
                }
                $size--;
        }

}
elsif ($ARGV[0]) {
        #$inputfile = $ARGV[0];
        @FILES = @ARGV;
}
else {
	usage();
}

# note filecount if 1 is sample mode, if > 1 is summarization mode
# with summarization mode, we expect a timestamp in the filename
#
$filecount = scalar @FILES;		# number of files

if ($filecount > 1) {
	$dbmfile_flag = 0;
}

# dump params for debug and exit
$dumpparams=0;
if ($dumpparams) {
	print "debug=$debug\n";
	print "portsiglog_filename=$portsiglog_filename\n";
	print "portsiglog_flag=$portsiglog_flag\n";
	print "sortbyinstance=$sortbyinstance\n";
	print "sortbytotal=$sortbytotal\n";
	print "dodns=$dodns\n";
	print "portreport=$portreport\n";
	print "doapps=$doapps\n";
	print "dbmfile=$dbmfile\n";
	print "dbmdoit=$dbmdoit\n";
	print "eventfile=$eventfile\n";
	print "eventflag=$eventflag\n";
	print "filecount=$filecount\n";
	print "portlimit=$portlimit\n";
	print  "FILES=@FILES\n";
	print  "$FILES[0] $FILES[1]\n";
	exit(1);
}

################################################### 

# init counters

# set various metric/weight constants
$wormconstant = 20;		# sins - fins - just used for nonsum stats
$econstant = 100000;		# E in EWORM constant
$ouchfins = 5;			# O in WORM constant
$Rresets = 5;			# R in WORM constant

# work weight used with port signature instance tuple
# we only store a port sig in the instance hash if
# it's weight is >= to this variable.
$ups_weight = 90;

# simple flag counters
$totaltuples = 0;
$Ecount = 0;	# number of E flags
$Wcount = 0;	# number of W flags
$wcount = 0;	# number of w flags
$Ocount = 0;	# number of O flags
$Rcount = 0;	# number of R flags
$Mcount = 0;	# number of M flags
$WWcount = 0;	# number of times S - F > C satisfied
# combo flag counters
$WOcount = 0;
$EWcount = 0;
$WORcount = 0;
$WORMcount = 0;
$WOMcount = 0;
$WMcount = 0;
$WRMcount = 0;
$comboTotal = 0;

#####################################################################
# averages for various weights
$allwork1 = 0;		# work average for all tuples
$WORMwork1 = 0;		# work average for all WORM tuples
$Wflagwork = 0;		# work average for any Wflag tuple

# work without fins in it
$allworkprime = 0;

# fins divided by syns
$allfinpercent = 0;
$WORMfinpercent = 0;

# packets recv divided by packets sent
$allrecvpercent = 0;
$WORMrecvpercent = 0;

# resets divided by syns sent
$allresetpercent = 0;
$WORMresetpercent = 0;

# work averages for ports 25 and port 80
# of the non-worm variety
$port25work = 0;
$port25count = 0;
$port80work = 0;
$port80count = 0;

$portSigReport = "";		# string concatenation report
######################################################
$start_time=`date`;
$start_time=~ s/\n//g;

if ($filecount == 1) {
	# get the UNIX time in seconds

	# banner
	tisaworm();

	# parse the input file
	doit($FILES[0]);

	# sort
	sortit();

	# final output printing
	finalstats();
}
else {
	while ($inputfile = shift @FILES) {
		# paranoia, not needed to avoid dot and dotdot
		if ($inputfile eq ".") {
			next;
		}
		if ($inputfile eq "..") {
			next;
		}
		# note that we expect a timestamp
		$curts = "";
		doit($inputfile);
	}
	sumreport();
	sumguide();
}

$etime=time();

$ftime = $etime - $stime;
print "\nelapsed time: $ftime seconds\n";
exit($ftime);

# end of main

######################################################################
# functions

#
# doit
# parse input file
#
#	1. parse and store syn tuple - file to variables
#	2. compute work weight for this tuple.
#	3. filter netmask/workweight to leave tuples out if desired
#	4. compute a few global stats  
#	5. call writeSynReport to compute rs string  
#	6. write out long stats (stdout) per tuple stats string in non-summarized mode.
#
sub doit() 
{ 
	
	my $inputfile = $_[0];
	my $curline;
	my @words;
	my $sysname;
	my $synacks;

	# summarization only uses this
	# extract timestamp from file into curts if any.
 	# e.g., Fri_May__6_16:07:34_PDT_2005.tcpworm.txt
	# 
	my $basename = File::Basename::basename($inputfile);
	my @PARTS = split(/\./, $basename);
	if (check_ts($PARTS[0])) {
		# curts is a global
		$cur_ts = $PARTS[0];
	} 

	open(FILE, $inputfile) || die "can't open $inputfile: $!\n";

	# line format:
	# ip src :syn:synacks:fins:finssent:rsts:tcpsent:tcpreturned:icmp error returned: 
	# 0       1   2       3    4        5    6       7           8
	# apphints:  apptags.lower : apptags.upper: edonkeyhits: ipdsts: tcpdsts: tcpsps: 
	# 9          10              11             12           13      14       15      
	# one src port:  sampled ip dst:  p25_syn count:  p25_ww: ports(1-n): (dst port, count) ... 
	# 16             17               18              19      20          21
	#
	if ($filecount == 1) {
		print "sorted syn tuples follow:\n\n";
	}
	while(<FILE>) {

		$totaltuples++;
		s/\n//g;
		$curline = $_;
		@words = split(/:/, $curline);
		if($words[0] =~ /^#./){
			next;
		}
		$flag="";
		$sysname = "";
		if ($dodns) {
			$sysname = gethostip($words[0]);
		}

		$syns = $words[1];
		$synacks = $words[2];
		$fins = $words[3];
		$finssent = $words[4];
		$rsts = $words[5];
		$tcpfrom = $words[6];		# tcp packets sent
		$tcpto = $words[7];		# tcp packets returned
		$icmpcount = $words[8];		# icmp packets (errors) returned
		$appflags = $words[9];		# L7/p2p appflags
		$apptagsLower = $words[10];
		$apptagsUpper = $words[11];
		$emulehits = $words[12];	# emulehits
		# suppress the above 2 if so desired
		if ($doapps == 0) {
			$appflags = 0;
			$emulehits = 0;
			$apptagsLower = 0;
			$apptagsUpper = 0;
		}
		$ipdsts = $words[13];		
		$tcpdsts = $words[14];		
		$tcpscount = $words[15];	# count of tcp src ports (10 max)
		$tcpsrc = $words[16];		# one sampled tcp src port
		$ipdst = $words[17];		# one sampled ip dst address
		# TBD now do something with it
		$p25_syns = $words[18];		# port 25 syns seen
		$p25_ww = $words[19];		# port 25 work weight seen
		$size0 = $words[20];		
		$size1 = $words[21];		
		$size2 = $words[22];		
		$size3 = $words[23];		
		$ports = $words[24];		# # port tuples
		$portstring = $words[25];	# portstring
		#$total = $tcpfrom + $tcpto;	# total tcp pkts exchanged

		# work equation at this point is:
		# syns_sent + fins_sent + resets_returned 
		# ---------------------------------------
		# total pkts sent + (total pkts returned)
		#
		# except we throw in the nmap check		
		# if total pkts returned - resets award 100%
		#
		# this is roughly 0..1 in range BUT we need to make sure
		# that is the case	

		$total = $tcpfrom + $tcpto;	# total tcp pkts exchanged 
		$tcontrol = $syns + $finssent + $rsts;	# control numerator

		# avoid divide by 0 and reduce likely result to 0 value (0/1)
		if ($total == 0) {
			$total = 1;
		}
		# make workweight into integer percent
                my $workweight = int(($tcontrol * 100.0)/$total);

		# now make sacred nmap test to rule out case of
		# only packets returned are mostly resets
		# if we return packets and they are mostly resets (95%)
		# we flag this case.  the front-end will tend to
		# force such cases to have more resets than the worm
		# weight.
		my $testtcpto = $tcpto;
		if ($testtcpto == 0) {
			$testtcpto = 1;
		}	
		if ($testtcpto && ($rsts/($testtcpto * 1.0)) > 0.95) {
			$workweight = 100;
		}
		# make sure it is not about 100% which is impossible
		# but make sure it is impossible!
		if ($workweight > 100) {
			$workweight = 100;
		}

		# filter netmask here if desired
		if ($home_netmaskflag) {
			my $rc;
			# rc is 1 if in subnet, 0 if not in subnet
			$rc = insubnet($home_netmask, $words[0]);
#print "$words[0] $rc\n";
			if ($rc == 1) {
				if ($home_netmaskflag == 2) {
#print "skipping: $words[0] $rc\n";
					next;
				}
			}
			elsif ($rc == 0) {
				if ($home_netmaskflag == 1) {
#print "skipping: $words[0] $rc\n";
					next;
				}
			}
		}

		# filter work weight here if desired
		if ($workweight < $filter_workweight) {
			next;
		}

		if ($filecount == 1) {

			# these are averages
			$allwork1 = $tcontrol/($total * 1.0) + $allwork1;
			# this average does not include fins
			$allworkprime = ($syns + rsts)/($total * 1.0) + $allworkprime;

			# this check is needed because excluding resets in total can
			# lead to value greater than 100%
			if ($allwork1 > 1.0) {
				$allwork1 = 1.0;
			}
			if ($allworkprime > 1.0) {
				$allworkprime = 1.0;
			}
			if (($syns - $fins) > $wormconstant) {
				$WWcount++;
			}
			$sumTotal{$words[0]} = $total;
		}

                # worm opinion metric
		$wopinion = "";
                if ($workweight >= 90) {
                        $wopinion = "W";
			$Wcount++;
                }
                elsif ($workweight > 40) {
                        $wopinion = "w";
			$wcount++;
                }

                $rs = writeSynReport($words[0], $syns, $synacks, 
			$fins, $finssent, $rsts, $tcpfrom, $tcpto, $total,
                        $icmpcount, $ports, $portstring, $workweight, 
			$wopinion, $appflags, $apptagsLower, $apptagsUpper, $emulehits, $ipdsts, 
			$tcpdsts, $tcpscount, $tcpsrc, $ipdst, $p25_syns, $p25_ww);

		# write out non summarized long per tuple stats string
		# note that this is the original "tcpreport" and is not
		# the port signature version.  it is sorted implicitly by
		# the front-end and is sorted by most syns.  hence topn_syn
		# syn tuple.  mostly for debug and details.  it is only
		# a 30second report too. not long summarization.
		if ($filecount == 1) {
			if ($sysname) {
				print "($sysname):$rs\n";
			}
			else {
				print "$rs\n";
			}
		}
	}  # end loop
	close(FILE);
}



# 
# writeSynReport()
#
# similar function in omupdate.pl
#
sub writeSynReport
{
	my $ipsrc = $_[0];
	my $syns = $_[1];
	my $synacks = $_[2];
	my $fins = $_[3];
	my $finssent = $_[4];
	my $rsts = $_[5];
	my $tcpfrom = $_[6];
	my $tcpto = $_[7];
	my $total = $_[8];		# tcpfrom + tcpto
	my $icmpcount = $_[9];
	my $ports = $_[10];
	my $portstring = $_[11];
	my $percent = $_[12];		# workweight
	my $wflag = $_[13];		# W flag or w flag
	my $appflags = $_[14];
	my $apptagsLower = $_[15];
	my $apptagsUpper = $_[16];
	my $emulehits = $_[17];
	my $ipdsts = $_[18];
	my $tcpdsts = $_[19];
	my $tcpscount = $_[20];
	my $tcpsrc = $_[21];
	my $sipdst = $_[22];		# sampled ip dst
	my $p25_syns = $_[23];
	my $p25_ww = $_[24];
	my $oflag = "";
	my $rflag = "";
	my $mflag = "";
	my $eflag = "";
	my $WORMflag = 0;
	my $JUSTWflag = 0;
	my $i;
	my $rs;
	my $sawtwo = 0;
	my $nps="";

     	@ps = split(/,/,$portstring);

	# just W flag
	if ($wflag eq "W") {
		$JUSTWflag = 1;
		$Wflagwork = ($workweight/100.0) + $Wflagwork;
	}
	# do compute of remaining weights and flags

	$allfinpercent = 0;
	$allresetpercent = 0;
	if ($syns != 0) {
		$allfinpercent = ($fins / ($syns * 1.0)) + $allfinpercent;
		$allresetpercent = ($rsts / ($syns * 1.0)) + $allresetpercent;
	}

	$allrecvpercent = 0; 
	if ($tcpfrom != 0) {
		$allrecvpercent = ($tcpto / ($tcpfrom * 1.0)) + $allrecvpercent;
	}

	# 1. compute E, the quadratic weight
	# (sins - fins) * (rst + icmp)
	my $quadw = ($syns - $fins) * ($rsts + $icmpcount);
	if ($quadw > $econstant) {
		$eflag = "E";
		$Ecount++;
	}

	# Ouch (oofffff?) flag - too few fins returned
	# 5. syns high (> 20), fins low < $ouchfins).  O is for Ouch.
	# really means no fins.
	if (($syns > 20) && ($fins < $ouchfins)) {
		$oflag = "O";
		$Ocount++;
	}

	# 6. no FINS, some resets flag: R is for Resets.
	if (($fins == 0) && ($rsts > $Rresets)) {
		$rflag = "R";
		$Rcount++;
	}
	# 7. no packets returned flag: M is for (keeping) Mum.
	if ($tcpto == 0) {
		$mflag = "M";
		$Mcount++;
	}
 
	# if summarization mode, call sumworm to summarize
	# and we are done for summarization, but not for single sample
	if ($filecount > 1) {
		sumworm(
		$ipsrc, $syns, $synacks, $fins, $finssent, 
		$rsts, $tcpfrom, $tcpto, $total, $icmpcount, 
		$ports, $portstring, $percent, $appflags, $apptagsLower, $apptagsUpper,$emulehits, 
		$ipdsts, $tcpdsts, $eflag, $wflag, $oflag, 
		$rflag, $mflag, $p25_syns, $p25_ww);
		return("");
	}

        # calculate combo flags

        # WO
        if (($wflag eq "W") && ($oflag eq "O")) {
                $WOcount++;
                $sawtwo = 1;
        }
        # EW
        if (($eflag eq "E") && ($wflag eq "W")) {
                $EWcount++;
                $sawtwo = 1;
        }
        # WOR
        if (($wflag eq "W") && ($oflag eq "O") && ($rflag eq "R")) {
                $WORcount++;
                $sawtwo = 1;
        }
        # WORM
        if (($wflag eq "W") && ($oflag eq "O") && ($rflag eq "R") && ($mflag eq "M")) {
                $WORMcount++;
                $sawtwo = 1;
                $WORMwork1 = ($workweight/100.0) + $WORMwork1;
                if ($syns) {
                        $WORMfinpercent = ($fins / ($syns * 1.0)) + $WORMfinpercent;
                        $WORMresetpercent = ($rsts / ($syns * 1.0)) + $WORMresetpercent;
                }
                if ($tcpfrom) {
                        $WORMrecvpercent = ($tcpto / ($tcpfrom * 1.0)) + $WORMrecvpercent;
                }
                $WORMflag = 1;    # WORM set
        }

	if (($wflag eq "W") && ($oflag eq "O") && ($mflag eq "M")) {
		$WOMcount++;
		$sawtwo = 1;
	}
	# WM
	if (($wflag eq "W") && ($mflag eq "M")) {
		$WMcount++;
		$sawtwo = 1;
	}
	# WRM
	if (($wflag eq "W") && ($rflag eq "R") && ($mflag eq "M")) {
		$WRMcount++;
		$sawtwo = 1;
	}

	if ($sawtwo) {
		$comboTotal++;
	}

#	print "parsing appflags=$appflags\n";
#	print "parsing hex appflags=",hex($appflags),"\n";
	$af = parseAppFlags(hex($appflags), $emulehits, $ipsrc);
	$af = parseLowerAppTags(hex($apptagsLower),$af);
	$af = parseUpperAppTags(hex($apptagsUpper),$af);

	$rs = "$ipsrc: s=$syns: sa=$synacks: f=$fins: fs=$finssent: r=$rsts: total=$total($tcpfrom:$tcpto): i=$icmpcount: ehits=$emulehits: ipdsts=$ipdsts: tcpdst = $tcpdsts\n";
	$rs = $rs . "\tweights/flags: ";
	$rs = $rs . "work=$percent%, Quad=$quadw, Flags:($eflag$wflag$oflag$rflag$mflag): Appflags: $af\n";
	my $flagstring = "($eflag$wflag$oflag$rflag$mflag)";

	#storeIpAddress($ipsrc, $WORMflag);
	# Wflag is good enough, bin IPs on that basis (if set)
	storeIpAddress($ipsrc, $JUSTWflag);

	# now loop through ports

	$rs = $rs . "\tdst ports=$ports: (dst port, count)\n";
	$pindex = 0;
	for ($i = 0; $i < $ports; $i++) {
		$pindex = $i * 2;
		$rs = $rs . "\t\tport[$i]=($ps[$pindex], $ps[$pindex+1])\n";

		# calculate some global stats. 
		# W flag port counts vs non-W flag port counts
		# also for non-W flag, do average work weight for
		# ports 80/25 as matter of interest
		if ($JUSTWflag) {
			$WORMipports{$ps[$pindex]} += $ps[$pindex+1];
		}
		# W flag is not set
		else {
			if ($ps[$pindex] == 80) {
				$port80work = ($workweight/100.0) + $port80work;
				$port80count++;
			}
			elsif ($ps[$pindex] == 25) {
				$port25work = $workweight/100.0 + $port25work;
				$port25count++;
			}
			$allipports{$ps[$pindex]} += $ps[$pindex+1];
		}
	}
	# call portsig to produce port signature report
	portSig($ipsrc, $flagstring, $percent, $ports, $ps, 
		$syns, $synacks, $fins, $rsts, $tcpfrom, 
		$tcpto, $total, $icmpcount, $appflags, $apptagsLower,$apptagsUpper,
        $emulehits, $ipdsts, $tcpdsts, $tcpscount, $tcpsrc, $sipdst, $p25_syns, $p25_ww);
	#$portSigReport = $portSigReport . "$ipsrc: \t$nps\n";
	# rs returned for stdout long report
	$rs = $rs . "\t\t$nps\n";
	return($rs);
}

#
# portSig
#
# take the unsorted port tuple and sort it
# according to the following math. notion
#
# 1. sort acc. to the port value itself from decreasing to increasing
# 2. determine the total port count frequency
# 3. express the count as a fraction of that frequency.
# 4. store the ipsrc, flagstring, percent(work) and portsig, etc.
#	in associative arrays.
#
# note: output is NOT done here, just some computation
# and storing of things in associative arrays for later output
#

sub portSig
{
	my $ipsrc=$_[0];		# ip src
	my $flags=$_[1];		# flags combined into one string
	my $percent=$_[2];		# work weight
	my $ports=$_[3];		# number of ports
	my $ps=$_[4];			# unprocessed as of yet port string
	my $syns=$_[5];			
	my $synacks=$_[6];			
	my $fins=$_[7];
	my $rsts=$_[8];
	my $tcpfrom=$_[9];
	my $tcpto=$_[10];
	my $total=$_[11];
	my $icmpcount=$_[12];
	my $appflags=$_[13];
	my $apptagsLower=$_[14];
	my $apptagsUpper=$_[15];
	my $emulehits=$_[16];
	my $ipdsts=$_[17];
	my $tcpdsts=$_[18];
	my $tcpscount=$_[19];
	my $tcpsrc=$_[20];
	my $sipdst=$_[21];
	my $p25_syns=$_[22];
	my $p25_ww=$_[23];
        my $tsend;
        my $trecv;
        my $pc;
	my $nps = "";			# nps is a signature (ports/frequency tuples)
	my $ups = "";			# ports only, no counts/frequency value
	my @pset;
	my @sp;
	my $pcount = 0;
	my $i;
	my @tuple;
	my $tmpx;

	# first compute the total port count
 	# meaning: the total number of dst sampled packets.	
	# which is NOT necessarily the total number of packets sent.
	for ($i = 0; $i < $ports; $i++) {
		my $pindex = $i * 2;
		$pcount += $ps[$pindex+1];
	}
	# now go thru and compute a percent for the 
	# destination port count (hits/total) and store it in an hash
	# array
	for ($i = 0; $i < $ports; $i++) {
		my $pindex = $i * 2;
		# ignore port if count is 1
		#if ($ps[pindex+1] <= 1) {
		#	next;
		#}
		if ($pcount) {
			$pset{$ps[$pindex]} = int(($ps[$pindex+1] * 100.0)/ $pcount);
		}
	}

	# now sort the array based on the port number itself. 
	# sortedPorts is a list of the ports themselves
	@sp  = sort p_by_key keys(%pset);

	# now loop thru the ports
	# putting the ports into one string that is a signature tuple
	for ($i = 0; $i < $ports; $i++) {
		$nps = $nps . "[$sp[$i],$pset{$sp[$i]}]";
		$ups = $ups . "[$sp[$i]]";			# ports only
	}

	foreach $key (keys(%pset)){
#print "pset:", $key, $pset{key}, "\n";
		delete $pset{$key};
	}
	#foreach $key (keys(%sp)){
	#	delete $sp{$key};
	#}
	#$nps = $nps . ")";
	$portSignature{$ipsrc} = $nps;		# key is ip, value is ports
	$portSignature_flags{$ipsrc} = $flags;	# per ip src, flags 
	$portSignature_work{$ipsrc} = $percent; # per ip src, work

        $tsend = $tcpfrom;
        if ($tcpfrom > 1000) {
                $tsend = (int $tcpfrom/1000) . "k";
        }
        $trecv = $tcpto;
        if ($tcpto > 1000) {
                $trecv = (int $tcpto/1000) . "k";
        }
        $pc = $pcount;
        if ($pcount > 1000) {
                $pc = (int $pcount/1000) . "k";
        }

	# put instance values into associative array variables, indexed by IP
	#

	# join is used so that output can be done as string in form  x/y
	$emailTotal{$ipsrc} = $p25_syns;
        $portSignature_fromto{$ipsrc} = join("/", $tsend, $trecv);
        $portSignature_sample{$ipsrc} = join("/", $pc, $tsend);
        $portSignature_p25sample{$ipsrc} = join("/", $p25_syns, $p25_ww);

	$portSignature_appinfo{$ipsrc} = parseAppFlags(hex($appflags), $emulehits, $ipsrc);
	$portSignature_appinfo{$ipsrc} = parseLowerAppTags(hex($apptagsLower), $portSignature_appinfo{$ipsrc});
	$portSignature_appinfo{$ipsrc} = parseUpperAppTags(hex($apptagsUpper), $portSignature_appinfo{$ipsrc});

	$portSignature_dsts{$ipsrc} = join("/", $ipdsts, $tcpdsts);
	$portSignature_srcs{$ipsrc} = join("/", $tcpscount, $tcpsrc);
	$portSignature_sipdst{$ipsrc} = $sipdst;

	# synacks divided by syns is expressed as a percent.
	# note: synacks are counted as syns too.  
	if ($syns == 0) {
		$portSignature_server{$ipsrc} = 0;
	}
	else {
		$portSignature_server{$ipsrc} = int(($synacks * 100.0) / $syns);
	}

# GUESS
        guesstimator(
                $ipsrc,
                $percent,               # workweight
                $portSignature_server{$ipsrc},   # sa/s
                $ports,                 # number of ports
                $portstring,            # port string
                $tcpfrom,
                $tcpto,
                $total,
                $appflags,   
                $apptagsLower, $apptagsUpper,  
                $ipdsts,
                $tcpdsts,
                $tcpscount,
                $tcpsrc
        );
# END GUESS

################################################################################ 

	# turned off for now
#	$portsig2 = 0;
#	if ($portsig2) {
#		# experiments coming out in "hidden" portreport2.txt
#		# syns+resets/total
#		if ($total == 0) {
#			$total = 1;
#		}
#		$portSignature_work2{$ipsrc} = 
#			int((( $syns + $rsts ) * 100)/$total); 
#
#		# syns - fins
#		$portSignature_work3{$ipsrc} = $syns - $fins; 
#
#		# syns - fins * (icmp + resets)
#		$portSignature_work4{$ipsrc} = ($syns - $fins) * ($icmpcount + $rsts);
#
#		# total sent/total sent+total recv
#		$tmpx = $tcpfrom + $tcpto;
#		if ($tmpx == 0) {
#			$tmpx = 1;
#		}
#			
#		$portSignature_work5{$ipsrc} = int(($tcpfrom*100.0)/$tmpx);
#	}
################################################################################ 

	# if work weight is above a configured constant,
	# we compute total signature stats and save info in a dbm database
	#
	# key=$ups, value=$ipsrc, $start_time, instance count
	# $ups is a set of dst ports, minus pkt counts.
	#
	# tuple plan for value
	# (ipsrc, time, instance_count)
	# ipsrc is colon separated (ip:ip:ip)
	# 
	if ( $percent >= $ups_weight ) {
		if ($upsSig{$ups}) {
			# take it apart, and then
			# split via commas
			@tuple = split(",", $upsSig{$ups});

			$ups_iplist = $tuple[0];
			$ups_time = $tuple[1];

			# instance count of this signature (ports only) in this particular tcpworm.txt
			$ups_instance = $tuple[2];

			# join it back together
			# ipsrc are separated by colon
			$ups_iplist = $ups_iplist . ":$ipsrc";
			$ups_instance++;
#print "XXX: $ups: $ups_iplist, $ups_time, $ups_instance\n";
			# note: time string has colons in it, we use comma
			# as delimiter
			$upsSig{$ups} = join(",", $ups_iplist, $ups_time, $ups_instance);
		}
		# no signature so make one
		# value is: (ipsrc, start time, instance count == 1)
		else {
			$upsSig{$ups} = join(",", $ipsrc, $start_time, 1); 
		}
	}
}

# sort by the key for the ports array 
# above
sub p_by_key 
{
        if ($a < $b) {
                -1;
        }
        elsif ($a == $b) {
                0;
        }
        elsif ($a > $b) {
                1;
        }
}


# you must be a worm if 
# do worms have red necks?
#
# print worm EWORM flag and weight guide

sub tisaworm 
{
	print "ourmon tcpworm syn tuple report at ($start_time):\n"; 
	print "\ttuples are output here iff for an ip src: SYNS - FINS > $wormconstant\n\n";
	print "syn tuple syntax: ip src: syns: fins: resets: tcp pkts total(sent/returned): icmp errors:\n";
	print "\tweight schemes, flags: (EWORM)\n";
	print "\tport tuples (TCP dst port/count of pkts)\n\n";

	print "Flag guide: it might be a worm if:\n";
	print "\tE is for Errors: quadratic weight: (SYNS - FINS) * (ICMP errors + RSTS) is > $econstant:\n";
	print "\tWflag (w/W): work (tcp control/total pkts) is high: W >= 90%,  w = > 50%\n";
	print "\tOuch flag (O): too few fins: SYNS > 20, FINS < $ouchfins\n";
	print "\tResets flag (R): no FINS, some RSTS > $Rresets\n";
	print "\tM is for mum flag: no TCP packets returned to ip src\n";
	print "\tWhat's that spell?  WORM (or EWORM)\n\n";
}

sub finalstats
{
	my $t;

	print "#########################################################################\n";
	print "final stats: \n";
	print "#########################################################################\n";

	print "total tuples: $totaltuples\n";
	print "#########################################################################\n";
	print "print simple flags:\n";
	print "\ttotal E flags: $Ecount\n";
	print "\ttotal W flags: $Wcount\n";
	print "\ttotal w flags: $wcount\n";
	print "\ttotal O flags: $Ocount\n";
	print "\ttotal R flags: $Rcount\n";
	print "\ttotal M flags: $Mcount\n";
	print "#########################################################################\n";
	print "print combo flags:\n";
	print "\ttotal WORM flags: $WORMcount\n";
	print "\ttotal WOR flags: $WORcount\n";
	print "\ttotal WOM flags: $WOMcount\n";
	print "\ttotal WRM flags: $WRMcount\n";
	print "\ttotal WO flags: $WOcount\n";
	print "\ttotal WM flags: $WMcount\n";
	print "\ttotal EW flags: $EWcount\n";
	print "\ttotal combo flags (2 or more, at least one W): $comboTotal\n";
	print "#########################################################################\n";
	print "averages/weight info:\n";
	print "\ttotal worm metric count (S-F>C): $WWcount\n";

	if ($totaltuples == 0) {
		print "\ttotal work average for all tuples: 0%\n";
	}
	else {
		my $t = int($allwork1/($totaltuples*1.0) * 100);
		print "\ttotal work average for all tuples: $t%\n";
	}

	if ($totaltuples == 0) {
		print "\ttotal work prime (s+r/total) average for all tuples: 0%\n";
	}
	else {
		my $t = int($allworkprime/($totaltuples*1.0) * 100);
		print "\ttotal work prime (s+r/total) average for all tuples: $t%\n";
	}

	if ($WORMcount == 0) {
		print "\ttotal work average for all WORM tuples: 0%\n";
	}
	else {
		my $t = int($WORMwork1/($WORMcount*1.0) * 100);
		print "\ttotal work average for all WORM tuples: $t%\n";
	}

	if ($Wcount == 0) {
		print "\ttotal work average for all W tuples: 0%\n";
	}
	else {
		my $t = int($Wflagwork/($Wcount*1.0) * 100);
		print "\ttotal work average for all W tuples: $t%\n";
	}


	# fin percent
	if ($totaltuples == 0) {
		print "\ttotal fin percent average for all tuples: 0%\n";
	}
	else {
		my $t = int($allfinpercent/($totaltuples*1.0) * 100);
		print "\ttotal fin percent average for all tuples: $t%\n";
	}

	if ($WORMcount == 0) {
		print "\ttotal fin percent average for all WORM tuples: 0%\n";
	}
	else {
		my $t = int($WORMfinpercent/($WORMcount*1.0) * 100);
		print "\ttotal fin percent average for all WORM tuples: $t%\n";
	}

	# tcp recv/sent percent
	if ($totaltuples == 0) {
		print "\ttotal recv (tcpto/tcpfrom) percent average for all tuples: 0%\n";
	}
	else {
		my $t = int($allrecvpercent/($totaltuples*1.0) * 100);
		print "\ttotal recv (tcpto/tcpfrom) percent average for all tuples: $t%\n";
	}

	if ($WORMcount == 0) {
		print "\ttotal recv percent average for all WORM tuples: 0%\n";
	}
	else {
		my $t = int($WORMrecvpercent/($WORMcount*1.0) * 100);
		print "\ttotal recv percent average for all WORM tuples: $t%\n";
	}

	# reset percent
	if ($totaltuples == 0) {
		print "\ttotal reset percent (reset/syn) average for all tuples: 0%\n";
	}
	else {
		my $t = int($allresetpercent/($totaltuples*1.0) * 100);
		print "\ttotal reset percent (reset/syn) average for all tuples: $t%\n";
	}

	if ($WORMcount == 0) {
		print "\ttotal reset percent average for all WORM tuples: 0%\n";
	}
	else {
		my $t = int($WORMresetpercent/($WORMcount*1.0) * 100);
		print "\ttotal reset percent average for all WORM tuples: $t%\n";
	}

	if ($port80count == 0) {
		print "\ttotal port 80 percent average for non Wflag tuples: 0%\n";
	}
	else {
		my $t = int($port80work/($port80count*1.0) * 100);
		print "\ttotal port 80 percent average for non Wflag tuples: $t%\n";
	}

	if ($port25count == 0) {
		print "\ttotal port 25 percent average for non Wflag tuples: 0%\n";
	}
	else {
		my $t = int($port25work/($port25count*1.0) * 100);
		print "\ttotal port 25 percent average for non Wflag tuples: $t%\n";
	}

	# figure out numeric distance of IP source addresses
	# synthetic count tuples for RRDTOOL graphs
	if ($xtramon) {
		open(tfd, ">$monlite2") || die "cannot open $monlite2: $!\n";
		print tfd "wsflag: $Ecount: $Wcount: $wcount: $Ocount: $Rcount: $Mcount:\n";
		print tfd "wcflag: $WORMcount: $WORcount: $WOMcount: $WRMcount: $WOcount: $WMcount:\n";
		close(tfd);
	}

	# first print out some stats
	print "#########################################################################\n";
	# port stats divided into Wflag ports and the rest
	printPorts();
	print "#########################################################################\n";
	# Wflag IP bins
	printIp();

	# print the port sig report, possibly optionally to the portreport file
	print "#########################################################################\n";
	printPortSigs(STDOUT, 1);

	# above was to main output file -- full detailed file
	# below goes to summarized portreport.txt
	if ($portreport) {
		open(tfd, ">$portreport") || die "cannot open $portreport: $!\n";
		printPortSigs(tfd, 0);
		printUpsSigs(tfd);
		printAppHints(tfd);

		close(tfd);

#		open(tfd, ">$portreport2") || die "cannot open $portreport2: $!\n";
#		printPortSigs2(tfd);
#		close(tfd);
	}

	# print out unique port signatures -- we need the db database for this
	printUpsSigs(STDOUT);
	printAppHints(STDOUT);
}

# print out sorted port stats
sub printPorts
{
	my $i;
	my $port;
	my $count;

	#@Allsortedports  = sort allipports_by_value keys(%allipports);
	#@WORMsortedports  = sort WORMipports_by_value keys(%WORMipports);
	# print worm ports
	print "WORM tuples: top $topnports ports/counts:\n";
	for ( $i = 0; $i < $topnports; $i++) {
		$port = $WORMsortedports[$i];
		if ($port eq "") {
			next;
		}
		$count = $WORMipports{$port};
		print "\tport[$port]: $count\n";
	}
	print "\n";
	# print all ports
	print "all tuples: top $topnports ports/counts:\n";
	for ( $i = 0; $i < $topnports; $i++) {
		$port = $Allsortedports[$i];
		if ($port eq "") {
			next;
		}
		$count = $allipports{$port};
		print "\tport[$port]: $count\n";
	}
}


# bin IPs according to first byte as crude method for
# detecting similarity
#
# note: two bins
#	W flag set
#	not set (not printing this -- may be buggy)
#
sub storeIpAddress
{
	my $ip = $_[0];
	my $wflag = $_[1];
	my @words;
	my $ipid;

	# just keep the first byte
	$ip=~s/\./:/g;
	@words = split(/:/, $ip);
	# if class C, then 3 bytes, else 2 bytes
	if ($words[0] >= 192) {
		$ipid = $words[0] . '.' . $words[1] . '.' . $words[2];
	}
	else {
		$ipid = $words[0] . '.' . $words[1];
	}
#print "$ip/$ipid, $wflag\n";
	if ($wflag) {
		if ($WORMipaddrs{$ipid}) {
			$WORMipaddrs{$ipid} += 1;
		}
		else {
			$WORMipaddrs{$ipid} = 1;
		}
	}
	else {
		if ($ipaddrs{$ipid}) {
			$ipaddrs{$ipid} += 1;
		}
		else {
			$ipaddrs{$ipid} = 1;
		}
	}
}

#
# print out IP bins
# currently only doing Wflag related IP addresses.
#
sub printIp
{
	my $i = 0;
	my $ipaddr;
	my $count;

	#@Allsortedipaddrs  = sort allipaddrs_by_value keys(%ipaddrs);
	#@WORMsortedipaddrs  = sort WORMipaddrs_by_value keys(%WORMipaddrs);

	# print worm ipaddrs
	print "W flag IP address tuples: < $topnipaddr (first two bytes of address)/counts:\n";
	for ( $i = 0; $i < $topnipaddr; $i++) {
		$ipaddr = $WORMsortedipaddrs[$i];
		if ($ipaddr eq "") {
			next;
		}
		$count = $WORMipaddrs{$ipaddr};
		print "\t$ipaddr: $count\n";
	}
	print "\n";

	#print "\nall IP address tuples other than WORM: $topnipaddr (two bytes)/counts:\n";
	#for ( $i = 0; $i < $topnipaddr; $i++) {
	#	$ipaddr = $Allsortedipaddrs[$i];
	#	if ($ipaddr eq "") {
	#		next;
	#	}
	#	$count = $ipaddrs{$ipaddr};
	#	print "\t$ipaddr: $count\n";
	#	$i++;
	#}
	#print "\n";
}

sub dumpIpaddrs {
	foreach $key (keys %ipaddrs) {
		printf("key %s: count %s\n", $key, $ipaddrs{$key});
	}
}


# printPortSigs
#
#	print out the port signature info
#
#	note parameter -- open handle where to print info
#
#	e.g., portSignature{$ipsrc} = $nps;
#	and other portSignature arrays used here.
#
#	note: if dohints is set, we put in apps/hints info.
#
# $doevents is used to indicate that a possible tcp ww event log message may
# be generated.

sub printPortSigs
{
	my $fd = $_[0];
	my $doevents = $_[1];
	my $ip;
	my $sbuf;
	my @sports;
	my $i;

	# sort by ip src in ascending order or by total pkt count

	if ($emailsyns) {
		@sports = sort sort_by_total_email keys(%portSignature);
	}
	elsif ($sortbytotal) {
		@sports = sort sort_by_total_value keys(%portSignature);
	}
	else {
		# sort by ip src is default
		@sports = sort p_by_ipkey keys(%portSignature);
	}

	print $fd "port signatures:\n";
	# ip src      flags    work weight   port sig tuples 1..10
	if ($emailsyns) {
		$sbuf = sprintf(
                "%-16.16s %-10.10s %-10.10s %-9.9s %-10.10s %-5.5s %-5.5s %-9.9s %-9.9s %-16.16s %-12.12s %-12.12s %s\n\n", 
		"ip src:", "flags", "apps", "esyn/eww", "guess", "work", "SA/S", "L3D/L4D", "L4S/src", "ip dst:",  "snt/rcv", "sdst/total", "port signature(dst port, frequency)"); 
	}
	elsif ($doapps) {
		$sbuf = sprintf(
                "%-16.16s %-10.10s %-10.10s %-10.10s %-5.5s %-5.5s %-9.9s %-9.9s %-16.16s %-12.12s %-12.12s %s\n\n", 
		"ip src:", "flags", "apps", "guess", "work", "SA/S", "L3D/L4D", "L4S/src", "ip dst:",  "snt/rcv", "sdst/total", "port signature(dst port, frequency)"); 
	}
	else {
		$sbuf = sprintf("ip src:          flags      work   SA/S   L3D/L4D    L4S/src ip dst:               snt/rcv sdst/total   port signature(dst port, frequency)\n\n"); 
	}
	
	print $fd $sbuf;

	$i = 0;
	while (1) {
		$ip = $sports[$i];
		if ($ip eq "") {
			return;
		}

# GUESS - not clean here
		# same as doapps case, but one more field for email data
		if ($emailsyns) {
			$sbuf = sprintf("%-16.16s %-10.10s %-10.10s %-9.9s %-10.10s %-5d %-5d %-9.9s %-9.9s %-16.16s %-12.12s %-12.12s %s", 
				"$ip", 
				$portSignature_flags{$ip},   
				$portSignature_appinfo{$ip},
				$portSignature_p25sample{$ip},
				$portSignature_guess{$ip},			# GUESS
				$portSignature_work{$ip} , 
				$portSignature_server{$ip}, 
				$portSignature_dsts{$ip},
				$portSignature_srcs{$ip},
				$portSignature_sipdst{$ip},
				$portSignature_fromto{$ip},	# total
				$portSignature_sample{$ip},
				$portSignature{$ip});
		}
		elsif ($doapps) {
			$sbuf = sprintf("%-16.16s %-10.10s %-10.10s %-10.10s %-5d %-5d %-9.9s %-9.9s %-16.16s %-12.12s %-12.12s %s", 
				"$ip", 
				$portSignature_flags{$ip},   
				$portSignature_appinfo{$ip},
				$portSignature_guess{$ip},			# GUESS
				$portSignature_work{$ip} , 
				$portSignature_server{$ip}, 
				$portSignature_dsts{$ip},
				$portSignature_srcs{$ip},
				$portSignature_sipdst{$ip},
				$portSignature_fromto{$ip},	# total
				$portSignature_sample{$ip},
				$portSignature{$ip});
		}
# end GUESS
		else {
			$sbuf = sprintf("%-16.16s %-10.10s %3d:   %3d:  %9.9s %9.9s %-16.16s %12.12s %-12.12s %s", 
				"$ip", 
				$portSignature_flags{$ip},   
				$portSignature_work{$ip} , 
				$portSignature_server{$ip}, 
				$portSignature_dsts{$ip},
				$portSignature_srcs{$ip},
				$portSignature_sipdst{$ip},
				$portSignature_fromto{$ip},
				$portSignature_sample{$ip},
				$portSignature{$ip});
		}

#		# possible ww event log message
 		if ($doevents && $eventflag) {
 			my $domessage = 0;

			# if we want an event for writes to darknet/potnet
			# then we write out such an event
			if ($restrict_topot && ($portSignature_pot{$ip} > 0)) {
				my $rc;
				$rc = insubnet($restrict_netmask, $ip);
				# if not in subnet, don't do it
				if ($rc == 0) {
					;
				}
				else {
 		 			storeEvent($eventfile, 
					"TCP port report darknet/honeynet violation:", 
						$sbuf, 0);
				}
			}
			# if we want an event for high ww possibly in a
			# restricted subnet, we write that out 
 			if ($restrict_flag && $portSignature_work{$ip} >= $ww_event_value) {
				my $rc;
				$rc = insubnet($restrict_netmask, $ip);
				# if not in subnet, don't do it
				if ($rc == 0) {
 					;
				}
 				else {
 		 			storeEvent($eventfile, 
					"TCP port report ww threshold violation:", 
						$sbuf, 0);
 				}
 			}
 		} 
		# print to report
		print $fd "$sbuf\n";
		$i++;
	}
}

# work2 syns+resets/total
# work3 syns-fins
# work4 syns-fins * errors
# work5 sent/sent+recv
# experimental - this is only used for stats. gathering
#	and for testing of ideas - not the "production" normal output.

sub printPortSigs2
{
	my $fd = $_[0];
	my $ip;
	my $sbuf;
	my @sports;
	my $i;

	# sort by ip src in ascending order

	@sports = sort p_by_ipkey keys(%portSignature);

	print $fd "port signatures:\n";
	# ip src      flags    work weight   port sig tuples 1..10
	$sbuf = sprintf("%-16.16s %-8.8s (%4s: %5s:   %3s:  %5s: %5s) %5s: %9s: %s\n\n", "ip src:", "flags", "work", 
		"s+r/t", "s-f", "s-f*e", "S/S+R",
		"SA/S",
		"dst/total", "port signature(dst port, frequency)"); 
	print $fd $sbuf;

	$i = 0;
	while (1) {
		$ip = $sports[$i];
		if ($ip eq "") {
			printWorkHistogram($fd);
			printWork2Histogram($fd);
			printWork5Histogram($fd);
			return;
		}

		$sbuf = sprintf("%-16.16s %-10.10s (%3d: %3d: %6d: %6d: %3d:)     %3d %-10.10s %s", "$ip", 
			$portSignature_flags{$ip},   
			$portSignature_work{$ip} , 
			$portSignature_work2{$ip} , 
			$portSignature_work3{$ip} , 
			$portSignature_work4{$ip} , 
			$portSignature_work5{$ip} , 
			$portSignature_server{$ip},
			$portSignature_sample{$ip},
			$portSignature{$ip});
		print $fd "$sbuf\n";
		storeWorkHistogram($ip);
		storeWork2Histogram($ip);
		storeWork5Histogram($ip);
		$i++;
	}
}


# Ups - unique port sig
#
# print out port sigs that are unique
# on this sample pass.  We essentially
# boil the single ip instances down into classes
# as a result.  E.g., maybe you are seeing 10 different
# sasser worms or related things from 10 diff. ip srcs.
#
sub
printUpsSigs
{
	my $fd = $_[0];
	my $key;
	my $value;
	my @newtuple;
	my @oldtuple;
	my $oldv;
	my $sbuf;
	my @iplist;
	my $ip;
	my $itime;
	my $instances;
	
	if (!$dbmfile_flag) {
		return;
	}

	# open the database
	dbmopen(%map, $dbmfile, 0666);

	# print out a report header
	print $fd "\n\nUnique port signature report:\n";
	$sbuf = sprintf("port signature > %d% work weight.\n", $ups_weight);
	print $fd $sbuf;
	$sbuf = sprintf("if portsig is old, first time seen, and original ip sources are given.\n\n");
	print $fd $sbuf;
	$sbuf = sprintf("%-7.7s %-14.14s %s\n", "New/Old", "instances(now)", "port_signature");
	print $fd $sbuf;
	$sbuf = sprintf("-------------------------------------\n\n");
	print $fd $sbuf;
 
	# loop thru the keys in the upsSig hash array
	# the keys are a unique port signature for this run
	#
	# $key is $ups, a set of ports
	# 
 	foreach $key (keys %upsSig) {
		# get the value associated with the key
 		$value = $upsSig{$key};
		#print "$key $value\n";
		@newtuple = split(",", $value);
		# newtuple[0] = ipsrcs, time, instances
		$ip = $newtuple[0];
		$itime = $newtuple[1];
		$instances = $newtuple[2];
		#print "$newtuple[0],$newtuple[1], $newtuple[2]\n";
		# see if the key is in the database, else we must store it
		#
		if ($map{$key}) {
			# fetch database stored instance
			$oldv = $map{$key};	
			#print "oldv is $oldv\n";

			$sbuf = sprintf("%-7.7s %-14.14s %s\n", "Old", @newtuple[2], $key);	
			print $fd "$sbuf";
			@oldtuple = split(",", $oldv);
			$sbuf = sprintf("\toriginal time seen: @oldtuple[1]\n");
			print $fd "$sbuf";
			$sbuf = sprintf("\toriginal ip_srcs: ");
			print $fd "$sbuf";
			@iplist = split(":", $oldtuple[0]);
			$i = 0;
			$j = 0;
			while($iplist[$i]) {
				if ($j < 4) {
					print $fd "$iplist[$i++] ";	
					$j++;
				}
				else {
					$j = 0;
					print $fd "\n\t\t";
				}
			}
			print $fd "\n";
		}
		# key is not in the database and is 
		# therefore a new worm port sig
		else {
			# store in database
			$map{$key} = $value;	

			# info is new
			$sbuf = sprintf("%-16.16s %s\n", $ip, $key);	
			print $fd "$sbuf";
			# this is not the event log, it is the new port sig log
			if ($portsiglog_flag) {
		 		#storeEvent($eventfile, "new worm signature from: $value :", $sbuf);				
		 		storeEvent($portsiglog_filename, 
					"TCP port report: new worm signature from:", $sbuf, 0);				
			}
		}
	}
	dbmclose(%map);
}

sub 
p_by_ipkey 
{
        if (ipToNumber($a) < ipToNumber($b)) { 
                -1;
        }
        elsif (ipToNumber($a) == ipToNumber($b)) {
                0;
        }
        elsif (ipToNumber($a) > ipToNumber($b)) {
                1;
        }
}

sub
ipToNumber
{
	$ipaddr = $_[0];
	@w = split(/\./, $ipaddr);
	$hip = ($w[0] << 24) | ($w[1] << 16) | ($w[2] << 8) | $w[3];
	return($hip);	
}

# sort for 30 second reports
sub sortit()
{
	# ipports
	@Allsortedports  = sort allipports_by_value keys(%allipports);
	@WORMsortedports  = sort WORMipports_by_value keys(%WORMipports);

	if ($sortbytotal) {
		# ipaddrs
		@Allsortedipaddrs  = sort sort_by_total_value keys(%ipaddrs);
		@WORMsortedipaddrs  = sort sort_by_total_value keys(%WORMipaddrs);
		return;
	}
	# ipaddrs
	@Allsortedipaddrs  = sort allipaddrs_by_value keys(%ipaddrs);
	@WORMsortedipaddrs  = sort WORMipaddrs_by_value keys(%WORMipaddrs);
}

sub allipaddrs_by_value 
{
        if ($ipaddrs{$a} < $ipaddrs{$b}) {
                1;
        }  
        elsif ($ipaddrs{$a} == $ipaddrs{$b}) {
                0;
        }
        elsif ($ipaddrs{$a} > $ipaddrs{$b}) {
                -1;
        }
}

sub WORMipaddrs_by_value 
{
        if ($WORMipaddrs{$a} < $WORMipaddrs{$b}) {
                1;
        }  
        elsif ($WORMipaddrs{$a} == $WORMipaddrs{$b}) {
                0;
        }
        elsif ($WORMipaddrs{$a} > $WORMipaddrs{$b}) {
                -1;
        }
}

sub allipports_by_value 
{
        if ($allipports{$a} < $allipports{$b}) {
                1;
        }  
        elsif ($allipports{$a} == $allipports{$b}) {
                0;
        }
        elsif ($allipports{$a} > $allipports{$b}) {
                -1;
        }
}

sub WORMipports_by_value 
{
        if ($WORMipports{$a} < $WORMipports{$b}) {
                1;
        }  
        elsif ($WORMipports{$a} == $WORMipports{$b}) {
                0;
        }
        elsif ($WORMipports{$a} > $WORMipports{$b}) {
                -1;
        }
}


#
# storeEvent(file, prependString, eventString, dupflag)
#
# if dupflag is on, don't print the message if already found
# in database
#
# -E flag triggers use of this, else ignored.
#
# eventString may or may not have a newline.  we strip it if it does.
#
sub
storeEvent
{

	my $eventfile = $_[0];
	my $prependString = $_[1];
	my $eventString = $_[2];
	my $nodups = $_[3];
	my $s;

	$eventString=~ s/\n//g;
	my $notime = $prependString . $eventString;
	$s = "$start_time: " . $notime;

	# if we have seen it before, let's not see it again
	# once a day is enough
	#
	if ($nodups) {
		dbmopen(%emap, $event_dbmfile, 0666);
		if ($emap{$notime}) {
			dbmclose(%emap);
			return;
		}
		else {
			$emap{$notime} = $notime;	
		}
		dbmclose(%emap);
	}

 	open(FILE, ">>$eventfile") || die "can't open $eventfile: $!\n";
 	print FILE "$s\n";
 	close(FILE);
}


# bittorrent 6881..6889
# kazaa 1214
# edonkey 4661, 4662, 4665
# gnutella 6346 6347
# winmx 6699
#
# B - bittorrent
# K - kazaa
# E - edonkey
# G - gnutella
# M - microsoft (virus in fact aimed at Microsoft file share)
# W - winmx
#sub
#idports
#{
#	$port = $_[0];
#
#	if ($port >= 6881 && $port <= 6889) {
#		return("B");
#	}
#	elsif ($port == 1214) {
#		return("K");
#	}
#	elsif ($port == 4661 || $port == 4662 || $port == 4665 ) {
#		return("E");
#	}
#	elsif ($port == 6346 || $port == 6347) {
#		return("G");
#	}
#	elsif ($port == 139 || $port == 135 || $port == 445 ) {
#		return("M");
#	}
#	elsif ($port == 6699 ) {
#		return("W");
#	}
#}

# ignoring emulehit count for now
sub parseAppFlags
{
	$appflags = $_[0];
	$emulehits = $_[1];
	$ipsrc = $_[2];

	#$appflags = hex($_[0]);
	my $string="";
	$portSignature_pot{$ipsrc} = 0;

	if ($appflags & 0x0f) {
		$string = "I";		#/* irc of some sort */
	}
	if ($appflags & 0x10) {
		$string .= "B";		#/* bittorrent */
	}
	if ($appflags & 0x20) {
		$string .= "G";		#/* GNUTELLA */
	}
	if ($appflags & 0x40) {
		$string .= "K";		#/* KAZAA */
	}
	if ($appflags & 0x80 && ($emulehits >= $emulecount)) {
		$string .= "e";		#/* edonkey */
	}
 	if ($appflags & 0x100) {
 		$string .= "P";		#/* honeypot dst */
		$portSignature_pot{$ipsrc} = 1;
 	}
	if ($appflags & 0x200) {
		$string .= "M";		#/* Morpheus */
	}
	if ($appflags & 0x400) {
		$string .= "H";		#/* web server based on tcp src port */
	}
	if ($appflags & 0x800) {
		$string .= "E";	  	# email  based on tcp dst port 
	}
	if ($appflags & 0x1000) {
		$string .= "D";	  	# darknet
		$portSignature_pot{$ipsrc} = 1;
	}
	return($string);
}

sub parseLowerAppTags 
{
	$apptag = $_[0];
	$string = $_[1];

	if ($apptag & 0x1 ) { $string .= "a"; }
	if ($apptag & 0x2 ) { $string .= "b"; }
	if ($apptag & 0x4 ) { $string .= "c"; }
	if ($apptag & 0x8 ) { $string .= "d"; }
	if ($apptag & 0x10 ) { $string .= "e"; }
	if ($apptag & 0x20 ) { $string .= "f"; }
	if ($apptag & 0x40 ) { $string .= "g"; }
	if ($apptag & 0x80 ) { $string .= "h"; }
	if ($apptag & 0x100 ) { $string .= "i"; }
	if ($apptag & 0x200 ) { $string .= "j"; }
	if ($apptag & 0x400 ) { $string .= "k"; }
	if ($apptag & 0x800 ) { $string .= "l"; }
	if ($apptag & 0x1000 ) { $string .= "m"; }
	if ($apptag & 0x2000 ) { $string .= "n"; }
	if ($apptag & 0x4000 ) { $string .= "o"; }
	if ($apptag & 0x8000 ) { $string .= "p"; }
	if ($apptag & 0x10000 ) { $string .= "q"; }
	if ($apptag & 0x20000 ) { $string .= "r"; }
	if ($apptag & 0x40000 ) { $string .= "s"; }
	if ($apptag & 0x80000 ) { $string .= "t"; }
	if ($apptag & 0x100000 ) { $string .= "u"; }
	if ($apptag & 0x200000 ) { $string .= "v"; }
	if ($apptag & 0x400000 ) { $string .= "w"; }
	if ($apptag & 0x800000 ) { $string .= "x"; }
	if ($apptag & 0x1000000 ) { $string .= "y"; }
	if ($apptag & 0x2000000 ) { $string .= "z"; }

	return($string);
}

sub parseUpperAppTags 
{
	$apptag = $_[0];
	$string=  $_[1];

	if ($apptag & 0x1 ) { $string .= "A"; }
	if ($apptag & 0x2 ) { $string .= "B"; }
	if ($apptag & 0x4 ) { $string .= "C"; }
	if ($apptag & 0x8 ) { $string .= "D"; }
	if ($apptag & 0x10 ) { $string .= "E"; }
	if ($apptag & 0x20 ) { $string .= "F"; }
	if ($apptag & 0x40 ) { $string .= "G"; }
	if ($apptag & 0x80 ) { $string .= "H"; }
	if ($apptag & 0x100 ) { $string .= "I"; }
	if ($apptag & 0x200 ) { $string .= "J"; }
	if ($apptag & 0x400 ) { $string .= "K"; }
	if ($apptag & 0x800 ) { $string .= "L"; }
	if ($apptag & 0x1000 ) { $string .= "M"; }
	if ($apptag & 0x2000 ) { $string .= "N"; }
	if ($apptag & 0x4000 ) { $string .= "O"; }
	if ($apptag & 0x8000 ) { $string .= "P"; }
	if ($apptag & 0x10000 ) { $string .= "Q"; }
	if ($apptag & 0x20000 ) { $string .= "R"; }
	if ($apptag & 0x40000 ) { $string .= "S"; }
	if ($apptag & 0x80000 ) { $string .= "T"; }
	if ($apptag & 0x100000 ) { $string .= "U"; }
	if ($apptag & 0x200000 ) { $string .= "V"; }
	if ($apptag & 0x400000 ) { $string .= "W"; }
	if ($apptag & 0x800000 ) { $string .= "X"; }
	if ($apptag & 0x1000000 ) { $string .= "Y"; }
	if ($apptag & 0x2000000 ) { $string .= "Z"; }

	return($string);
}
sub printAppHints
{
	my $fd = $_[0];
	print $fd "\nL7 app flag hints\n";
	print $fd "I - irc\n";
	print $fd "B - bittorrent\n";
	print $fd "G - gnutella\n";
	print $fd "K - kazaa\n";
	print $fd "e - edonkey\n";
	print $fd "M - morpheus\n";
	print $fd "H - web server based on src port (80/443)\n";
	print $fd "E - email based on dst port\n";
 	print $fd "P - honeynet ip dest addr\n";

# GUESS
# END GUESS
}

# given a work weight store a count
# in 10 possible bins to give us a rough
# idea of the work scatter
sub
storeWorkHistogram
{
	my $ip = $_[0];
	my $work = $portSignature_work{$ip}; 
 	my $workindex = $work - ($work % 10);
	$workCount{$workindex} = $workCount{$workindex} + 1;
}

sub
printWorkHistogram
{
	my $fd = $_[0];
	my $index = 0;

	print $fd "\nwork histogram:\n";
	for ($index = 0; $index <= 100;  $index = $index + 10) {
		print $fd "$index: $workCount{$index}\n";
	}
}

sub
storeWork2Histogram
{
	my $ip = $_[0];
	my $work = $portSignature_work2{$ip}; 
 	my $workindex = $work - ($work % 10);
	$work2Count{$workindex} = $work2Count{$workindex} + 1;
}


sub
printWork2Histogram
{
	my $fd = $_[0];
	my $index = 0;

	print $fd "\nwork2 (s+r/t) histogram:\n";
	for ($index = 0; $index <= 100;  $index = $index + 10) {
		print $fd "$index: $work2Count{$index}\n";
	}
}

sub
storeWork5Histogram
{
	my $ip = $_[0];
	my $work = $portSignature_work5{$ip}; 
 	my $workindex = $work - ($work % 10);
	$work5Count{$workindex} = $work5Count{$workindex} + 1;
#print "GOO $ip $work $workindex $workCount5{$workindex}\n";
}

sub
printWork5Histogram
{
	my $fd = $_[0];
	my $index = 0;

	print $fd "\nwork5 (total sent/total) histogram:\n";
	for ($index = 0; $index <= 100;  $index = $index + 10) {
		print $fd "$index: $work5Count{$index}\n";
	}
}

sub usage
{
	print "syntax: tcpworm.pl [-i -t -d -N -p(P) port_report.txt -D dbfile -E eventfile -w workweight -h/H homenet/mask -4 port inputs > output_file\n";
	print "\t inputs may be 1 input syn file or a list of files or -z to supply a directory of files\n";
	print "\t\t - 1 file gives us normal sample output, 2 or more files gives us summarization mode\n";
	print "\t-d debug on\n";
	print "\t-e topn_count - email style report, input is emaildump.txt, or directory of emaildump.txt if -z\n";
	print "\t-N dns lookup on\n";
	print "\t-p short summary file version of port signature report here\n";
	print "\t-P same as -p, but no P2P \"application flags\" info\n";
	print "\t-D db file to place new port sigs in \n";
	print "\t-E event file to place event messages in\n";
	print "\t-w filter output based on workweight. default is 0. range is 0..100\n";
	print "\t-h filter and classify output based on subnet/mask (use slash notation).  no default.\n";
	print "\t-H filter and classify output based on IP not in subnet/mask (use slash notation).  no default.\n";
	print "\t-4 port - limit outputs to this port - summarization mode.\n";
	print "\t-i sort by instance as opposed to ip src ascending value - summarization mode.\n";
	print "\t-t sort ipaddrs by total pkt count - 30 sec/summarization (not long 30 second).\n";
	print "\t-z process an entire directory of input files - summarization mode.\n";
	print "\t-S port signature log file.\n";
	print "\t-W work weight for filtering events at 30 sec period placed in event file.\n";
	print "\t-r netmask - if IP address falls in net/mask and ww > -W value issue event log message. \n"; 
	print "\t-R netmask - if IP address falls in net/mask and has P/D flag set, issue event log message \n";
	exit(1);
}

# insubnet
#
# note: input param netmask may have these forms
#	1.0.0.0/8
#	1.0.0.0/8,2.0.0.0/8
#	1/8
#	10.10/16
#	1/8,2/8,200.1.2.0/24
#
# a comma-separated list is possible.
# so called netmask variable below is really net/mask.
#
# compare a subnet/mask to an ip
# return:
#	1 if ip within subnet/s
#	0 otherwise

sub insubnet
{
	my $netmask = $_[0];
	my $ip = $_[1];
	my $hexmask;
	my $rnet1;
	my $rnet2;
	my $net1;
	my $net2;
	my $index=0;
	
	my @nets = split(",", $netmask);

	while ($nets[$index]) {

		my @words = split("\/", $nets[$index]);
		my $subnet = $words[0];
		my $mask = $words[1];

		# fix subnet if too short - we need full dot quad notation
		my @quads=split('\.',$subnet);
		my $suffixquads = 3 - $#quads;  # number of .zero quads needed at end
		# append enough dot quads if needed
		if ($suffixquads == 3) {
			$subnet = $subnet . ".0.0.0";
		}
		elsif ($suffixquads == 2) {
			$subnet = $subnet . ".0.0";
		}
		elsif ($suffixquads == 1) {
			$subnet = $subnet . ".0";
		}

		# if slash notation
		# 1. convert the specified mask
		# to a hex value (a real mask)
		$hexmask = 0xffffffff << (32 - $mask);
		# limit to 32 bits please
		$hexmask = $hexmask & 0xffffffff;

		# ip address based netmask as opposed to slash
		#$hexmask = unpack('N', inet_aton($mask));

		# 2. convert the ipaddr to a host int
		$net1 = unpack('N', inet_aton($ip));
		$net2 = unpack('N', inet_aton($subnet));
		# printf ("inputip: %x inputnet %x mask %x\n", $net1, $net2, $hexmask);

		# 3. compare the results
		$rnet1 = $net1 & $hexmask;
		$rnet2 = $net2 & $hexmask;

		# print "$rnet1, $rnet2:", "$rnet1 == $rnet2\n"; 
		if ($rnet1 == $rnet2) {
			return (1);
		}
		$index++;
	}
	return(0);
}

# weak check to see if timestamp is in name
sub check_ts
{
	my @DAYS=
	my $candidate_ts = $_[0];
	my @TSPARTS = split("_", $candidate_ts);
	if ($TSPARTS[0] eq "Sun") {
		return(1);
	}
	if ($TSPARTS[0] eq "Mon") {
		return(1);
	}
	if ($TSPARTS[0] eq "Tue") {
		return(1);
	}
	if ($TSPARTS[0] eq "Wed") {
		return(1);
	}
	if ($TSPARTS[0] eq "Thu") {
		return(1);
	}
	if ($TSPARTS[0] eq "Fri") {
		return(1);
	}
	if ($TSPARTS[0] eq "Sat") {
		return(1);
	}
	return(0);
}

# internal dns cache if reverse dns in use.
#
# return dns from ip within timeout period of 5 seconds
# dnscache{} is associative array so that we can
# 	cache our own dns lookups
sub gethostip {
	my $ipaddr = $_[0];
	my $dnsname;

	if ($dnscache{$ipaddr}) {
		if ($debug) {
 			print "cached $ipaddr: $dnscache{$ipaddr}\n";
		}
		return ($dnscache{$ipaddr});
	}


	if ($debug) {
		print "gethostip: lookup of $ipaddr\n";
	}
  	eval {
		$SIG{ALRM} = \&sigalrm; 
 		alarm(5);
		eval {
			$dnsname = gethostbyaddr(inet_aton($ipaddr), 2);
		};
		alarm(0);
 	};
	alarm(0);
#print "ip: $ipaddr dns: $dnsname\n";
	if ($dnsname) {
		$dnscache{$ipaddr} = $dnsname;
		if ($debug) {
			print "gethostip: found $dnsname\n";
		}
	}
	else {
		$dnscache{$ipaddr} = $ipaddr;
		if ($debug) {
			print "gethostip: did not find $ipaddr\n";
		}
	}
	return($dnsname);
}

sub sigalrm {
	#alarm(0);
	#print "ALARM TIMEOUT\n";
}

#########################################################################
# summarization code from here on out
#
sub sumworm
{
	my $ipsrc=$_[0];		# ip src
	my $syns=$_[1];			
	my $synacks=$_[2];			
	my $fins=$_[3];
	my $finssent=$_[4];
	my $rsts=$_[5];			# 
	my $tcpfrom=$_[6];
	my $tcpto=$_[7];
	my $total=$_[8];
	my $icmpcount=$_[9];
	my $ports=$_[10];
	my $portstring=$_[11];
	my $workweight=$_[12];
	my $appflags=$_[13];
	my $apptagsLower = $_[14];
	my $apptagsUpper = $_[15];
	my $emulehits=$_[16];
	my $ipdsts=$_[17];
	my $tcpdsts=$_[18];
	my $eflag=$_[19];
	my $wflag=$_[20];
	my $oflag=$_[21];
	my $rflag=$_[22];
	my $mflag=$_[23];
	my $p25_syns=$_[24];
	my $p25_ww=$_[25];
	my $timestamp = $curts;		# curts is a global
	my $curport;
	my $curportcount;
	  
	# handle portstring first as we may filter on it
	# two parts to this:
	#	1. check to see if port matches possible supplied
	#		port filter parameter
	#	2. if that check was not make, or it passed store port info
     	my @ps = split(/,/,$portstring);
	if ($portlimit != -1) {
		my $sawport = 0;
		my $pindex = 0;
		# loop thru port string to see if filter is mentioned
		for ($i = 0; $i < $ports; $i++) {
			$pindex = $i * 2;
			$curport = $ps[$pindex]; 
			if ($curport == $portlimit) {
				$sawport = 1;
			}
		}
		# ignore tuple if port of interest not mentioned
		if ($sawport == 0) {
			return;
		}
	}

	# part 2 for ports: loop through ports and store info.
	my $pindex = 0;
	for ($i = 0; $i < $ports; $i++) {
		$pindex = $i * 2;
		$curport = $ps[$pindex]; 	# the port
		$curportcount = $ps[$pindex+1]; # the pkt count
		if ($sumPS{$ipsrc}->{$curport}) {
			$sumPS{$ipsrc}->{$curport} += $curportcount;
		}
		else {
			$sumPS{$ipsrc}->{$curport} = $curportcount;
		}
	}

        $appflags = hex($appflags);	# convert to hex mask
        $apptagsLower = hex($apptagsLower);
        $apptagsUpper = hex($apptagsUpper);

	# hex flags for dealing with worm flags
	# W = 0x1
	# w = 0x2
	# E = 0x4
	# O = 0x8
	# R = 0x10
	# M = 0x20
	$wormmask = 0;
	if ($wflag eq "W") {
		$wormmask |= 0x1;
	}
	elsif ($wflag eq "w") {
		$wormmask |= 0x2;
	}
	if ($eflag) {
		$wormmask |= 0x4;
	}
	if ($oflag) {
		$wormmask |= 0x8;
	}
	if ($rflag) {
		$wormmask |= 0x10;
	}
	if ($mflag) {
		$wormmask |= 0x20;
	}

	# seen it before or first time init
	if ( $sumSyns{$ipsrc} ) {
		# summations
		$sumInstances{$ipsrc}++;	# how many times seen 
		$sumSyns{$ipsrc} += $syns;
		$sumSynacks{$ipsrc} += $synacks;
		$sumFins{$ipsrc} += $fins;
		$sumFinssent{$ipsrc} += $finssent;
		$sumRsts{$ipsrc} += $rsts;
		$sumTcpfrom{$ipsrc} += $tcpfrom;
		$sumTcpto{$ipsrc} += $tcpto;
		$sumTotal{$ipsrc} += $total;
		$emailTotal{$ipsrc} += $p25_syns;
		$emailWW{$ipsrc} += $p25_ww;		# will produce average
		$sumIcmpcount{$ipsrc} += $icmpcount;
		$sumPorts{$ipsrc} += $ports;
		$sumEmulehits{$ipsrc} += $emulehits;
		$sumIPdsts{$ipsrc} += $ipdsts;
		$sumTCPdsts{$ipsrc} += $tcpdsts;
		$sumWorkweight{$ipsrc} += $workweight;	# for computing average
		# keep max and min workweights seen
		if ($workweight > $sumWorkweightMax{$ipsrc}) {
			$sumWorkweightMax{$ipsrc} = $workweight;	
		} 
		if ($workweight < $sumWorkweightMin{$ipsrc}) {
			$sumWorkweightMin{$ipsrc} = $workweight;	
		} 

		# assignments

		# note how we must take files in time
		# order for this timestamp variable to work 
		$sumLastTS{$ipsrc} = $cur_ts;	
		$sumAppflags{$ipsrc} |= $appflags;
		$sumAppTagsLower{$ipsrc} |= $apptagsLower;
		$sumAppTagsUpper{$ipsrc} |= $apptagsUpper;
		$sumWormmask{$ipsrc} |= $wormmask;
	}
	# first time init
	else {
		$sumInstances{$ipsrc} = 1;
		$emailTotal{$ipsrc} = $p25_syns;
		$emailWW{$ipsrc} = $p25_ww;		# will produce average
		$sumSyns{$ipsrc} = $syns;
		$sumSynacks{$ipsrc} = $synacks;
		$sumFins{$ipsrc} = $fins;
		$sumFinssent{$ipsrc} = $finssent;
		$sumRsts{$ipsrc} = $rsts;
		$sumTcpfrom{$ipsrc} = $tcpfrom;
		$sumTcpto{$ipsrc} = $tcpto;
		$sumTotal{$ipsrc} = $tcpfrom + $tcpto;
		$sumIcmpcount{$ipsrc} = $icmpcount;
		$sumPorts{$ipsrc} = $ports;
		$sumEmulehits{$ipsrc} = $emulehits;
		$sumIPdsts{$ipsrc} = $ipdsts;
		$sumTCPdsts{$ipsrc} = $tcpdsts;
		$sumFirstTS{$ipsrc} = $cur_ts;
		$sumAppflags{$ipsrc} = $appflags;
		$sumAppTagsLower{$ipsrc} = $apptagsLower;
		$sumAppTagsUpper{$ipsrc} = $apptagsUpper;
		$sumWormmask{$ipsrc} |= $wormmask;
		$sumWorkweight{$ipsrc} = $workweight;	# for computing average
		$sumWorkweightMax{$ipsrc} = $workweight;	
		$sumWorkweightMin{$ipsrc} = $workweight;
	}
}

sub buildEwormFlags
{
	my $eflags = $_[0];
	my $string="";

	if ($eflags & 0x4) {
		$string = "E";		
	}
	# if W is set, we use that
	if ($eflags & 0x1) {
		$string .= "W";		
	}
	elsif ($eflags & 0x2) {
		$string .= "w";		
	}
	if ($eflags & 0x8) {
		$string .= "O";		
	}
	if ($eflags & 0x10) {
		$string .= "R";	
	}
	if ($eflags & 0x20) {
		$string .= "M";		
	}
	return($string);
}

# sum report
#
# this is like the port signature report BUT the format is slightly different
#
# for the first iteration:
# 1. ip addresses are sorted as before
# 2. port sig is NOT sorted in ascending order, but sorted in terms of
#	top ports used.  e.g., if port 65555 has the most pkts, it will
#	be shown first.  
#
# for each tuple, 2 lines:
#
# ip addr   wflags  appflags  workweight  sa/s  l3d/l4d: ports:portsig in sorted order 
# 	total instances: start time: stop: time
#
# wflags is or'red together
# appflags is or'red together
# workweight is min/avg/max
# sa/s is average
# l3d/l4d are averages
# average pkts sent and recv
# ports: total ports in port sig
# portsig: is topN basically, shown in terms of port, total pkt count 
#	don't show more than 10.
#
# 2nd iteration like first iteration but we sort on instance count.
# 

sub sumguide
{
print <<END_OF_TEXT;
------------------------------------------------------------------------------
port summarization guide:
-------------------------

instances count 30-second sample periods
Line 1 shows the IP src and various flags and counters.
Line 2 shows time related information, instance count, start and stop times.
Line 3 shows the top 10 dest port samples in terms of pkt counts.  
On line 1 the worm flags are the logical OR of all samples.
          the app flags are the logical OR of all samples.
          the work weight is given as a min, average, max.
          the sa/s field is given as an average for all samples.
	  L3 dst and L4 dst counts are averaged across all samples.
	  syns/fins/resets are shown and are averaged across all samples.
          total pkts sent and recv are averaged across all samples.
On line 3 destination port counts are sorted by count.
	  The ports are thus a top N of sample ports seen.
	  No more than 10 port, count tuples are shown.
	  The count is the total pkts sent to a port bucket across all samples.
END_OF_TEXT
}

sub sumstart
{
	my $ipaddrcount = $_[0];
	print "tcp syn tuple summarization:\n";
	print "----------------------------\n";
        print "start time: $start_time\n";     
	print "\tswitches if any:\n"; 
	if ($sortbytotal) {
		print "\t-t(sort by total pkts)\n";
	}
	if ($sortbyinstance) {
		print "\t-i(sort by total instances)\n";
	}
	if ($portlimit > -1) {
		print "\t-4(port filter): $portlimit\n";
	}
	if ( $home_netmaskflag == 1) {
		print "\t-h(net filter): $home_netmask\n";
	}
	if ( $home_netmaskflag == 2) {
		print "\t-H(!net filter): $home_netmask\n";
	}
	if ( $filter_workweight > 0) {
		print "\t-w(work filter): $filter_workweight\n";
	}
	printf ("\ttotal ipsrcs: %u\n\n", $ipaddrcount);
}

sub suminstances_by_value 
{
        if ($sumInstances{$a} < $sumInstances{$b}) {
                1;
        }  
        elsif ($sumInstances{$a} == $sumInstances{$b}) {
                0;
        }
        elsif ($sumInstances{$a} > $sumInstances{$b}) {
                -1;
        }
}

sub sort_by_total_value
{
        if ($sumTotal{$a} < $sumTotal{$b}) {
                1;
        }  
        elsif ($sumTotal{$a} == $sumTotal{$b}) {
                0;
        }
        elsif ($sumTotal{$a} > $sumTotal{$b}) {
                -1;
        }
}

# sort by total email syns
sub sort_by_total_email
{
        if ($emailTotal{$a} < $emailTotal{$b}) {
                1;
        }  
        elsif ($emailTotal{$a} == $emailTotal{$b}) {
                0;
        }
        elsif ($emailTotal{$a} > $emailTotal{$b}) {
                -1;
        }
}
	   
# print out summarization report
sub sumreport
{
	my $ipaddrcount = 0;

	# decide sorting scheme to use
	if ($emailsyns) {
		@sports = sort sort_by_total_email keys(%sumInstances);
	}
	elsif ($sortbyinstance) {
		@sports = sort suminstances_by_value keys(%sumInstances);
	}
	else {
		if ($sortbytotal) {
			@sports = sort sort_by_total_value keys(%sumTotal);
		}
		else {
			@sports = sort p_by_ipkey keys(%sumSyns);
		}
	}

	# total number of ip sources seen
        $ipaddrcount = scalar @sports;

	# if -e form, limit results to topn requested
	if ($emailsyns) {
		if ($emailcount) {
			$ipaddrcount = $emailcount;
		}
	}

	sumstart($ipaddrcount);

	printf("-----------------------------------------------------------------------\n");
	printf("%-16.16s %-8.8s %-5.5s %s %s %s %s %s\n",
		"ip_src", "flags", "apps", "ww:min/avg/max", "sa/s", "l3d/l4d",  "syn/fin/rst",
		"tsent/trecv");
	printf("\t:dns: dnsname if any\n");
	printf("\t:instances:start:stop\n");
	if ($emailsyns) {
		printf("\t:email syn total:email syn avg:email ww avg\n");
	}
	printf("\ttop-N port tuples (dst port, pkt count)...\n\n");
		
	# loop through sorted ip addresses
	# 	print out work weight info
	#	wflags and appflags
	#	print out instances/times
	$i = 0;
	for ($i = 0; $i < $ipaddrcount; $i++) {
		my $ipsrc = $sports[$i];
		my $dnsname = gethostip($ipsrc);       
		my $instances = $sumInstances{$ipsrc};
		if ($instances == 0) {
			$instances = 1;
		}
		my $ww = $sumWorkweight{$ipsrc} / $instances;
		my $wwmax = $sumWorkweightMax{$ipsrc};
		my $wwmin = $sumWorkweightMin{$ipsrc};
		my $ap = $sumAppflags{$ipsrc};
		my $eh  = $sumEmulehits{$ipsrc};
		my $appflags = parseAppFlags($ap, $eh, $ipsrc);
		
		$appflags= parseLowerAppTags($sumAppTagsLower{$ipsrc},$appflags);
		$appflags= parseUpperAppTags($sumAppTagsUpper{$ipsrc},$appflags);
		my $ewormflags = buildEwormFlags($sumWormmask{$ipsrc});
		my $s = $sumSyns{$ipsrc};
		my $sa = $sumSynacks{$ipsrc};
		if ($s == 0) {
			$s = 1;
		}
		my $sas = ($sa * 100.0)/$s;
		# pkts sent by host divided by instance count
		# pkts sent into sample buckets divided by instance count
		my $sent = $sumTcpfrom{$ipsrc}/$instances;
		my $recv = $sumTcpto{$ipsrc}/$instances;
		my $l3count = $sumIPdsts{$ipsrc}/$instances;
		my $l4count = $sumTCPdsts{$ipsrc}/$instances;
		my $syns = $sumSyns{$ipsrc}/$instances;
		my $fins = $sumFins{$ipsrc}/$instances;
		my $resets = $sumRsts{$ipsrc}/$instances;
		if ($emailsyns) {
			$p25syns = int($emailTotal{$ipsrc}/$instances);
			$p25ww = int($emailWW{$ipsrc}/$instances);
		}

		my $sbuf = sprintf("%-16.16s %-8.8s %-5.5s (%3d:%3d:%3d:) %3d: (%u/%u) (%u:%u:%u) (%u:%u)",
				"$ipsrc", $ewormflags, $appflags, $wwmin, 
				 $ww, $wwmax, $sas, $l3count, $l4count, $syns, $fins, $rsts, $sent, $recv);
		print "$sbuf\n";
		print "\tdns: $dnsname\n";
		print "\t:$instances: $sumFirstTS{$ipsrc}: $sumLastTS{$ipsrc}:\n";
		if ($emailsyns) {
			print "\temail: syns: $emailTotal{$ipsrc}, synavg: $p25syns, wwavg: $p25ww\n";	
		}
# CROCK   
		
		sumports($ipsrc);
		print "\n";
	}
}

$curip = 0;

sub sumportsby_value 
{
        if ($sumPS{$curip}->{$a} < $sumPS{$curip}->{$b}) {
                1;
        }  
        elsif ($sumPS{$curip}->{$a} == $sumPS{$curip}->{$b}) {
                0;
        }
        elsif ($sumPS{$curip}->{$a} > $sumPS{$curip}->{$b}) {
                -1;
        }
}

sub sumports
{
	my $ipsrc = $_[0];
	my $portkey;
	my $nps="";

	$curip = $ipsrc;
	#my @sp  = sort p_by_key keys %{$sumPS{$ipsrc}};  
	my @sp  = sort sumportsby_value keys %{$sumPS{$ipsrc}};  
        my $ports = scalar @sp;
	# top 10 ports, no more
 	if ($ports > 10) {
 		$ports = 10;
 	}
	for (my $i = 0; $i < $ports; $i++) {
		$nps = $nps . "[$sp[$i], $sumPS{$ipsrc}->{$sp[$i]}]";
	}
	print "\tportuples[$ports]: $nps\n";

	#foreach $portkey (keys %{ $sumPS{$ipsrc} }) {
	#	print "$ipkey: $portkey:", $sumPS{$ipsrc}->{$portkey}, "\n";
	#}
	#$sumPS{$ipsrc}->{$curport} += $curportcount;
}

# GUESS 
#
# classifications
#	worm
#	stworm
#	ipworm
#	portworm
#	p2p
#	server
#
# allowed one guess per ip src at this time
#
# guesstimator stats:
#	percent sigs but no guess/p2psig
#	percent guess/total
#	percent p2psig/total
#	percent p2pguess but no p2psig/total pop

sub
guesstimator
{
	my $ipsrc = $_[0];

	$portSignature_guess{$ipsrc} = "NA";
} 
